// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class body                      fully inline                                |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_body_h
#define included_body_h

#ifdef  USE_BODIES

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_flag_h
#include <tree/flag.h>
#endif

////////////////////////////////////////////////////////////////////////////////

#define NOT_LONGER  16                        // use 5th bit of class flag
#define NOT_SHORTER 32                        // use 6th bit of class flag

namespace fasttree {
  class soul;

  class body : public flag {
    //   class for a body; to be used in N-body codes; FULLY INLINE
  public:
    struct iodata {                           // data used in I/O
      real          MASS;                     //   mass
      vect          TPOS;                     //   true position
      vect          VEL;                      //   velocity
      real          EPS;                      //   gravity softening length
      real          SIZE;  };                 //   size of SPH/sticky particle
  private:
    iodata          IOD;                      // data used in I/O
    mutable indx    LEVEL;                    //   time-step level
    mutable soul   *MYSOUL;                   //   soul and its data
    vect            PPOS;                     //   predicted position
    vect            OACC;                     //   last acceleration
  public:
    // some simple manipulations
    void set_soul          (soul* S)      const { MYSOUL = S; }
    // boolean informations via members
    bool is_sink           ()             const { return  is_set(SINK); }
    bool is_source         ()             const { return IOD.MASS!=zero; }
    bool is_test_body      ()             const { return IOD.MASS==zero; }
    bool is_in_tree        ()             const { return !is_set(NOT_IN_TREE); }
    bool is_sph            ()             const { return  is_set(SPH); }
    bool is_sticky         ()             const { return  is_set(STICKY); }
    bool may_go_longer     ()             const { return !is_set(NOT_LONGER); }
    bool may_go_shorter    ()             const { return !is_set(NOT_SHORTER); }
    // boolean informations via friends
    friend bool is_source     (const body*B) { return B->IOD.MASS!=zero; }
    friend bool is_test_body  (const body*B) { return B->IOD.MASS==zero; }
    friend bool is_sink       (const body*B) { return B->is_set(SINK); }
    friend bool is_in_tree    (const body*B) { return !B->is_set(NOT_IN_TREE); }
    friend bool is_sph        (const body*B) { return B->is_set(SPH); }
    friend bool is_sticky     (const body*B) { return B->is_set(STICKY); }
    friend bool may_go_longer (const body*B) { return !B->is_set(NOT_LONGER); }
    friend bool may_go_shorter(const body*B) { return !B->is_set(NOT_SHORTER); }
    // non-const flag manipulations via members
    void flag_as_sink      ()                   { addflag(SINK); }
    void unflag_sink       ()                   { unflag(SINK); }
    void allow_tree_usage  ()                   { unflag(NOT_IN_TREE); }
    void forbid_tree_usage ()                   { addflag(NOT_IN_TREE); }
    void flag_as_sph       ()                   { addflag(SPH); }
    void unflag_sph        ()                   { unflag(SPH); }
    void flag_as_sticky    ()                   { addflag(STICKY); }
    void unflag_sticky     ()                   { unflag(STICKY); }
    void allow_longer      ()                   { unflag(NOT_LONGER); }
    void forbid_longer     ()                   { addflag(NOT_LONGER); }
    void allow_shorter     ()                   { unflag(NOT_SHORTER); }
    void forbid_shorter    ()                   { addflag(NOT_SHORTER); }
    // non-const data access via members
    iodata& io_data        ()                   { return IOD; }
    real& mass             ()                   { return IOD.MASS; }
    vect& tpos             ()                   { return IOD.TPOS; }
    real& tpos             (const indx i)       { return IOD.TPOS[i]; }
    vect& vel              ()                   { return IOD.VEL; }
    real& vel              (const indx i)       { return IOD.VEL[i]; }
    real& eps              ()                   { return IOD.EPS; }
    real& size             ()                   { return IOD.SIZE; }
    vect& ppos             ()                   { return PPOS; }
    real& ppos             (const indx i)       { return PPOS[i]; }
    vect& oacc             ()                   { return OACC; }
    real& oacc             (const indx i)       { return OACC[i]; }
    indx& level            ()                   { return LEVEL; }
    soul* &mysoul          ()                   { return MYSOUL; }
    vect& acc              ();
    real& pot              ();
    real& rho              ();
    // const data access via friends
    friend const iodata&io_data(const body*const&B) { return B->IOD; }
    friend const real&  mass   (const body*const&B) { return B->IOD.MASS; }
    friend const vect&  tpos   (const body*const&B) { return B->IOD.TPOS; }
    friend const vect&  vel    (const body*const&B) { return B->IOD.VEL; }
    friend const real&  eps    (const body*const&B) { return B->IOD.EPS; }
    friend const real&  size   (const body*const&B) { return B->IOD.SIZE; }
    friend const indx&  level  (const body*const&B) { return B->LEVEL; }
    friend soul* const& mysoul (const body*const&B) { return B->MYSOUL; }
    friend const vect&  ppos   (const body*const&B) { return B->PPOS; }
    friend const vect&  oacc   (const body*const&B) { return B->OACC; }
    friend const vect&  acc    (const body*const&);
    friend const real&  pot    (const body*const&);
    friend const real&  rho    (const body*const&);
    friend const real   energy (const body*const&);
    friend const amom   angmom (const body*const&);
  };
}
//////////////////////////////////////////////////////////////////////////////////
// class body

#ifndef included_soul_h
#include <tree/soul.h>
#endif

namespace fasttree {
  inline       vect& body::acc()                   { return MYSOUL->ACC; }
  inline       real& body::pot()                   { return MYSOUL->POT; }
  inline       real& body::rho()                   { return MYSOUL->RHO; }
  inline const vect& acc   (const body*const&B) { return acc(B->MYSOUL); }
  inline const real& pot   (const body*const&B) { return pot(B->MYSOUL); }
  inline const real& rho   (const body*const&B) { return rho(B->MYSOUL); }
  inline const real  energy(const body*const&B) { return pot(B)+0.5*norm(vel(B));}
  inline const amom  angmom(const body*const&B) { return tpos(B) ^ vel(B); }
}
//////////////////////////////////////////////////////////////////////////////////
#endif // USE_BODIES
#endif // included_body_h
