// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// See the detailled comments for how to use this code. Some abbreviations:    |
//                                                                             |
// R:       return value                                                       |
// I:       function argument is input                                         |
// O:       function argument is output                                        |
// I/O:     function argument is input & output                                |
// [I]:     function argument is optional                                      |
//                                                                             |
// Note that in case of optional function arguments, one may only omit the     |
// last one (and the second but last if the last is already omitted ...).      |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// Note that it is often more convenient to use the wrapper provided by the    |
// classes defined in files yanc.h and ssgs.h                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_nbdy_h                    // ensure this file is seen once    |
#define included_nbdy_h 1                  // only by the compiler             |
extern "C" {                               //                                  |
#ifndef included_time_h                    //                                  |
#include <time.h>                          // CPU timing                       |
#define included_time_h                    //                                  |
#endif                                     //                                  |
                                           //                                  |
#ifndef included_math_h                    //                                  |
#include <math.h>                          // standard math library            |
#define included_math_h                    //                                  |
#endif                                     //                                  |
}                                          //                                  |
#ifndef included_tree_h                    //                                  |
#include <tree.h>                          // the tree code                    |
#endif                                     //                                  |
                                           //                                  |
#ifndef included_step_h                    //                                  |
#include <tree/step.h>                     // the time stepping schemes        |
#endif                                     //                                  |
                                           //                                  |
#ifndef included_pext_h                    //                                  |
#include <tree/pext.h>                     // an external potential            |
#endif                                     //                                  |
//-----------------------------------------------------------------------------+
namespace fasttree {                       // class nbody is in fasttree       |
  class nbody :                            // a tree-based N-body code         |
    protected tree,                        // derived from class tree          |
    protected BlockStep                    // as well as from class BlockStep  |
  {                                        //                                  |
  public:                                  //                                  |
    enum soft_type {                       // type of softening                |
      global_fixed        = 0,             // globally time-constant           |
      individual_fixed    = 1,             // individual but time-constant     |
      individual_adaptive = 2              // individual and adaptive          |
    };                                     //                                  |
//-----------------------------------------------------------------------------+
//                                                                             |
//   1 CONSTRUCTION & RE-SETTING OF PARAMETERS                                 |
//   =========================================                                 |
//                                                                             |
    nbody(      body*,                     // I:   array with bodies           |
	  const unsigned,                  // I:   N = # bodies                |
	  const soft_type,                 // I:   type of softening           |
	  const unsigned,                  // I:   Nsoft                       |
	  const unsigned,                  // I:   Nref                        |
	  const real,                      // I:   eps                         |
	  const real,                      // I:   f                           |
	  const real     =two,             // [I]: tol^2                       |
	  const real     =half,            // [I]: theta                       |
	  const unsigned =8,               // [I]: Ncrit                       |
	  const kern_type=p1,              // [I]: softening kernel            |
	  const potext*  =0);              // [I]: P_ex                        |
//                                                                             |
//   the meaning of some of the parameters are as follows:                     |
//                                                                             |
//   Nsoft    is the number of bodies within each bodies softening volume;     |
//            to be used only for adapting individual softening lengths        |
//   Nref     is the critical cell size for density estimation in time-step    |
//            scheme 2 (see below)                                             |
//   Ncrit    is the number of bodies a cell must have before it is splitted   |
//            in tree::grow(); Sensibly, Ncrit <= Nref.                        |
//   eps      is either the globally constant softening length                 |
//            or the maximum allowed softening length, in the case of          |
//            individual softening lengths                                     |
//   f        if the time-step controlling factor                              |
//            the sign of f controls which stepping scheme is used:            |
//            for f < 0, the scheme described in 2.1 is used                   |
//            for f > 0, the scheme described in 2.2 is used                   |
//   tol      is a parameter used for scheme 1 only (see below)                |
//                                                                             |
    void reset_softening(                  // resets softening parameters      |
			 const soft_type,  // I:   type of softening           |
			 const unsigned,   // I:   Nsoft                       |
			 const real,       // I:   eps                         |
			 const kern_type); // [I]: softening kernel            |
//                                                                             |
    void reset_opening  (                  // resets opening parameter         |
			 const real) const;// I:   theta                       |
//                                                                             |
    void reset_Ncrit    (                  // resets Ncrit                     |
			 const unsigned);  // I:    Ncrit                      |
//                                                                             |
    void reset_stepping (                  // resets stepping parameters       |
			 const real);      // I:   f                           |
//-----------------------------------------------------------------------------+
//                                                                             |
//  2 ON THE TIME-STEPPING SCHEMES                                             |
//  ==============================                                             |
//                                                                             |
//  There are two possible options, the nearly time symmetric and the exactly  |
//  time symmetric scheme of Dehnen (2001b). Both use a block step scheme with |
//  hierarchically nested time steps that differ by a factor of two between    |
//  two levels of the hierarchy.                                               |
//                                                                             |
//  2.1 The nearly time symmetric scheme                                       |
//  ------------------------------------                                       |
//  This scheme is based on an ESTIMATE, made from the first half of each      |
//  step, for the relative change in each bodies acceleration. The idea is to  |
//  limit this change to be less than f:                                       |
//                                                                             |
//         |delta acc| < f |acc|                                               |
//                                                                             |
//  Besides computing the acceleration at the beginning and end of each step,  |
//  which is necessary for the leap-frog time integrator, the code also        |
//  computes it in the MIDDLE, to (i) check the criterion and (ii) to use it   |
//  in the time integration in case a shorter step is required.                |
//  There is also another parameter, tol, which can be used to reduce the      |
//  of these extra computations of acceleration in the middle of each step:    |
//  Only when                                                                  |
//                                                                             |
//         |delta acc| < f * tol |acc|                                         |
//                                                                             |
//  will the criterion actually be checked. Thus tol=2 results in the strict   |
//  case with checking every time, while tol=1 means no checking in the middle |
//  of each step. The default value, tol=sqrt[2], controls half the bodies.    |
//                                                                             |
//  2.2 The exactly time symmetric scheme                                      |
//  -------------------------------------                                      |
//  This scheme uses the mass density rho to assign time steps, according to   |
//                                                                             |
//         tau < f / rho                                                       |
//                                                                             |
//  where rho is estimated in the middle of each step (which makes the scheme  |
//  time symmetric) as the mean density of the smalles cell that (i) contains  |
//  the body in question and (ii) at least Nref bodies in total.               |
//  With this scheme, no acceleration is computed without being used in the    |
//  time integration.                                                          |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//   3 PREPARATION FOR TIME INTEGRATION                                        |
//   ==================================                                        |
//                                                                             |
// Before we can start the time integration, we have                           |
// 1. in the case of adaptive individual softening lengths: to determine the   |
//    individual eps_i for each body;                                          |
// 2. to pre-compute the accelerations at time = t_ini;                        |
// 3. to determine the time-step levels for each body.                         |
//                                                                             |
    void prepare(const real,               // I:   t_ini                       |
		 const int,                // I:   h0 =-log2(tau_max)          |
		 const indx,               // I:   Nsteps                      |
		 const bool = true,        // I:   need to set levels?         |
		 const bool = true,        // I:   need to set eps_i?          |
		       vect* = 0);         // [I]: center of root              |
//                                                                             |
// NOTE that if the accelerations and/or softening lengths are non-zero FOR ALL|
// bodies, we won't pre-compute / determine them.  Use this latter possibility |
// ONLY when re-starting an old simulation that has been interupted (but even  |
// then you must call nbody::prepare()).                                       |
//                                                                             |
// The meaning of some of the parameters are as follows:                       |
//                                                                             |
//   t_ini:      initial global time (preferrable an integer)                  |
//   h0:         determines longest time step: tau_max = (1/2)^h0              |
//   Nstps:      # of time steps, differ by factor 1/2 from one another        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//   4 TIME INTEGRATION                                                        |
//   ==================                                                        |
//                                                                             |
    void long_step(                        // perform one long step            |
		   const unsigned = 0u,    // [I]: #subsequent tree-reuses     |
		         vect*    = 0);    // [I]: center of root              |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//   5 STATISTIC & OUTPUT                                                      |
//   ====================                                                      |
//                                                                             |
// In addition to the few routines listed here, all public member functions of |
// class tree, a base class of class nbody, are available, see file tree.h     |
    enum step_crit {
      dacc = 0,
      rhos = 1
    };
    void        estimate_mass_densities();         // estimate every bodies rho
    void        estimate_surf_densities();         // estimate every bodies SD 
    void        dump_bodies     (std::ostream&) const; // dump body data, ASCII
    void        dump_steps      (std::ostream&) const; // dump steps data, ASCII
    void        long_stats      (std::ostream&) const; // statistic of long_step
    void        short_stats     (std::ostream&) const; // statistic of long_step
    void        short_stat_head (std::ostream&) const; // statistic of long_step
    void        short_stat_line (std::ostream&) const; // line of proper size
    const step_crit& stepping   ()         const;  // actual stepping criteriom
    const real& time            ()         const;  // actual time
    const long unsigned& N_forces()        const;  // # forces in long_step()
    const real  kin_energy      ()         const;  // kinetic energy
    const real  pot_energy      ()         const;  // potential energy
    const real  total_energy    ()         const;  // total energy
    const amom& total_angmom    ()         const;  // total angular momentum
    const vect& total_momentum  ()         const;  // total momentum
    const vect& center_of_mass  ()         const;  // center of mass
    const real& kin_energy      (const indx, const indx) const; // K_ij
    const real& pot_energy      (const indx, const indx) const; // W_ij
    const real  as_angmom       (const indx, const indx) const; // anti-sym AM
    const real& cpu_predict     ()         const;  // time for pred positions
    const real& cpu_build       ()         const;  // time for tree::build
    const real& cpu_tree_force  ()         const;  // time for tree forces
    const real& cpu_ext_force   ()         const;  // time for external forces
    const real  cpu_force       ()         const;  // time for all forces
    const real& cpu_density     ()         const;  // time for density estimates
    const real& cpu_softening   ()         const;  // time for setting eps_i
    const real& cpu_longstep    ()         const;  // time for long_step()
    const real& cpu_total       ()         const;  // total time so far
    const real& diagnose_time   ()         const;  // time of last diagnose

//-----------------------------------------------------------------------------+
//                                                                             |
//   VII. PRIVATE MEMBERS                                                      |
//   ====================                                                      |
//                                                                             |
//   These are indeed private, and hence none of your business!                |
//                                                                             |
//===================NO PUBLIC ACCESS BEYOND THIS POINT ========================
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
  private:
    enum tree_use {
      un_used  = 0,
      re_used  = 1,
      built    = 2
    };
    typedef real tensor[NDIM][NDIM];             // not necessarily symmetric

    soft_type      SOFTENING;                    // type of softening method
    step_crit      MYSCHEME;                     // time stepping scheme
    const potext  *PEX;                          // external potential
    bool           READY;                        // are we ready to go?
    uint           NSOFT;                        // # bodies in softening volume
    uint           NREF;                         // # bodies in cell for rho est.
    uint           NCRIT;                        // max # bodies in cell
    real           EMAX;                         // max softening length
    real           TINI;                         // initial simulation time
    real           FAC,TQ;                       // time-step contolling factors
    mutable real   CPU_PREDICT,                  // time for advancing Xs
                   CPU_BUILD,                    // time for tree::build & re_use
                   CPU_GRAV,                     // time for tree::forces & gravs
                   CPU_PEX,                      // time for external potential
                   CPU_DENS,                     // time for density estimates
                   CPU_EPSI,                     // time for setting eps_i
                   CPU_STEP,                     // total time for a longstep
                   CPU_TOTAL;                    // total time so far
    mutable bool   DIAG;                         // diagnose up-to-date ?
    mutable real   M,Ktot,Utot,TU;               // mass, kin & pot E, vir ratio
    mutable amom   L;                            // total angular momentum
    mutable tensor KT,WT;                        // kin & pot energy, AM tensors
    mutable vect   CMX,CMV;                      // center of mass pos & vel
    mutable long unsigned N_FORCES;
    // general routines
    void           set_pposs();
    void           set_allow_tree();
    tree_use       set_tree(const tree_use, vect* =0);
    void   	 setup_eps(body*) const;
    void           set_softening_lengths();
    bool           may_go_longer(const body*) const;
    void           set_forces();
    void           record_cpu        (clock_t&, real&) const;
    void           reset_cpus        () const;
    void           diagnose          () const;
    void           update_diagnostics() const;
    // routines to use with rhos scheme
    bool           test_for_longer(const indx, const body*) const;
    bool           test_for_shorter(const indx, const body*) const;
    tree_use       smallest_step_rhos(const indx, const tree_use, vect*);
    void           estimate_mass_density();
    // routines to use with dacc scheme
    bool           step_too_long(const body*) const;
    too_short      step_too_short(const body*) const;
    uint           set_sink_flags  (const indx);
    tree_use       smallest_step_dacc(const indx, const tree_use, vect*);
  };
  //----------------------------------------------------------------------------
  inline void nbody::reset_opening(const real th)const{ tree::reset_opening(th); }
  inline void nbody::reset_Ncrit(const uint nc) { NCRIT = max(1u,nc); }
  inline void nbody::dump_steps(std::ostream& to)  const {  BlockStep::dump(to); }
  inline const nbody::step_crit& nbody::stepping()const{ return MYSCHEME; }
  inline const real& nbody::time           () const { return BlockStep::time(); }
  inline const real& nbody::cpu_predict    () const { return CPU_PREDICT; }
  inline const real& nbody::cpu_build      () const { return CPU_BUILD; }
  inline const real& nbody::cpu_tree_force () const { return CPU_GRAV; }
  inline const real& nbody::cpu_ext_force  () const { return CPU_PEX; }
  inline const real  nbody::cpu_force      () const { return CPU_GRAV+CPU_PEX; }
  inline const real& nbody::cpu_density    () const { return CPU_DENS; }
  inline const real& nbody::cpu_softening  () const { return CPU_EPSI; }
  inline const real& nbody::cpu_longstep   () const { return CPU_STEP; }
  inline const real& nbody::cpu_total      () const { return CPU_TOTAL; }
  inline const long unsigned& nbody::N_forces() const { return N_FORCES; }
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_nbdy_h
