// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// bman.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class manager<BOX>                      fully inline                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_bman_h
#define included_bman_h

namespace fasttree {

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::manager<BOX>                                               
  //                                                                            
  // - allocates boxes in up to N_BLOCKS blocks                                 
  // - gives boxes out either singly or in small blocks                         
  // - provides iterators for sequential access to all boxes given out sofar    
  //                                                                            
  // In order to (i) guarantee the correct amount of boxes (not known a priori) 
  // and (ii) not to waste memory by allocating too many, we use the following  
  // simple strategy, similar to that of std::string.                           
  //                                                                            
  // We first allocate N_body/4 boxes, assuming that in almost any case we need 
  // at least as many (however this is easily changed). Once these are consumed,
  // we use the fraction of bodies already handled, to estimate the remaining   
  // number of boxes needed, and allocate the corresponding amount (plus a small
  // margin for security). If this still wasn't enough, we repeat the process   
  // up to N_BLOCKS times. In almost all cases, repetition is not required.     
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
#define N_BLOCKS 10                                // # blocks hold             
  template<class BOX>
  class manager {
  private:
    ////////////////////////////////////////////////////////////////////////////
    // 1. member types of class fasttree::manager                               
    ////////////////////////////////////////////////////////////////////////////
    // 1.1 class fasttree::manager::block                                       
    //     - allocates a large block of boxes                                   
    //     - gives boxes out either singly or in small blocks                   
    //     - provides pointes to first and last box given out sofar             
    ////////////////////////////////////////////////////////////////////////////
    class block {
    private:
      uint NUSED;                                  // # boxes used              
      uint NALLOC;                                 // # boxes allocated         
      BOX  ZEROTH[2];                              // dummy to guarantee last   
      BOX *FIRST;                                  // first box in block        
      BOX *END;                                    // beyond last box used      
      BOX *ENDTOT;                                 // beyond last box alloc     
    public:
      block() :                                    // default constructor       
	NUSED(0),                                  //   # boxes used            
	NALLOC(0),                                 //   # boxes available       
	FIRST(ZEROTH+1),                           //   dummy first             
	END(FIRST),                                //   dummy end               
	ENDTOT(FIRST) {}                           //   dummy end_tot           
      ~block()                                     // destructor: de-allocate   
      {                                            // >                         
	if(NALLOC) delete[] FIRST;                 //   IF alloc > de-alloc <   
      }                                            // <                         
      void reset()                                 // destruct & construct      
      {                                            // >                         
	if(NALLOC) {                               //   IF allocate()ed   >     
	  delete[] FIRST;                          //     de-allocate           
	  NUSED  = 0;                              //     # boxes used          
	  NALLOC = 0;                              //     # boxes available     
	  FIRST  = ZEROTH+1;                       //     dummy first           
	  END    = FIRST;                          //     dummy end             
	  ENDTOT = FIRST;                          //     dummy end_tot         
	}                                          //   <                       
      }                                            // <                         
      void allocate(const uint n)                  // allocate N boxes          
      {                                            // >                         
	if(n==0) error(" cannot allocate 0");      //   IF N==0: FATAL ERROR    
	if(NALLOC) delete[] FIRST;                 //   de-alloc if necessary   
	NUSED = 0;                                 //   reset N_used            
	NALLOC= n;                                 //   set N_allocated         
	FIRST = new BOX[NALLOC];                   //   allocate N boxes        
	END   = FIRST;                             //   set beyond last used    
	ENDTOT= FIRST+NALLOC;                      //   set beyond last alloc   
      }                                            // <                         
      BOX* new_box()                               // give another box out      
	// no check for overflow; has to be done by caller: manager::new_box()  
      {                                            // >                         
	++NUSED;                                   //   keep track of # used    
	return END++;                              //   return end ++           
      }                                            // <                         
      BOX* new_boxes(const uint& New)              // give another New boxes out
	// no check for overflow; has to be done by caller: manager::new_boxes()
      {                                            // >                         
	NUSED+=New;                                //   keep track of # used    
	register BOX* CR=END;                      //   reserve boxes           
	END+=New;                                  //   increment next free     
	return CR;                                 //   return old end          
      }                                            // <                         
      bool has_free(const uint& n) const           // Do we have n free spaces ?
      {                                            // >                         
	return NUSED + n <= NALLOC;                //   true if NUSED+n<= NALLOC
      }                                            // >                         
      bool is_full() const { return END>=ENDTOT; } // is block full?            
      BOX* front  () const { return FIRST; }       // first box used            
      BOX* back   () const { return END-1; }       // last box used             
      BOX* end    () const { return END; }         // beyond last box used      
      const uint& N_used () const { return NUSED; }// # boxes given out sofar   
      const uint& N_alloc() const { return NALLOC;}// # boxes allocated         
    };
  public:
    ////////////////////////////////////////////////////////////////////////////
    // 1.2 class fasttree::manager::iterator                                    
    //     for forward sequential iteration through all boxes                   
    ////////////////////////////////////////////////////////////////////////////
    class iterator{
    private:
      block *B;                                    // current block             
      BOX   *C;                                    // current box               
    public:
      iterator(block*b, BOX*c) : B(b), C(c) {}
      iterator(const iterator&I) : B(I.B), C(I.C) {}
      iterator& operator= (const iterator&I) { B=I.B; C=I.C; return *this; }
      iterator& operator++()                       // prefix ++                 
      {                                            // >                         
	if(C < B->back() )  ++C;                   //   IF still in block, ++C  
	else              { ++B; C=B->front(); }   //   ELSE: 1st of next block 
	return *this;                              //   return *this            
      }                                            // <                         
      iterator& operator--()                       // prefix --                 
      {                                            // >                         
	if(C > B->front() )  --C;                  //   IF still in block, --C  
	else               { --B; C=B->back(); }   //   ELSE: last of prev block
	return *this;                              //   return *this            
      }                                            // <                         
      iterator  operator++(int)                    // postfix ++                
      {                                            // >                         
	iterator tmp(*this);                       //   temp copy *this         
	this->operator++();                        //   prefix ++ *this         
	return tmp;                                //   return temp copy        
      }                                            // <                         
      iterator  operator--(int)                    // postfix --                
      {                                            // >                         
	iterator tmp(*this);                       //   temp copy *this         
	this->operator--();                        //   prefix -- *this         
	return tmp;                                //   return temp copy        
      }                                            // <                         
      bool operator== (const iterator&I) { return B == I.B  &&  C == I.C; }
      bool operator!= (const iterator&I) { return B != I.B  ||  C != I.C; }
      operator   BOX*()                 { return C; }
      BOX& operator* ()                 { return*C; }
      BOX* operator->()                 { return C; }
    };
    ////////////////////////////////////////////////////////////////////////////
    // 1.3 class fasttree::manager::reverse_iterator                            
    //     for backward sequential iteration through all boxes                  
    ////////////////////////////////////////////////////////////////////////////
    class reverse_iterator{
    private:
      block *B;
      BOX  *C;
    public:
      reverse_iterator(block*b, BOX*c) : B(b), C(c) {}
      reverse_iterator(const reverse_iterator&I) : B(I.B), C(I.C) {}
      reverse_iterator& operator= (const reverse_iterator&I)
      {
	B=I.B;
	C=I.C;
	return *this;
      }
      reverse_iterator& operator++()               // prefix ++                 
      {                                            // >                         
	if(C > B->front() )  --C;                  //   IF still in block, --C  
	else               { --B; C=B->back(); }   //   ELSE: last of prev block
	return *this;                              //   return *this            
      }                                            // <                         
      reverse_iterator& operator--()               // prefix --                 
      {                                            // >                         
	if(C < B->back() )  ++C;                   //   IF still in block, ++C  
	else              { ++B; C=B->front(); }   //   ELSE: 1st of next block 
	return *this;                              //   return *this            
      }                                            // <                         
      reverse_iterator  operator++(int)            // postfix ++                
      {                                            // >                         
	reverse_iterator tmp(*this);               //   temp copy *this         
	this->operator++();                        //   prefix ++ *this         
	return tmp;                                //   return temp copy        
      }                                            // <                         
      reverse_iterator  operator--(int)            // postfix --                
      {                                            // >                         
	reverse_iterator tmp(*this);               //   temp copy *this         
	this->operator--();                        //   prefix -- *this         
	return tmp;                                //   return temp copy        
      }                                            // <                         
      bool operator== (const reverse_iterator&I) { return B == I.B  &&  C == I.C; }
      bool operator!= (const reverse_iterator&I) { return B != I.B  ||  C != I.C; }
      operator   BOX*()                 { return C; }
      BOX& operator* ()                 { return*C; }
      BOX* operator->()                 { return C; }
    };
    ////////////////////////////////////////////////////////////////////////////
    // 2 member data of class fasttree::manager                                 
    //   hold & manager the blocks, keep track of boxes used and allocated      
    ////////////////////////////////////////////////////////////////////////////
  private:
    uint         NBOD;                             // # bodies in total         
    uint         NTOT;                             // # boxes allocated         
    uint         NUSED;                            // # boxes used              
    block        BA[N_BLOCKS+1];                   // we have N_BLOCKS blocks   
    block       *BZ;                               // before 1st active block   
    block       *B0;                               // first active block        
    block       *BI;                               // currently active block    
    block       *BN;                               // beyond last block         
    ////////////////////////////////////////////////////////////////////////////
    // 3 member functions of class fasttree::manager                            
    ////////////////////////////////////////////////////////////////////////////
  public:
    manager(                                       // constructor:              
	    const uint&n) :                        // I: # bodies in total      
      NBOD(n),                                     //   set # bodies in total   
      NUSED(0),                                    //   set # boxes given out   
      BZ(BA),                                      //   set pter to 0th block   
      B0(BZ+1),                                    //   set pter to 1st block   
      BI(B0),                                      //   set pter to active block
      BN(BZ+N_BLOCKS)                              //   set pter to end block   
    {                                              // >                         
      NTOT = 1+NBOD/4;                             //   set # boxes allocated   
      BI->allocate(NTOT);                          //   N/4 boxes in 1st block  
    }                                              // <                         
    void reset(                                    // destruct & construct      
	       const uint&n)                       // I: # bodies in total      
    {                                              // >                         
      for(block*B=B0; B<BN; B++) B->reset();       //   reset all blocks        
      NBOD  = n;                                   //   set # bodies in total   
      NUSED = 0u;                                  //   set # boxes given out   
      B0    = BZ+1;                                //   set pter to 1st block   
      BI    = B0;                                  //   set pter to active block
      NTOT  = NBOD/4;                              //   set # boxes allocated   
      BI->allocate(NTOT);                          //   N/4 boxes in 1st block  
    }                                              // <                         
    BOX* new_box(                                  // make new box              
		 const uint&Nsofar)                // I: # bodies added sofar   
    {                                              // >                         
      if(BI->is_full()) {                          //   IF active block full >  
	BI++;                                      //     take next block       
	if(BI >= BN)                               //     IF beyond last block  
	  error("too few boxes allocated a");       //       FATAL ERROR        
	if(Nsofar > NBOD)                          //     IF N_sofar > NBOD     
	  error("more bodies added than exist");   //       FATAL ERROR         
	register real x   = NUSED*(real(NBOD)/real(Nsofar)-one);
	register uint New = int(x+4*SQRT(x)+16);   //     # boxes to allocate   
	BI->allocate(New);                         //     allocate in next block
	NTOT += New;                               //     # boxes allocated     
      }                                            //   <                       
      NUSED++;                                     //   increment # boxes used  
      return BI->new_box();                        //   return new box          
    }                                              // <                         
    BOX* new_boxes(                                // make new boxes            
		  const uint&Nboxes,               // I: # boxes wanted         
		  const uint&Nsofar)               // I: # bodies added sofar   
    {                                              // >                         
      if(!BI->has_free(Nboxes)) {                  //   IF active black full >  
	BI++;                                      //     take next block       
	if(BI >= BN)                               //     IF beyond last block  
	  error("too few boxes allocated a");      //       FATAL ERROR         
	if(Nsofar > NBOD)                          //     IF N_sofar > NBOD     
	  error("more bodies added than exist");   //       FATAL ERROR         
	register real x   = NUSED*(real(NBOD)/real(Nsofar)-one);
	register uint New = int(x+5*SQRT(x)+20);   //     # boxes to allocate   
	New = max(New,Nboxes);                     //     guarantee Nboxes      
	BI->allocate(New);                         //     allocate in next block
	NTOT += New;                               //     # boxes allocated     
      }                                            //   <                       
      NUSED+=Nboxes;                               //   increment # boxes used  
      return BI->new_boxes(Nboxes);                //   return new box          
    }                                              // <                         
    int number_of_box(const box* C) const          // number of given box*      
    {                                              // >                         
      if(C==0) return -99;                         //   IF null: return -99     
      register int num=0;                          //   reset counter num       
      for(register block*cb=B0; cb<=BI; cb++)      //   loop blocks         >   
	if(C >= cb->front() && C < cb->end())      //     IF in block         > 
	  return num + int(C-cb->front());         //       return num + n(in b)
	else                                       //     < ELSE >              
	  num += cb->N_used();                     //   <   add to counter    < 
      return -99;                                  //   not found: return -99   
    }                                              // <                         
    box*             root  () const { return B0->front(); }
    iterator         front () const { return iterator(B0,B0->front()); }
    iterator         back  () const { return iterator(BI,BI->back()); }
    iterator         begin () const { return iterator(B0,B0->front()); }
    iterator         end   () const { return iterator(BI+1,(BI+1)->front()); }
    reverse_iterator rbegin() const { return reverse_iterator(BI,BI->back()); }
    reverse_iterator rend  () const { return reverse_iterator(BZ,BZ->back()); }
    const uint& N_bodies   () const { return NBOD; }
    const uint& N_used     () const { return NUSED; }
    const uint& N_allocated() const { return NTOT; }
    const uint  N_blocks   () const { return 1+int(BI-B0); }
  };
#undef N_BLOCKS
}
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// class box {                                                                  
// public:                                                                      
//   void some_member_function();                                               
// void some_function(box*);                                                    
// manager<box>*CM;                                                             
//                                                                              
// In order to loop the boxes downwards (root being the first), use             
//                                                                              
//   for(boxmanager::iterator Ci=CM->begin(); Ci!=CM->end(); ++Ci) {            
//     some_function(Ci);                                                       
//     Ci->some_member_of_box();                                                
//   }                                                                          
//                                                                              
// In order to loop the boxes upwards (root being the last), use                
//                                                                              
//   for(boxmanager::reverse_iterator Ci=CM->rbegin(); Ci!=CM->rend(); ++Ci) {  
//     some_function(Ci);                                                       
//     Ci->some_member_of_box();                                                
//   }                                                                          
//                                                                              
////////////////////////////////////////////////////////////////////////////////
#endif // included_bman_h
