// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// cell.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class cell                         fully inline                             |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_cell_h
#define included_cell_h 1

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_flag_h
#include <tree/flag.h>
#endif

#ifndef included_node_h
#include <tree/node.h>
#endif

#ifndef included_soul_h
#include <tree/soul.h>
#endif

#ifndef included_matr_h
#include <tree/matr.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// some macros                                                                  
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// macro NGRAVP: number of reals needed in grav-part of cell = 6 + size of matr 
//                                                                              
#if NDIM==3
#define NGRAVP 12
#else
#define NGRAVP  9
#endif
//                                                                              
// macro steering the way the sub-cells are organized                           
//                                                                              
#undef SUBCELL_LLIST
#undef SUBCELL_ARRAY
// #define SUBCELL_LLIST        // as linked list                               
#define SUBCELL_ARRAY           // as array                                     

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  class leaf;                                      // forward declaration       
  class TaylorCoeffs;                              // forward declaration       
  template<class X> class manager;                 // forward declaration       
  //////////////////////////////////////////////////////////////////////////////
  // fasttree::class cell                                                       
  //////////////////////////////////////////////////////////////////////////////
  class cell : 
    private node,
    public  flag
  {
    // class cell is a element in the linked structure of a tree                
    friend class box;
    friend class builder;
#ifndef C_AND_F
    friend void dump_cell(const cell*, std::ostream&, const cell*, const soul*);
#endif
  private:
    cell(const cell&);                             // not implemented           
    cell& operator=(const cell&);                  // not implemented           
                                                   // general properties        
    vect             COFM;                         //    center of mass         
    real             RADIUS;                       //    half size of the cube  
    real             RMAX;                         //    max distance from CofM 
    real             MASS;                         //    mass in cell           
    uint             NUMBER;                       //    # souls (total)        
    indx             DEPTH;                        //    tree depth downward    
#ifndef SUBSOUL_LLIST
    indx             NSOULS;                       //    # sub-souls            
#endif
#ifdef SUBCELL_LLIST
    cell            *NEXT;                         //    pter to sister cell    
#else
    indx             NCELLS;                       //    # sub-cells            
#endif
    soul            *SOULS;                        //    pointer to 1st soul    
    cell            *CELLS;                        //    pointer to 1st sub-cell
    mutable real     GRAVP[NGRAVP];                //    holds grav-part of cell
    TaylorCoeffs    *COEFFS;                       //    pointer to coefficients
  public:
    //                                                                          
    // constructor                                                              
    //                                                                          
    cell() {}
    //                                                                          
    // simple manipulations                                                     
    //                                                                          
    void set_rcrit(const real ith)    { rcrit()  = ith * rmax();
                                        rcrit2() = rcrit() * rcrit(); }
    //                                                                          
    // non-const data access via member functions                               
    //                                                                          
  protected:
    vect&         center()            { return pos(); }
    real&         radius()            { return RADIUS; }
  public:
    real&         rmax  ()            { return RMAX; }
    real&         mass  ()            { return MASS; }
    vect&         cofm  ()            { return COFM; }
    real&         imass ()            { return GRAVP[0]; }
    real&         rcrit ()            { return GRAVP[1]; }
    real&         rcrit2()            { return GRAVP[2]; }
    real&         eps   ()            { return GRAVP[3]; }
    real&         emin  ()            { return GRAVP[4]; }
    real&         emax  ()            { return GRAVP[5]; }
    emat          quadru()            { return emat(GRAVP+6); }
    TaylorCoeffs*&coeffs()            { return COEFFS; }
    //                                                                          
    // access to sub-cells via subcell_iterator                                 
    //                                                                          
#ifdef SUBCELL_LLIST
    cell*     &next()       { return NEXT; }
    cell*const&next() const { return NEXT; }
#endif
    class subcell_iterator {
      cell* X;
    public:
      subcell_iterator(cell*x) : X(x) {}
      subcell_iterator(const subcell_iterator&I) : X(I.X) {}
      subcell_iterator&operator=(const subcell_iterator&I) { X=I.X; return *this; }
#ifdef SUBCELL_LLIST
      subcell_iterator&operator++() { X=X->next(); return *this; }
      subcell_iterator next      () const { return subcell_iterator(X->next()); }
#else
      subcell_iterator&operator++() { ++X; return *this; }
      subcell_iterator next      () const { return subcell_iterator(X+1); }
#endif
      subcell_iterator operator++(int) { subcell_iterator tmp(*this);
                                         operator++(); return tmp; }
      bool operator==(const subcell_iterator&I) const { return X == I.X; }
      bool operator!=(const subcell_iterator&I) const { return X != I.X; }
      operator cell*  ()                        { return X; }
      cell& operator* ()                        { return*X; }
      cell* operator->()                        { return X; }
    };
    subcell_iterator begin_subcells() const{return subcell_iterator(CELLS); }
#ifdef SUBCELL_LLIST
    subcell_iterator end_subcells() const {return subcell_iterator(0); }
#else
    subcell_iterator end_subcells() const {return subcell_iterator(CELLS+NCELLS);}
#endif
    //                                                                          
    // access to sub-souls via subsoul_iterator                                 
    //                                                                          
    class subsoul_iterator {
      soul* X;
    public:
      subsoul_iterator(soul*x) : X(x) {}
      subsoul_iterator(const subsoul_iterator&I) : X(I.X) {}
      subsoul_iterator&operator=(const subsoul_iterator&I) { X=I.X; return *this; }
#ifdef SUBSOUL_LLIST
      subsoul_iterator&operator++() { X=X->next(); return *this; }
      subsoul_iterator next      () const { return subsoul_iterator(X->next()); }
#else
      subsoul_iterator&operator++() { ++X; return *this; }
      subsoul_iterator next      () const { return subsoul_iterator(X+1); }
#endif
      subsoul_iterator operator++(int) { subsoul_iterator tmp(*this);
                                         operator++(); return tmp; }
      bool operator==(const subsoul_iterator&I) const { return X == I.X; }
      bool operator!=(const subsoul_iterator&I) const { return X != I.X; }
      operator soul*  ()                        { return X; }
      soul& operator* ()                        { return*X; }
      soul* operator->()                        { return X; }
    };
    subsoul_iterator begin_subsouls() const{return subsoul_iterator(SOULS); }
#ifdef SUBSOUL_LLIST
    subsoul_iterator end_subsouls() const {return subsoul_iterator(0); }
#else
    subsoul_iterator end_subsouls() const {return subsoul_iterator(SOULS+NSOULS);}
#endif
    //                                                                          
    // access to all souls via allsoul_iterator                                 
    //                                                                          
    class allsoul_iterator {
      soul* X;
    public:
      allsoul_iterator(soul*x) : X(x) {}
      allsoul_iterator(const allsoul_iterator&I) : X(I.X) {}
      allsoul_iterator&operator=(const allsoul_iterator&I) { X=I.X; return *this; }
      allsoul_iterator&operator++()       { ++X; return *this; }
      allsoul_iterator operator++(int)    { allsoul_iterator tmp(*this);
                                            operator++(); return tmp; }
      allsoul_iterator&operator+=(const int i) { X+=i; return *this; }
      allsoul_iterator&operator-=(const int i) { X-=i; return *this; }
      allsoul_iterator operator+ (const int i) { return allsoul_iterator(X+i); }
      allsoul_iterator operator- (const int i) { return allsoul_iterator(X-i); }
      bool operator==(const allsoul_iterator&I) const { return X == I.X; }
      bool operator!=(const allsoul_iterator&I) const { return X != I.X; }
      bool operator==(const soul*S)             const { return X == S; }
      bool operator!=(const soul*S)             const { return X != S; }
      operator soul*  ()                        { return X; }
      soul& operator* ()                        { return*X; }
      soul* operator->()                        { return X; }
      friend bool operator==(const soul*S,const allsoul_iterator&I) {return I==S;}
      friend bool operator!=(const soul*S,const allsoul_iterator&I) {return I!=S;}
    };
    allsoul_iterator begin_allsouls() const{return allsoul_iterator(SOULS); }
    allsoul_iterator last_allsouls()const{return allsoul_iterator(SOULS+NUMBER-1);}
    allsoul_iterator end_allsouls()const{return allsoul_iterator(SOULS+NUMBER);}
    //                                                                          
    // const data access via friends                                            
    //                                                                          
#define CCCC const cell*const&C
    friend const real&radius    (CCCC) {return C->RADIUS; }
    friend const real&rmax      (CCCC) {return C->RMAX; }
    friend const real&mass      (CCCC) {return C->MASS; }
    friend const vect&cofm      (CCCC) {return C->COFM; }
    friend const vect&center    (CCCC) {return C->pos(); }
    friend const uint&number    (CCCC) {return C->NUMBER; }
    friend const indx&depth     (CCCC) {return C->DEPTH; }
    friend const real&imass     (CCCC) {return C->GRAVP[0]; }
    friend const real&rcrit     (CCCC) {return C->GRAVP[1]; }
    friend const real&rcrit2    (CCCC) {return C->GRAVP[2]; }
    friend const real&eps       (CCCC) {return C->GRAVP[3]; }
    friend const real&emin      (CCCC) {return C->GRAVP[4]; }
    friend const real&emax      (CCCC) {return C->GRAVP[5]; }
    friend const emat quadru    (CCCC) {return emat(C->GRAVP+6); }
    friend const TaylorCoeffs*const&coeffs (CCCC) {return C->COEFFS; }
    //                                                                          
    // boolean information via friends                                          
    //                                                                          
    friend bool       has_subsouls    (CCCC) {return C->NSOULS!=0; }
#ifdef SUBCELL_LLIST
    friend bool       has_subcells    (CCCC) {return C->CELLS !=0; }
#else
    friend bool       has_subcells    (CCCC) {return C->NCELLS!=0; }
#endif
    friend bool       is_twig         (CCCC) {return !has_subcells(C); }
    friend bool       is_sink         (CCCC) {return C->is_set(SINK); }
    friend bool       al_sink         (CCCC) {return C->is_set(AL_SINK); }
    friend bool       in_subtree      (CCCC) {return C->is_set(SUBTREE); }
    friend bool       is_subtreecell  (CCCC) {return C->is_set(SUBTREECELL); }
    friend bool       is_subtreezombie(CCCC) {return C->is_set(SUBTREEZOMBIE);}
#ifdef TESTBODIES
    friend bool       is_source       (CCCC) {return C->MASS!=zero; }
    friend bool       al_source       (CCCC) {return C->is_set(AL_SOURCE); }
#endif
#undef CCCC
    //                                                                          
    // flag manipulation (non-const)                                            
    //                                                                          
    void flag_for_subtree()            { flag::addflag(SUBTREE); }
    void flag_as_subtreecell()         { flag::addflag(SUBTREECELL); }
    void flag_as_subtreezombie()       { flag::addflag(SUBTREEZOMBIE); }
    void unflag_subtree()              { flag::unflag(SUBTREE); }
    void unflag_subtreecell()          { flag::unflag(SUBTREECELL); }
    void unflag_subtreezombie()        { flag::unflag(SUBTREEZOMBIE); }
    void unflag_subtree_flags()        { flag::unflag(SUBTREE_FLAGS); }
    void reset_sink_flag()             { flag::unflag(SINK);
                                         flag::addflag(AL_SINK);}
    void add_sink_flag  (const soul*s) { flag::addpartflag(s,SINK);
                                         if(!is_sink(s)) flag::unflag(AL_SINK); }
    void add_sink_flag  (const cell*c) { flag::addpartflag(c,SINK);
                                         if(!al_sink(c)) flag::unflag(AL_SINK); }
#ifdef TESTBODIES
    void reset_source_flag()           { flag::addflag(AL_SOURCE); }
    void add_source_flag(const soul*s) {if(!is_source(s))
                                          flag::unflag(AL_SOURCE);}
    void add_source_flag(const cell*c) {if(!al_source(c))
                                          flag::unflag(AL_SOURCE);}
#endif
  };
}
#undef NGRAVP
////////////////////////////////////////////////////////////////////////////////
#endif // included_cell_h
