// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// flag.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class flag                           fully inline                           |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_flag_h
#define included_flag_h 1

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// class flag                                                                   
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  class flag {
    // flag to be used in the tree, indicating whether a node is sink
    // of gravity/density or not used in the tree (first 3 bits). The
    // higher bits are free for other usages.
    //
    // base class of classes body & node.
    // 
    // bit value    meaning
    //   1    1     tree node is sink
    //   2    2     body shall be ignored by class tree
    //   3    4     body is a SPH particle
    //   4    8     body is a sticky particle
    // i>4  2^(i+1) free for usage by derived classes
  public:
    enum {
      SINK          =   1,  // tree node is sink
      NOT_IN_TREE   =   2,  // body shall be ignored by class tree
      SPH           =   4,  // body is a SPH particle
      STICKY        =   8,  // body is a sticky particle
      BODY_FLAGS    =  15,  // sum of all body flags
#ifdef TESTBODIES
      AL_SOURCE     =  32,  // all leafs of cell are sources (no test-bodies)
#endif
      AL_SINK       =  64,  // all leafs of cell are sink
      SUBTREE       = 128,  // node is to become node of subtree
      SUBTREECELL   = 256,  // cell is to become cell of subtree
      SUBTREEZOMBIE = 512,  // cell with only one subtree subcell
      SUBTREE_FLAGS = 896   // sum of all subtree flags
    };
  private:
    indx FLAG;
  protected:
    void reset      ()                           { FLAG  =  0; }
    void setflag    (const indx i)               { FLAG  =  i; }
    void setpartflag(const int  f, const indx i) { FLAG  =  f&i; }
    void setpartflag(const flag*f, const indx i) { FLAG  =  f->FLAG&i; }
    void addflag    (const indx i)               { FLAG |=  i; }
    void addpartflag(const int  f, const indx i) { FLAG |=  f&i; }
    void addpartflag(const flag*f, const indx i) { FLAG |=  f->FLAG&i; }
    void unflag     (const indx i)               { FLAG &= ~i; }
    void set        (const indx i)               { FLAG=i; }
    void set        (const flag f)               { FLAG=f.FLAG; }
    void set        (const flag*f)               { FLAG=f->FLAG; }
    bool is_set     (const indx i) const         { return FLAG & i; }
    flag() : FLAG(0)                 {}    // default: not sink but in tree
    flag (const indx f) : FLAG(f)    {}
  public:
    //   operator indx&()                 { return FLAG; }
    operator const indx&()    const  { return FLAG; }
    friend bool is_in_tree(const int  F) { return !(F & flag::NOT_IN_TREE); }
    friend bool is_sph    (const int  F) { return   F & flag::SPH; }
    friend bool is_sticky (const int  F) { return   F & flag::STICKY; }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_flag_h
