// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// iact.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class A, class B> struct iaction                                   |
// template<class A, class B> class iastack                                    |
// template<class INTERACTOR, class CELL, class SOUL> void interact()          |
// template<class INTERACTOR, class CELL, class SOUL> class interaction        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_iact_h
#define included_iact_h

namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::iaction<A,B>                                               
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  struct iaction
  // holds a pair of objects, very simple                                       
  {
    A     first;                                       // first object          
    B     second;                                      // second object         
    iaction (A a=0, B b=0) :                           // constructor           
      first (a),                                       //   set first           
      second(b) {}                                     //   set second          
    void set(A a, B b)                                 // set pair              
    {                                                  // >                     
      first =a;                                        //   set first           
      second=b;                                        //   set second          
    }                                                  // <                     
    iaction& operator= (const iaction&I)               // assignment            
    {                                                  // >                     
      first =I.first;                                  //   set first           
      second=I.second;                                 //   set second          
      return *this;                                    //   return              
    }                                                  // <                     
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::iastack<A,B>                                               
  //////////////////////////////////////////////////////////////////////////////
  template<class A, class B>
  class iastack
  // manages a stack of iaction<>s, very simple                                 
  {
  private:
    typedef iaction<A,B>    iact;                      // type of stack objects 
    iact    *IA, *pi;                                  // first & active element
  public:
    iastack (const unsigned int M) :                   // constructor           
      IA(new iact[M]),                                 //   allocate memory     
      pi(IA-1) {}                                      //   set pter to active  
    ~iastack () { delete[] IA; }                       // destructor: deallocate
    bool     is_empty  () const   { return pi<IA;}     // if stack empty?       
    iact     pop       ()         { return *(pi--); }  // give last active: pop 
    void     push      (A a, B b) { (++pi)->set(a,b); }// add element:      push
  };
  //////////////////////////////////////////////////////////////////////////////
  // functions fasttree::interact<I,A,B>()                                      
  //////////////////////////////////////////////////////////////////////////////
#define LoopSubCells(C,A)    for(register typename CELL::subcell_iterator	\
			     A=C->begin_subcells(); A!=C->end_subcells(); ++A)
#define LoopSubPairs(C,A,B)  for(register typename CELL::subcell_iterator	\
                             B=A;                   B!=C->end_subcells(); ++B)
#define LoopSubSouls(C,A)    for(register typename CELL::subsoul_iterator	\
			     A=C->begin_subsouls(); A!=C->end_subsouls(); ++A)
#define LoopSoulPairs(C,A,B) for(register typename CELL::subsoul_iterator	\
			     A=C->begin_subsouls(); A!=C->end_subsouls(); ++A)	\
                               for(register typename CELL::subsoul_iterator	\
			       B=A.next();          B!=C->end_subsouls(); ++B)
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL>
  inline void interact(      CELL *I1,                 // I:   1st iacting cell 
			     CELL *I2,                 // I:   2nd iacting cell 
		       const INTERACTOR *K)            // I:   interactor       
  {
    const unsigned int  NN = (I1==I2)?                 // self-interaction?     
                        NSUB*(NSUB+1)*depth(I1)/2      // max size of stacks    
                      : NSUB*(depth(I1)+depth(I2));    // max size of stacks    
    register iastack<CELL*,CELL*> CCIAS(NN);           // stack for C-C iacts   
    register iastack<CELL*,SOUL*> CBIAS(NN);           // stack for C-B iacts   
    register iaction<CELL*,CELL*> cc;                  // current C-C iact      
    register iaction<CELL*,SOUL*> cb;                  // current C-B iact      
    CCIAS.push(I1,I2);                                 // push I1-I2 onto CC    
    while(!CCIAS.is_empty()) {                         // while CC non-empty  > 
      cc = CCIAS.pop();                                //   pop new C-C iact    
      if(cc.first == cc.second) {                      //   IF self iact      > 
	if(K->interact(cc.first)) continue;            //     success? -> next  
	LoopSubCells(cc.first,c1) {                    //     ELSE: split     > 
	  LoopSubSouls(cc.first,s2)    CBIAS.push(c1,s2); //    push sub C-B    
	  LoopSubPairs(cc.first,c1,c2) CCIAS.push(c1,c2); //    push sub C-C    
	}                                              //     <                 
	LoopSoulPairs(cc.first,s1,s2) K->interact(s1,s2); //  do sub B-B        
      } else {                                         //   < ELSE: mutual    > 
	if(K->interact(cc.first,cc.second)) continue;  //     success? -> next  
	if(K->split_first(cc.first,cc.second)) {       //     IF split 1st    > 
	  LoopSubSouls(cc.first,s2) CBIAS.push(cc.second,s2); // push sub C-B   
	  LoopSubCells(cc.first,c2) CCIAS.push(c2,cc.second); // push sub C-C   
	} else {                                       //     < ELSE split 2nd> 
	  LoopSubSouls(cc.second,s2) CBIAS.push(cc.first,s2); // push sub C-B   
	  LoopSubCells(cc.second,c2) CCIAS.push(cc.first,c2); // push sub C-C   
	}                                              //     <                 
      }                                                //   <                   
      while(!CBIAS.is_empty()) {                       //   while CB non-empty >
	cb = CBIAS.pop();                              //     pop new C-B iact  
	if(K->interact(cb.first,cb.second)) continue;  //     success? -> next  
	LoopSubCells(cb.first,c1) CBIAS.push(c1,cb.second); //push sub C-B      
	LoopSubSouls(cb.first,s2) K->interact(cb.second,s2);//do sub B-B        
      }                                                //   <                   
    }                                                  // <                     
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL>
  inline void interact(      CELL *I1,               // I:   self-iacting cell
		       const INTERACTOR *K)          // I:   interactor
  {
    interact<INTERACTOR,CELL,SOUL>(I1,I1,K);
  }
#if 0 // not (yet?) implemented
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::interation<I,A,B>                                          
  //////////////////////////////////////////////////////////////////////////////
  template<class INTERACTOR, class CELL, class SOUL>
  class interaction
  {
  public:
#ifdef USE_BODIES
    typedef const fasttree::body* body_pair[2];
#else  // USE_BODIES
    typedef unsigned int          body_pair[2];
#endif // USE_BODIES
  private:
    unsigned int         NN;                         // maximal stack size
    iastack<CELL*,CELL*> CCIAS;                      // cell-cell iactions
    iastack<CELL*,SOUL*> CBIAS;                      // cell-body iactions
    INTERACTOR          *K;                          // mutual interactor
    body_pair           *BLIST;                      // interaction list
    const unsigned int   NL;                         // physical size of list
    const unsigned int   I;                          // actual element on list

    static unsigned int stack_size(CELL*, CELL*);
    void                copy_body_pair();
    void                empty_buffer();
    void                empty_CB();
    void                empty_CC();
    interaction();
  public:
    interaction(CELL*, CELL*, INTERACTOR*, body_pair*, const unsigned int);
    void fill_list ();
    void reset_list() { I=0; }
    bool list_full () { return I>=NL; }
    bool finished  () { return CCIAS.is_empty() && CBIAS.is_empty(); }
  };
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  unsigned int interaction<INTERACTOR,CELL,SOUL>::stack_size(CELL* I1, CELL *I2)
  {
    return (I1==I2)?  NSUB*(NSUB+1)*depth(I1)/2    // self-interaction
                    : NSUB*(depth(I1)+depth(I2));  // mutual interaction
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  interaction<INTERACTOR,CELL,SOUL>::interaction(CELL* i1, CELL *i2,
						 INTERACTOR *k,
						 body_pair *bp,
						 const unsigned int nl)
    : NN(stack_size(i1,i2)), CCIAS(NN), CBIAS(NN), K(k), BLIST(bp), NL(nl), I(0)
  { CCIAS.push(I1,I2); }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  void interaction<INTERACTOR,CELL,SOUL>::copy_body_pair()
  {
    register body_pair BP = K->get_pair();
    BLIST[I][0] = BP[0];
    BLIST[I][1] = BP[1];
    I++;
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  void interaction<INTERACTOR,CELL,SOUL>::empty_buffer()
  {
    while(!K->list_empty() && !list_full() ) copy_body_pair();
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  void interaction<INTERACTOR,CELL,SOUL>::empty_CB()
  {
    register iaction<CELL*,SOUL*> cb;
    empty_buffer();
    while(!CBIAS.is_empty() && !list_full() ) {
      cb = CBIAS.pop();
      if(K->interact(cb.first,cb.second))
	empty_buffer();
      else {
	LoopSubCells(cb.first,c1) CBIAS.push (c1,cb.second);
	LoopSubSouls(cb.first,s2) K->interact(cb.second,s2);
	empty_buffer();
      }
    }
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  void interaction<INTERACTOR,CELL,SOUL>::empty_CC()
  {
    register iaction<CELL*,CELL*> cc;
    empty_CB();
    while(!CCIAS.is_empty() && !list_full() ) {
      cc = CCIAS.pop();
      if(cc.first == cc.second) {
	if(K->interact(cc.first)) 
	  empty_buffer();
	else {
	  LoopSubCells(cc.first,c1) {
	    LoopSubSouls(cc.first,s2)    CBIAS.push(c1,s2);
	    LoopSubPairs(cc.first,c1,c2) CCIAS.push(c1,c2);
	  }
	  LoopSoulPairs(cc.first,s1,s2) K->interact(s1,s2);
	  empty_buffer();
	}
      } else {
	if(K->interact(cc.first,cc.second)) 
	  empty_buffer();
	else if(K->split_first(cc.first,cc.second)) {
	  LoopSubSouls(cc.first,s2) CBIAS.push(cc.second,s2);
	  LoopSubCells(cc.first,c2) CCIAS.push(c2,cc.second);
	} else {
	  LoopSubSouls(cc.second,s2) CBIAS.push(cc.first,s2);
	  LoopSubCells(cc.second,c2) CCIAS.push(cc.first,c2);
	}
      }
      empty_CB();
    }
  }
  //----------------------------------------------------------------------------
  template<class INTERACTOR, class CELL, class SOUL> inline
  void interaction<INTERACTOR,CELL,SOUL>::fill_list()
  {
    if(list_full()) fasttree::error("interaction() called when list was full");
    empty_CC();
  }
#endif // 0
}
//==============================================================================
#undef LoopSubSouls
#undef LoopSubCells
#undef LoopSubPairs
#undef LoopSoulPairs
////////////////////////////////////////////////////////////////////////////////
#endif // included_iact_h
