// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// nmio.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_nmio_h
#define included_nmio_h

#ifdef  ALLOW_NEMO

//------------------------------------------------------------------------------

#ifdef TWODIMENSIONAL
#define NDM 2
#define TDM 4
#else
#define NDM 3
#define TDM 6
#endif

//------------------------------------------------------------------------------
namespace fasttree {
  class nemo_io {
  public:
    enum Scalar {time,cputime};
    enum Vector {energy};
    enum Matrix {KinT,PotT,AmT};
    enum Array  {mass,acc,pot,aux};
    enum Set    {snap,param,bodies,diags};
    enum CoSys  {cart,spher,scat};
  private:
    nemo_io();
    nemo_io(const nemo_io&);
  protected:
    void  *STREAM;
    mutable int    N,CS;
    mutable float *AUXX, *PHASE, MATRIX[NDM*NDM], CMPHASE[TDM], VECTOR[NDM];

    nemo_io(char*, char*);
   ~nemo_io();
    void  open (char*, char*);
    void  close();
    bool  is_open() { return STREAM != 0; }
    void  open_set     (const Set, const bool) const;
    void  close_set    (const Set, const bool) const;
    bool  is_present   (const Scalar) const;
    bool  is_present   (const Vector) const;
    bool  is_present   (const Matrix) const;
    bool  is_present   (const Array)  const;
    bool  is_present   (const Set)    const;

    int   read_N       () const;
    float read_scalar  (const Scalar) const;
    void  read_vector  (const Vector) const;
    void  read_cofm    () const;;
    void  read_matrix  (const Matrix) const;
    void  read_array   (const Array) const;
    void  read_posvel  () const;
    void  read_history () const;

    void  write_N      (const int);
    void  write_coords (const CoSys);
    void  write_scalar (const Scalar, const float);
    void  write_vector (const Vector);
    void  write_cofm   ();
    void  write_matrix (const Matrix);
    void  write_array  (const Array);
    void  write_posvel ();
    void  write_history() const;

  public:
    void  reset_N     (const int) const;
  };
  //----------------------------------------------------------------------------
  class nemo_input : public nemo_io {
  public:
    nemo_input(char* file) : nemo_io(file,"r") {}

    const float& aux(const int i) const              { return AUXX[i]; }
    const float* vec()            const              { return VECTOR; }
    const float& vec(const int j) const              { return VECTOR[j]; }
    const float* cmx()            const              { return CMPHASE; }
    const float* cmv()            const              { return CMPHASE+NDM; }
    const float& cmx(const int j) const              { return CMPHASE[j]; }
    const float& cmv(const int j) const              { return CMPHASE[NDM+j]; }
    const float& mat(const int i, const int j) const { return MATRIX[i*NDM+j]; }
    const float* pos(const int i) const              { return PHASE+i*TDM; }
    const float& pos(const int i, const int j) const { return PHASE[i*TDM+j]; }
    const float* vel(const int i) const              { return PHASE+i*TDM+NDM; }
    const float& vel(const int i, const int j) const { return PHASE[i*TDM+NDM+j];}

    nemo_io::is_present;
    void  open_set    (const Set S)   const { return nemo_io::open_set(S,true); }
    void  close_set   (const Set S)   const { return nemo_io::close_set(S,true); }
    int   read_N      ()              const { return nemo_io::read_N(); }
    float read_scalar (const Scalar S)const { return nemo_io::read_scalar(S); }
    void  read_vector (const Vector V)const { return nemo_io::read_vector(V); }
    void  read_cofm   ()              const { return nemo_io::read_cofm(); }
    void  read_matrix (const Matrix M)const { return nemo_io::read_matrix(M); }
    void  read_array  (const Array A) const { return nemo_io::read_array(A); }
    void  read_posvel ()              const { return nemo_io::read_posvel(); }
    void  read_history()              const { return nemo_io::read_history(); }
  };
  //----------------------------------------------------------------------------
  class nemo_output : public nemo_io {
  public:
    nemo_output(char* file) : nemo_io(file,"w") {}

    float& aux(const int i)                    { return AUXX[i]; }
    float& vec(const int j)                    { return VECTOR[j]; }
    float& cmx(const int j)                    { return CMPHASE[j]; }
    float& cmv(const int j)                    { return CMPHASE[NDM+j]; }
    float& mat(const int i, const int j)       { return MATRIX[i*NDM+j]; }
    float& pos(const int i, const int j)       { return PHASE[i*TDM+j]; }
    float& vel(const int i, const int j)       { return PHASE[i*TDM+NDM+j]; }
  
    void  open_set    (const Set S)            { nemo_io::open_set(S,false);}
    void  close_set   (const Set S)            { nemo_io::close_set(S,false);}
    void  write_N     (const int n)            { nemo_io::write_N(n); }
    void  write_coords(const CoSys c)          { nemo_io::write_coords(c); }
    void  write_scalar(const Scalar S,
		       const float t)          { nemo_io::write_scalar(S,t); }
    void  write_vector(const Vector V)         { nemo_io::write_vector(V); }
    void  write_cofm  ()                       { nemo_io::write_cofm(); }
    void  write_matrix(const Matrix M)         { nemo_io::write_matrix(M); }
    void  write_array (const Array A)          { nemo_io::write_array(A); }
    void  write_posvel()                       { nemo_io::write_posvel(); }
    void  write_history()                      { nemo_io::write_history(); }
  };
}
//------------------------------------------------------------------------------
#undef NDM
#undef TDM
//------------------------------------------------------------------------------
#endif // ALLOW_NEMO
#endif // included_nmio_h
