// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// soul.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class soul                fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_soul_h
#define included_soul_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_flag_h
#include <tree/flag.h>
#endif

#ifndef included_node_h
#include <tree/node.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// some macros                                                                  
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// macro steering the way the sub-souls are organized                           
//                                                                              
#undef SUBSOUL_LLIST
#undef SUBSOUL_ARRAY
// #define SUBSOUL_LLIST        // as linked list                               
#define SUBSOUL_ARRAY           // as array (default)                           

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
#ifdef USE_BODIES
  class body;                                      // forward declare class body
#endif // USE_BODIES
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::soul                                                       
  //////////////////////////////////////////////////////////////////////////////
  class soul :
    private node,
    public  flag
  {
    soul(const soul&);                             // not implemented           
    soul& operator=(const soul&);                  // not implemented           
#ifdef USE_BODIES                                  // friends:                  
    friend class body;                             //   class body (body.h)     
#endif // USE_BODIES
    friend class cell;                             //   class cell (cell.h)     
    friend class builder;                          //   class builder (boxx.h)  
  private:                                         // data members:             
    real        MASS;                              //   mass of body            
    vect        ACC;                               //   acceleration            
    real        POT;                               //   potential               
    real        RHO;                               //   mass density            
    real        EPS;                               //   individual softening    
#ifdef SUBSOUL_LLIST
    soul       *NEXT;                              //   sister soul             
  public:
    soul* const& next() const { return NEXT; }
  private:
#endif
#ifdef USE_BODIES
    const body *MYBODY;                            //   pointer to my body      
#else  // USE_BODIES
    uint        MYELEM;                            //   array index of my "body"
#endif // USE_BODIES
  public:
    // constructor
    soul() {}
    // some manipulators
#ifdef USE_BODIES            //  for use with bodies only                       
    void link_and_set_up(const body*);             //   set my body, cofm & mass
    void set_up         ();                        //   set cofm, mass from body
    void copy_body_flag ();                        //   copy full body flag     
    void add_body_flag  ();                        //   add full body flag      
    void reset_flag     ();                        //   copy flag               
    void reset_grav     ();                        //   set acc,pot=0, copy flag
#else                        // for use with arrays only                        
    void link_and_set_up(const areal,              //   set my elem, cofm & mass
			 const vect, const uint);  //   I: mass, cofm, elem     
    void copy_body_flag (const int*);              //   copy full body flag     
    void add_body_flag  (const int*);              //   add full body flag      
    void set_up         (const areal*,             //   set cofm, mass from body
			 const areal*[NDIM]);      //   I: arrays with mass,cofm
    void update_grav    (areal*[NDIM],areal*);     //   copy acc,pot into arrays
    void update_dens    (areal*);                  //   copy rho into array     
    void reset_flag     (const int*);              //   copy MYELEM's sink flag 
    void reset_rho      (const int*);              //   set rho=0,  copy flag   
    void reset_grav     (const int*);              //   set acc,pot=0, copy flag
    void reset_grav     (const int*, const areal*);//   set acc,pot=0, eps, flag
#endif
    void normalize_grav ();                        // acc,pot     /= mass       
    // non-const data access via members
    vect& acc() { return ACC; }
    real& pot() { return POT; }
    real& rho() { return RHO; }
    // const data access via friends
#ifdef USE_BODIES
    friend const body*const& mybody (const soul*const&S) { return S->MYBODY; }
#else
    friend const uint&myelem    (const soul*const&S) { return S->MYELEM; }
#endif
    friend const vect&cofm      (const soul*const&S) { return S->pos(); }
    friend const real&mass      (const soul*const&S) { return S->MASS; }
    friend const real&eps       (const soul*const&S) { return S->EPS; }
    friend const vect&acc       (const soul*const&S) { return S->ACC; }
    friend const real&pot       (const soul*const&S) { return S->POT; }
    friend const real&rho       (const soul*const&S) { return S->RHO; }
    // boolean information
    friend       bool is_sink   (const soul*const&S) { return S->is_set(SINK); }
    friend       bool is_source (const soul*const&S) { return S->MASS != zero; }
    friend       bool in_subtree(const soul*const&S) { return S->is_set(SUBTREE);}
    // flag manipulations
    void set_sink_flag (const int  f) { flag::setpartflag(f,SINK); }
    void set_sink_flag (const flag*f) { flag::setpartflag(f,SINK); }
    void flag_for_subtree()           { flag::addflag(SUBTREE); }
    void unflag_subtree()             { flag::unflag(SUBTREE); }
  };
}
////////////////////////////////////////////////////////////////////////////////
#ifdef USE_BODIES
#ifndef included_body_h
#include <body.h>
#endif
#endif
////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
#ifdef USE_BODIES
  //----------------------------------------------------------------------------
  inline void soul::set_up()
    // set up my mass and position                                              
  {
    MASS  = mass(MYBODY);
    pos() = ppos(MYBODY);
  }
  //----------------------------------------------------------------------------
  inline void soul::link_and_set_up(const body* B)
    // link to body and set up my mass and position                             
  {
    MYBODY = B;
    set_up();
  }
  //----------------------------------------------------------------------------
  inline void soul::copy_body_flag()
    // copy flags from my body                                                  
  {
    setpartflag(MYBODY,BODY_FLAGS);
  }
  //----------------------------------------------------------------------------
  inline void soul::add_body_flag()
    // add in flags from my body                                                
  {
    addpartflag(MYBODY,BODY_FLAGS);
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_flag()
    // reset flag to the sink flag of my body                                   
  {
    set_sink_flag(MYBODY);
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav()
    // reset flag to the sink flag of my body, set softening, reset acc & pot   
  {
    set_sink_flag(MYBODY);
    EPS = eps(MYBODY);
    ACC = zero;
    POT = zero;
  }
  //----------------------------------------------------------------------------
#else  // USE_BODIES
  //----------------------------------------------------------------------------
  inline void soul::link_and_set_up(const areal m,
				    const vect x,
				    const uint i)
  {
    MYELEM = i;
    MASS   = m;
    pos()  = x;
  }
  //----------------------------------------------------------------------------
  inline void soul::set_up(const areal*m, const areal*x[NDIM])
  {
    MASS     = m[MYELEM];
    pos()[0] = x[0][MYELEM];
    pos()[1] = x[1][MYELEM];
#if NDIM==3
    pos()[2] = x[2][MYELEM];
#endif
  }
  inline void soul::update_grav(areal*a[NDIM], areal*p)
  {
    if(p) p[MYELEM] = POT;
    a[0][MYELEM] = ACC[0];
    a[1][MYELEM] = ACC[1];
#if NDIM==3
    a[2][MYELEM] = ACC[2];
#endif
  }
  //----------------------------------------------------------------------------
  inline void soul::update_dens(areal*rh)
  { rh[MYELEM] = RHO; }
  //----------------------------------------------------------------------------
  inline void soul::copy_body_flag(const int* F)
  { setpartflag(F[MYELEM],BODY_FLAGS); }
  //----------------------------------------------------------------------------
  inline void soul::add_body_flag(const int* F)
  { addpartflag(F[MYELEM],BODY_FLAGS); }
  //----------------------------------------------------------------------------
  inline void soul::reset_flag(const int* F)
  { 
    set_sink_flag(F[MYELEM]);
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_rho(const int* F)
  { 
    set_sink_flag(F[MYELEM]);
    RHO = zero;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav(const int* F)
  { 
    set_sink_flag(F[MYELEM]);
    ACC = zero;
    POT = zero;
  }
  //----------------------------------------------------------------------------
  inline void soul::reset_grav(const int* F, const areal* E)
  { 
    set_sink_flag(F[MYELEM]);
    EPS = E[MYELEM];
    ACC = zero;
    POT = zero;
  }
  //----------------------------------------------------------------------------
#endif // USE_BODIES
  //----------------------------------------------------------------------------
  inline void soul::normalize_grav()
  {
    if(MASS>zero) {
      register real im = one/MASS;
      ACC *= im; POT *= im;
    }
  }
}
////////////////////////////////////////////////////////////////////////////////
#endif  // included_soul_h
