// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// step.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class TimeStep            fully inline                                      |
// class BlockStep           fully inline                                      |
// class DaccScheme          fully inline                                      |
// class RhoScheme           fully inline                                      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_step_h
#define included_step_h

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif

#ifndef included_iomanip
#include <iomanip>
#define included_iomanip
#endif

#ifndef included_body_h
#include <body.h>
#define included_body_h
#endif

////////////////////////////////////////////////////////////////////////////////
namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::TimeStep                                                   
  //////////////////////////////////////////////////////////////////////////////
  class TimeStep {
    friend class BlockStep;
  private:
    mutable uint  N,                     // # bodies with this step
                  NC;                    // # bodies with this and shorter steps
    real          TAU, TAUSQ;            // time step
    mutable real  TPOS,                  // time of true position of bodies
                  DT, DTH;               // DT = actual time - TPOS

    void re_set(const int f=0, const real t=zero){ TAU =pow_int(half,f);
    TAUSQ = TAU*TAU;
    TPOS= t; DT=DTH=zero; }
    TimeStep   (const int f=0, const real t=zero){ N=0u; NC=0u; re_set(f,t); }
    void        reset_count ()             const { N=0u; NC=0u; }
    void        set_predict (const real t) const { DT = t-TPOS; DTH=half*DT; }
    void        dump   (std::ostream&to)   const { to<<N<<" "<<NC<<" "<<TAU<<" "
						     <<TPOS<<" "<<DT<<"\n"; }
    void        add               ()       const { N++; NC++;   }
    void        add_to_shorter    ()       const {      NC++;   }
    void        join_from_shorter ()       const { N++;         }
    void        join_from_longer  ()       const { N++; NC++;   }
    void        leave_to_shorter  ()       const { N--;         }
    void        leave_to_longer   ()       const { N--; NC--;   }
    void        move_time         ()       const { TPOS += TAU; }
    const real& time_tpos         ()       const { return TPOS; }
    const real& tau               ()       const { return TAU;  }
    const real& tausq             ()       const { return TAUSQ;}
    const real& dt                ()       const { return DT;   }
    const real& dt_half           ()       const { return DTH;  }
    const uint& number            ()       const { return N;    }
    const uint& acc_number        ()       const { return NC;   }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::BlockStep                                                  
  //////////////////////////////////////////////////////////////////////////////
  class BlockStep {
    friend class DaccScheme;
    friend class RhoScheme;
  public:
    enum too_short {
      go_longer    = 1,          // step is/was too short
      forbid_short = 2,          // step okay, next one unlikely to be too long
      allow_short  = 4           // step okay, but next one may be too long
    };
  private:
    int            H0;
    indx           NSTEPS;
    TimeStep      *TS, *SHORTEST;
    mutable real   TIME;
    real               F, FQ, FQTOL, FQSH, FQSHTOL; 
    void    join_from_shorter(const indx l) const { (TS+l)->join_from_shorter(); }
    void    join_from_longer (const indx l) const { (TS+l)->join_from_longer();  }
    void    leave_to_shorter (const indx l) const { (TS+l)->leave_to_shorter();  }
    void    leave_to_longer  (const indx l) const { (TS+l)->leave_to_longer();   }
  protected:
    void    add              (const indx)   const;
    void    reset_counts     () const;
    void    reset_steps      (const int,  const indx, const real=zero);
    void    reset_fac        (const real, const real=two);
  public:
    BlockStep                () : TS(0) {}
    BlockStep                (const int,             // h0, tau_max = 2^-h0
			      const uint,            // #steps
			      const real,            // actual time
			      const real,            // f
			      const real=two);       // tol^2
    ~BlockStep                ()              { delete[] TS;}

    void        clock_on        ()             const;
    void        clock_back      ()             const;
    void        set_predict     ()             const;
    void        dump            (std::ostream&)const;
    void        predict_pos     (body*)        const;
    void        move_times      (const indx)   const;
    indx        longest_moving  (const uint)   const;
    bool        is_leap_frog    ()             const {return NSTEPS == 1; }
    const real& time            ()             const {return TIME; }
    const real& tau             (const indx l) const {return(TS+l)->tau(); }
    const real& tausq           (const indx l) const {return(TS+l)->tausq(); }
    const real& dt              (const indx l) const {return(TS+l)->dt(); }
    const real& dt_half         (const indx l) const {return(TS+l)->dt_half(); }
    const real& time_tpos       (const indx l) const {return(TS+l)->time_tpos();}
    const uint& number          (const indx l) const {return(TS+l)->number(); }
    const real& tau             (const body*B) const {return tau(level(B)); }
    const real& tausq           (const body*B) const {return tausq(level(B)); }
    const real& dt              (const body*B) const {return dt(level(B)); }
    const real& dt_half         (const body*B) const {return dt_half(level(B)); }
    const real& time_tpos       (const body*B) const {return time_tpos(level(B));}
    const real& factor          ()             const {return F; }
    void        move_proper     (body*)        const;
    void        goto_shorter    (body*)        const;
    void        goto_longer     (body*)        const;
    indx        shortest_step   ()             const {return NSTEPS-1; }
    const real& shortest_tau    ()             const {return SHORTEST->tau(); }
    const uint& No_in_level     (const indx l) const {return(TS+l)->number(); }
    const uint& No_shorter_level(const indx l) const {return(TS+l)->acc_number();}
    const  int& h0              ()             const {return H0; }
    void        short_stats     (std::ostream&)const;
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::DaccScheme                                                 
  //////////////////////////////////////////////////////////////////////////////
  class DaccScheme {
  private:
    typedef BlockStep::too_short tooshort;
  public:
    static void     assign_to_level(const BlockStep*, body*);
    static bool     step_too_long  (const BlockStep*, const body*);
    static tooshort step_too_short (const BlockStep*, const body*);
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::RhoScheme                                                  
  //////////////////////////////////////////////////////////////////////////////
  class RhoScheme {
  private:
    static bool     too_long       (const BlockStep*, const real&, const indx);
  public:
    static void     assign_to_level(const BlockStep*, body*);
    static bool     step_too_long  (const BlockStep*, const body*);
    static bool     step_too_short (const BlockStep*, const body*);
  };

  //MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM//
  //WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW//
  //                                                                          //
  //                      INLINE FUNCTION DEFINITIONS                         //
  //                                                                          //
  //MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM//
  //WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW//

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::BlockStep                                                  
  //////////////////////////////////////////////////////////////////////////////
  inline void BlockStep::reset_steps(const int hz, const indx Ns, const real t)
  {
    if(TS) delete[] TS;
    NSTEPS   = Ns;
    TS       = new TimeStep[NSTEPS];
    SHORTEST = TS+NSTEPS-1;
    TIME     = t;
    H0       = hz;
    register int  h;
    register indx i;
    for(i=0, h=H0; i<NSTEPS; i++,h++) (TS+i)->re_set(h,t);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::reset_counts() const
  {
    for(register TimeStep* s=TS; s<=SHORTEST; s++) s->reset_count();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::reset_fac(const real f, const real tq)
  {
    F=f; FQ=F*F; FQTOL=FQ*tq; FQSH=0.0625*FQ; FQSHTOL=FQSH*tq;
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::predict_pos(body* B) const
  {
    register vect newpos(oacc(B));
    newpos   *= dt_half(B);
    newpos   += vel    (B);
    newpos   *= dt     (B);
    newpos   += tpos   (B);
    B->ppos() = newpos;
  }
  //----------------------------------------------------------------------------
  inline BlockStep::BlockStep(const int h0, const uint Ns, const real t,
			      const real f1, const real tq)
    : TS(0)
  {
    reset_steps(h0,Ns,t);
    reset_fac  (f1,tq);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::move_times(const indx l) const
  {
    for(register indx i=l; i<NSTEPS; i++) (TS+i)->move_time();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::dump(std::ostream&to)  const
  {
    for(register indx i=0; i<NSTEPS; i++) {
      to<<i<<": ";
      (TS+i)->dump(to);
    }
  }
  //----------------------------------------------------------------------------
  inline indx BlockStep::longest_moving(const uint T) const
  {
    register indx  i=NSTEPS-1;
    register uint  t=T;
    for(; !(t&1) && i>0; t>>=1, i--);
    return i;
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::clock_on() const
  {
    TIME += shortest_tau();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::clock_back() const
  {
    TIME -= shortest_tau();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::set_predict() const
  {
    for(register TimeStep* s=TS; s<=SHORTEST; s++) s->set_predict(TIME);
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::goto_shorter(body* B) const
  {
    leave_to_shorter((B->level())++);
    join_from_longer(level(B));
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::goto_longer(body* B) const
  {
    leave_to_longer((B->level())--);
    join_from_shorter(level(B));
    B->forbid_longer();
    B->allow_shorter();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::move_proper(body* B) const
    // note that dt(B) = t_acc - t_oacc = t_ppos - t_tpos
  {
    register vect dvel(oacc(B));
    dvel     += acc    (B);
    dvel     *= dt_half(B);
    B->vel() += dvel;
    B->tpos() = ppos(B);
    B->oacc() = acc(B);
    B->allow_longer();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::add(const indx l) const
  {
    (TS+l)->add();
    for(register TimeStep* s=TS; s<TS+l; s++) s->add_to_shorter();
  }
  //----------------------------------------------------------------------------
  inline void BlockStep::short_stats(std::ostream& to) const
  {
    for(register indx l=0; l<NSTEPS; l++) to<<std::setw(7)<<number(l)<<" ";
  }
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::DaccScheme                                                 
  //////////////////////////////////////////////////////////////////////////////
  inline bool DaccScheme::step_too_long (const BlockStep* s, const body* B)
  {
    return dist_sq(oacc(B),acc(B)) > s-> FQSH * sum_sq(oacc(B),acc(B));
  }
  //----------------------------------------------------------------------------
  inline BlockStep::too_short DaccScheme::step_too_short(const BlockStep* s,
							 const body* B)
  {
    register real dacc=dist_sq(oacc(B),acc(B)), macc=sum_sq(oacc(B),acc(B));
    if(dacc < s->FQSH    * macc) return BlockStep::go_longer;
    if(dacc < s->FQSHTOL * macc) return BlockStep::forbid_short;
    return BlockStep::allow_short;
  }
  //----------------------------------------------------------------------------
  inline void DaccScheme::assign_to_level(const BlockStep* s, body* B)
  {
    register indx lev;
    register real FQSH_x_accm = sum_sq (oacc(B),acc(B)) * s->FQSH,
      dacc = dist_sq(oacc(B),acc(B));
    for(lev = s->shortest_step();
	lev > 0u && dacc < FQSH_x_accm;
	lev--, dacc *= 4.);
    s->add(B->level()=lev);
  }

  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::RhoScheme                                                  
  //////////////////////////////////////////////////////////////////////////////
  inline bool RhoScheme::too_long(const BlockStep*s, const real&rh, const indx l)
  {
    return rh * s->tausq(l) > s->F;
  }
  //----------------------------------------------------------------------------
  inline bool RhoScheme::step_too_long(const BlockStep*s, const body* B)
  {
    return too_long(s,rho(B),level(B));
  }
  //----------------------------------------------------------------------------
  inline bool RhoScheme::step_too_short(const BlockStep*s, const body* B)
  {
    return level(B)>0 && !too_long(s,rho(B),level(B)-1);
  }
  //----------------------------------------------------------------------------
  inline void RhoScheme::assign_to_level(const BlockStep*s, body* B)
  {
    register indx lev;
    for(lev = 0;
	lev < s->shortest_step() && too_long(s,rho(B),lev);
	lev++);
    s->add(B->level()=lev);
  }
}
////////////////////////////////////////////////////////////////////////////////
#endif   // included_step_h
