// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// stic.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class Ssoul                       fully inline                              |
// class Scell                       fully inline                              |
// class Stree                       see stic.cc                               |
// class sticky_finder               see stic.cc                               |
// class neighbour_finder            see stic.cc                               |
// class neighbour_counter           see stic.cc                               |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_stic_h
#define included_stic_h

#ifndef included_cell_h
#include <tree/cell.h>
#endif

#ifndef included_subt_h
#include <tree/subt.h>
#endif

#ifndef included_soul_h
#include <tree/soul.h>
#endif

////////////////////////////////////////////////////////////////////////////////
#define LoopDims  for(register indx dim=0; dim<NDIM; dim++) 
////////////////////////////////////////////////////////////////////////////////

namespace fasttree {
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::Ssoul                                                      
  //////////////////////////////////////////////////////////////////////////////
  class Ssoul : public bsoul<Ssoul> {
  private:
    real SIZE;
    real SIZEQ;
    vect VEL;
    uint NUM;
  public:
#ifdef USE_BODIES
    void set_sticky() { SIZE=fasttree::size(MYBODY); VEL=fasttree::vel(MYBODY); }
    void set_sph   () { SIZE=fasttree::size(MYBODY); SIZEQ=SIZE*SIZE; }
#else  // USE_BODIES
    void set_sticky(const areal*s, const areal*v[NDIM])
    { SIZE=s[MYELEM]; LoopDims VEL[dim] = v[dim][MYELEM]; NUM=0; }
    void set_sph   (const areal*s) { SIZE=s[MYELEM]; SIZEQ=SIZE*SIZE; NUM=0; }
#endif // USE_BODIES
#ifndef C_AND_F
    static void dump_head(std::ostream&o)
    {
      bsoul<Ssoul>::dump_head(o);
      o<<"           size [velocity]";
    }
    void dump(std::ostream&o, const Ssoul*S0) const
    {
      bsoul<Ssoul>::dump(o,S0);
      o<<" "<<std::setw(5)<<std::setprecision(4)<<SIZE;
      if( is_sticky())
	for(register indx i=0;i<NDIM;i++)
	  o<<" "<<std::setw(7)<<std::setprecision(4)<<VEL[i];
    }
#endif
    void        inc  ()        { NUM++; }
    const real& size ()  const { return SIZE; }
    const real& sizeq()  const { return SIZEQ; }
    const vect& vel  ()  const { return VEL; }
    const uint& num  ()  const { return NUM; }
//     // friends
//     friend Ssoul* next   (Ssoul*X)  { return X->next(); }
//     friend void   incsoul(Ssoul*&X) { X = X->next(); }
  };
  //----------------------------------------------------------------------------
  inline             bool   is_sink(const Ssoul*S) { return S->is_sink(); }
#ifdef USE_BODIES
  inline const body* const &mybody (const Ssoul*S) { return S->mybody(); }
#else  // USE_BODIES
  inline uint        const &myelem (const Ssoul*S) { return S->myelem(); }
#endif // USE_BODIES
  inline const       vect  &pos    (const Ssoul*S) { return S->pos(); }
  inline const       vect  &vel    (const Ssoul*S) { return S->vel(); }
  inline const       real  &size   (const Ssoul*S) { return S->size(); }
  inline const       real  &sizeq  (const Ssoul*S) { return S->sizeq(); }
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::Scell                                                      
  //////////////////////////////////////////////////////////////////////////////
  class Scell : public bcell<Ssoul,Scell> {
  private:
    vect VEL;
    real SIZE;
    union {
      real VRAD;
      real RMAX;
    };
  public:
    vect& vel () { return VEL; }
    real& size() { return SIZE; }
    real& rmax() { return RMAX; }
    real& vrad() { return VRAD; }
#ifndef C_AND_F
    static void dump_head(std::ostream&o)
    {
      bcell<Ssoul,Scell>::dump_head(o);
      o<<"           size     rmax/ velocity           [vrad]";
    }
    void dump(std::ostream&o, const Ssoul*S0, const Scell*C0) const
    {
      bcell<Ssoul,Scell>::dump(o,S0,C0);
      o<<" "<<std::setw(6)<<std::setprecision(4)<<SIZE;
      if(is_sticky()) {
	for(register indx i=0; i<NDIM; i++)
	  o<<" "<<std::setw(9)<<std::setprecision(4)<<VEL[i];
	o<<" "<<std::setw(5)<<std::setprecision(4)<<VRAD;
      }
      else if (is_sph())
	o<<" "<<std::setw(5)<<std::setprecision(4)<<RMAX;
    }
#endif
    const real& vrad()  const { return VRAD; }
    const real& size()  const { return SIZE; }
    const real& rmax()  const { return RMAX; }
    const vect& vel ()  const { return VEL; }
  };
  //----------------------------------------------------------------------------
  inline        bool   is_sink(const Scell*C) { return C->is_sink(); }
  inline        bool   al_sink(const Scell*C) { return C->al_sink(); }
  inline        bool   is_twig(const Scell*C) { return C->is_twig(); }
  inline const  vect  &pos    (const Scell*C) { return C->pos(); }
  inline const  vect  &vel    (const Scell*C) { return C->vel(); }
  inline const  real  &rmax   (const Scell*C) { return C->rmax(); }
  inline const  real  &size   (const Scell*C) { return C->size(); }
  inline const  real  &vrad   (const Scell*C) { return C->vrad(); }
  inline const  uint  &number (const Scell*C) { return C->number(); }
  inline const  indx  &depth  (const Scell*C) { return C->depth(); }
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::Stree                                                      
  //////////////////////////////////////////////////////////////////////////////
  class Stree : public subtree<Ssoul,Scell> {
  private:
#ifndef USE_BODIES
    const areal *S,**V;             // arrays with sizes and velocities
#endif
    Stree();                        // not implemented
    Stree(const Stree&);            // not implemented
  public:
#ifdef USE_BODIES
    Stree(cell*C0, cell*CN)
      : subtree<Ssoul,Scell>(C0,CN) {}
    void set_num(body*) const;
#else  // USE_BODIES
    Stree(cell*C0, cell*CN, const areal*s, const areal**v)
      : subtree<Ssoul,Scell>(C0,CN), S(s), V(v) {}
    void set_num(areal*) const;
#endif // USE_BODIES
    void prepare_sticky();
    void prepare_sph   ();
    Scell*       &root()       { return C0; }
    Scell*       &end ()       { return CN; }
    Scell* const &root() const { return C0; }
    Scell* const &end () const { return CN; }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::sticky_finder                                              
  //////////////////////////////////////////////////////////////////////////////
  class sticky_finder {
  public:
#ifdef USE_BODIES
    typedef const body* body_pair[2];
#else  // USE_BODIES
    typedef       uint  body_pair[2];
#endif // USE_BODIES
  private:
    const real    TAU;
    const uint    NMAX;
    body_pair    *BLIST;
    mutable uint  N;

    void add_pair(Ssoul*, Ssoul*) const;
    void many    (const bool, Ssoul*, Ssoul*, const Ssoul*) const;
    bool many    (Scell*, Ssoul*) const;
    bool many    (Scell*, Scell*) const;
    bool many    (Scell*) const;
    bool discard (Scell*, Ssoul*) const;
    bool discard (Scell*, Scell*) const;
  public:
    sticky_finder   (const real t,   	   // time step tau
		     const uint n,   	   // max size of pair list
		     body_pair* l)   	   // list of pairs
      : TAU(t), NMAX(n), BLIST(l), N(0) {}
    const uint&      actual_size_of_list() { return N; }
    bool interact   (Scell*) const;
    bool interact   (Scell*, Scell*) const;
    bool interact   (Scell*, Ssoul*) const;
    void interact   (Ssoul*, Ssoul*) const;
    bool split_first(const Scell*A, const Scell*B) const
    { return is_twig(B) || size(A)+TAU*vrad(A) > size(B)+TAU*vrad(B); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::neighbour_finder                                           
  //////////////////////////////////////////////////////////////////////////////
  class neighbour_finder {
  public:
#ifdef USE_BODIES
    typedef const body* body_pair[2];
#else  // USE_BODIES
    typedef       uint  body_pair[2];
#endif // USE_BODIES
  private:
    const uint    NMAX;
    body_pair    *BLIST;
    mutable uint  N;

    void add_pair(Ssoul*, Ssoul*) const;
    void many    (const bool, Ssoul*, Ssoul*, const Ssoul*) const;
    bool many    (Scell*, Ssoul*) const;
    bool many    (Scell*, Scell*) const;
    bool many    (Scell*) const;
    bool discard (Scell*, Ssoul*) const;
    bool discard (Scell*, Scell*) const;
  public:
    neighbour_finder   (const uint n,   	   // max size of pair list
			body_pair *l)   	   // list of pairs
      : NMAX(n), BLIST(l), N(0) {}
    const uint&      actual_size_of_list() { return N; }
    bool interact   (Scell*) const;
    bool interact   (Scell*, Scell*) const;
    bool interact   (Scell*, Ssoul*) const;
    void interact   (Ssoul*, Ssoul*) const;
    bool split_first(const Scell* A, const Scell* B) const
    { return is_twig(B) || size(A) > size(B); }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class fasttree::neighbour_counter                                          
  //////////////////////////////////////////////////////////////////////////////
  class neighbour_counter {
  private:
    void many    (Ssoul*, const bool, Ssoul*, const Ssoul*, const bool) const;
    bool many    (Scell*, Ssoul*) const;
    bool many    (Scell*, Scell*) const;
    bool many    (Scell*) const;
    bool discard (Scell*, Ssoul*) const;
    bool discard (Scell*, Scell*) const;
  public:
    bool interact   (Scell*) const;
    bool interact   (Scell*, Scell*) const;
    bool interact   (Scell*, Ssoul*) const;
    void interact   (Ssoul*, Ssoul*) const;
    bool split_first(const Scell* A, const Scell* B) const
    { return is_twig(B) || size(A) > size(B); }
  };
}
#undef LoopDims
//------------------------------------------------------------------------------
#endif // included_stic_h
