// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// subt.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// template<class ssoul> class bsoul                    fully inline           |
// template<class ssoul, class scell> class bcell       fully inline           |
// template<class ssoul, class scell> class subtree     fully inline           |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_subt_h
#define included_subt_h

#ifndef included_auxx_h
#include <tree/auxx.h>
#endif

#ifndef included_cell_h
#include <tree/cell.h>
#endif

#ifndef  C_AND_F

#ifndef included_iomanip
#include <iomanip>
#define included_iomanip
#endif

#endif // C_AND_F

////////////////////////////////////////////////////////////////////////////////
// some macros                                                                  
////////////////////////////////////////////////////////////////////////////////
//                                                                              
// macro steering the way the sub-souls are organized                           
//                                                                              
#undef SUBTSUBSOUL_LLIST
#undef SUBTSUBSOUL_ARRAY
#define SUBTSUBSOUL_LLIST         // as linked list                             
//#define SUBTSUBSOUL_ARRAY       // as array                                   

//                                                                              
// macro steering the way the sub-cells are organized                           
//                                                                              
#undef SUBTSUBCELL_LLIST
#undef SUBTSUBCELL_ARRAY
#define SUBTSUBCELL_LLIST          // as linked list                            
// #define SUBTSUBCELL_ARRAY       // as array                                  

namespace fasttree {

  //============================================================================
  // base class for a subtree soul;                                            |
  // ssoul=soul type of subtree, derived from bsoul                            |
  //                                                                           |
  // create a subtree soul type by adding your features, e.g.                  |
  //                                                                           |
  // class Tsoul : public bsoul<Tsoul> {                                       |
  // private:                                                                  |
  //   real SIZE;                                 // added data                |
  // public:                                                                   |
  //   const real& size() const { return SIZE; }  // added function            |
  //         real& size()       { return SIZE; }  // added function            |
  // }                                                                         |
  template<class ssoul>
  class bsoul : public flag {
  protected:
    vect        POS;                                  //  position              
#ifdef SUBTSUBSOUL_LLIST
    ssoul      *NEXT;                                 //  pter to sister soul   
  public:
    ssoul*        next()       { return NEXT; }
    ssoul* const& next() const { return NEXT; }
    void  thread(ssoul*s)      { NEXT   = s; }
  protected:
#endif
#ifdef USE_BODIES                   // versions for use of bodies               
    const body *MYBODY;                               //  where is my body      
  public:
#ifdef SUBTSUBSOUL_LLIST
    bsoul() : NEXT(0), MYBODY(0) {}
#else
    bsoul() : MYBODY(0) {}
#endif
    const body* const& mybody() const  { return MYBODY; }
    void  link(const body*b)           { MYBODY = b; }
#else  // USE_BODIES                // versions for use of arrays               
    uint MYELEM;                                      //  array index of my body
  public:
    bsoul() : NEXT(0), MYELEM(0) {}
    uint const&       myelem() const { return MYELEM; }
    void link  (const uint i)        { MYELEM = i; }
#endif // USE_BODIES                // common methods                           
    void copy_body_flag(const flag*f)
    {
      flag::setpartflag(f,BODY_FLAGS);
    }
    vect& pos()            { return POS; }
    bool  is_sink()  const { return flag::is_set(SINK); }
    bool  is_sticky()const { return flag::is_set(STICKY); }
    bool  is_sph()   const { return flag::is_set(SPH); }
    const vect& pos()const { return POS; }
#ifndef C_AND_F
    static void dump_head(std::ostream&);
    void        dump     (std::ostream&, const ssoul*) const;
#endif // C_AND_F
  };
  //============================================================================
  // base class for a subtree cell;                                            |
  // ssoul=soul type of subtree, derived from bsoul                            |
  // scell=cell type of subtree, derived from bcell                            |
  //                                                                           |
  // create a subtree cell type by adding your features, e.g.                  |
  //                                                                           |
  // class Tcell : public bcell<Tsoul,Tcell> {                                 |
  // private:                                                                  |
  //   real SIZE;                                 // added data                |
  // public:                                                                   |
  //   const real& size() const { return SIZE; }  // added function            |
  //         real& size()       { return SIZE; }  // added function            |
  // }                                                                         |
  template<class ssoul, class scell> class subtree;
  template<class ssoul, class scell>
  class bcell : public flag {
    friend class subtree<ssoul,scell>;
  private:
    indx   DEPTH;                           //    subtree depth downward
    uint   NUMBER;                          //    number of subtree souls
    vect   POS;                             //    position
    ssoul *SOULS;                           //    pter to first soul child
    ssoul *FIRST;                           //    pter to first soul descendent
    scell *CELLS;                           //    pter to first cell child
#ifndef SUBTSUBSOUL_LLIST
    indx   NSOULS;                          //    # sub-souls
#endif
#ifdef  SUBTSUBCELL_LLIST
    scell *NEXT;                            //    pter to next sister cell
#else
    indx   NCELLS;                          //    # sub-cells
#endif
  public:
    bcell() : DEPTH(0), NUMBER(0), POS(zero), 
      SOULS(0), FIRST(0), CELLS(0), NEXT(0)  {}
    void   thread(scell* c) { NEXT   = c; }
#ifndef C_AND_F
    static void   dump_head(std::ostream&);
    void   dump  (std::ostream&, const ssoul*, const scell*) const;
#endif // C_AND_F
    void        reset_flag()     { flag::reset();
    flag::addflag(AL_SINK); }
    bool        is_sink()  const { return flag::is_set(SINK); }
    bool        al_sink()  const { return flag::is_set(AL_SINK);}
    bool        is_twig()  const { return CELLS==0; }
    bool        is_sticky()const { return flag::is_set(STICKY); }
    bool        is_sph()   const { return flag::is_set(SPH); }
    const indx& depth()    const { return DEPTH; }
    const uint& number()   const { return NUMBER; }
          vect& pos()            { return POS; }
    const vect& pos()      const { return POS; }
    void add_body_flag(const ssoul*s)
    {
      flag::addpartflag(s,BODY_FLAGS);
      if(!s->is_sink()) flag::unflag(AL_SINK);
    }
    void add_body_flag(const scell*c)
    {
      flag::addpartflag(c,BODY_FLAGS);
      if(!c->al_sink()) flag::unflag(AL_SINK);
    }
    //                                                                          
    // access to sub-souls                                                      
    //                                                                          
    class subsoul_iterator {
      ssoul* X;
    public:
      subsoul_iterator(ssoul*x) : X(x) {}
      subsoul_iterator(const subsoul_iterator&I) : X(I.X) {}
      subsoul_iterator&operator=(const subsoul_iterator&I) { X=I.X; return *this; }
#ifdef SUBTSUBSOUL_LLIST
      subsoul_iterator&operator++() { X=X->next(); return *this; }
      subsoul_iterator next      () const { return subsoul_iterator(X->next()); }
#else
      subsoul_iterator&operator++() { ++X; return *this; }
      subsoul_iterator next      () const { return subsoul_iterator(X+1); }
#endif
      subsoul_iterator operator++(int) { subsoul_iterator tmp(*this);
                                         operator++(); return tmp; }
      bool operator==(const subsoul_iterator&I) const { return X == I.X; }
      bool operator!=(const subsoul_iterator&I) const { return X != I.X; }
      operator ssoul*  ()                       { return X; }
      ssoul& operator* ()                       { return*X; }
      ssoul* operator->()                       { return X; }
    };
    subsoul_iterator begin_subsouls() const{return subsoul_iterator(SOULS); }
#ifdef SUBTSUBSOUL_LLIST
    subsoul_iterator end_subsouls() const {return subsoul_iterator(0); }
#else
    subsoul_iterator end_subsouls() const {return subsoul_iterator(SOULS+NSOULS);}
#endif
    //                                                                          
    // access to sub-cells                                                      
    //                                                                          
#ifdef SUBTSUBCELL_LLIST
    scell*     &next()       { return NEXT; }
    scell const&next() const { return NEXT; }
#endif
    class subcell_iterator {
      scell* X;
    public:
      subcell_iterator(scell*x) : X(x) {}
      subcell_iterator(const subcell_iterator&I) : X(I.X) {}
      subcell_iterator&operator=(const subcell_iterator&I) { X=I.X; return *this; }
#ifdef SUBTSUBCELL_LLIST
      subcell_iterator&operator++() { X=X->next(); return *this; }
      subcell_iterator next      () const { return subcell_iterator(X->next()); }
#else
      subcell_iterator&operator++() { ++X; return *this; }
      subcell_iterator next      () const { return subcell_iterator(X+1); }
#endif
      subcell_iterator operator++(int) { subcell_iterator tmp(*this);
                                         operator++(); return tmp; }
      bool operator==(const subcell_iterator&I) const { return X == I.X; }
      bool operator!=(const subcell_iterator&I) const { return X != I.X; }
      operator scell*  ()                       { return X; }
      scell& operator* ()                       { return*X; }
      scell* operator->()                       { return X; }
    };
    subcell_iterator begin_subcells() const{return subcell_iterator(CELLS); }
#ifdef SUBTSUBCELL_LLIST
    subcell_iterator end_subcells() const {return subcell_iterator(0); }
#else
    subcell_iterator end_subcells() const {return subcell_iterator(CELLS+NCELLS);}
#endif
    //                                                                          
    // access to all souls                                                      
    //                                                                          
    class allsoul_iterator {
      ssoul* X;
    public:
      allsoul_iterator(ssoul*x) : X(x) {}
      allsoul_iterator(const allsoul_iterator&I) : X(I.X) {}
      allsoul_iterator&operator=(const allsoul_iterator&I) { X=I.X; return *this; }
      allsoul_iterator&operator++()              { ++X; return *this; }
      allsoul_iterator operator++(int)           { allsoul_iterator tmp(*this);
                                                   operator++(); return tmp; }
      allsoul_iterator&operator+=(const int i)   { X+=i; return *this; }
      allsoul_iterator&operator-=(const int i)   { X-=i; return *this; }
      allsoul_iterator operator+ (const int i)   { return allsoul_iterator(X+i); }
      allsoul_iterator operator- (const int i)   { return allsoul_iterator(X-i); }
      bool operator==(const allsoul_iterator&I) const { return X == I.X; }
      bool operator!=(const allsoul_iterator&I) const { return X != I.X; }
      bool operator==(const ssoul*S)            const { return X == S; }
      bool operator!=(const ssoul*S)            const { return X != S; }
      operator ssoul*  ()                        { return X; }
      ssoul& operator* ()                        { return*X; }
      ssoul* operator->()                        { return X; }
      friend bool operator==(const ssoul*S,const allsoul_iterator&I) {return I==S;}
      friend bool operator!=(const ssoul*S,const allsoul_iterator&I) {return I!=S;}
    };
    allsoul_iterator begin_allsouls() const{return allsoul_iterator(FIRST); }
    allsoul_iterator last_allsouls()const{return allsoul_iterator(FIRST+NUMBER-1);}
    allsoul_iterator end_allsouls() const {return allsoul_iterator(FIRST+NUMBER);}
  };
  //                                                                          
  //============================================================================
  // template class subtree for a subtree of subsouls and subcells
  // use as base for your subtree (simply add features at your desire)
  template<class ssoul, class scell>
  class subtree {
  protected:
    uint   Ns,Nc;
    ssoul *S0,*SN;
    scell *C0,*CN;
  private:
    static indx link_sub_tree(const cell*,scell**, ssoul**);
    void        cut          (fasttree::cell*, fasttree::cell*);
  public:
    subtree(fasttree::cell*tC0,fasttree::cell*tCN) :
      S0(0), C0(0) { cut(tC0,tCN); }
    ~subtree() { if(S0) delete[] S0; if(C0) delete[] C0; }
    bool is_empty() const { return S0==0; }
#ifndef C_AND_F
    void dumpcells(std::ostream&) const;
    void dumpsouls(std::ostream&) const;
    void dumpcells(const char*) const;
    void dumpsouls(const char*) const;
#endif // C_AND_F
  };
  //============================================================================
#define LoopSubCells(C,X)  for(register cell::subcell_iterator			\
			     X=C->begin_subcells(); X!=C->end_subcells(); X++)
#define LoopSubSouls(C,X)  for(register cell::subsoul_iterator			\
			     X=C->begin_subsouls(); X!=C->end_subsouls(); X++)
  //----------------------------------------------------------------------------
  inline soul* find_soul(const cell* tC)
    // cell with one subtree soul  
  {
    LoopSubCells(tC,tc) if(in_subtree(tc)) return find_soul(tc);
    LoopSubSouls(tC,ts) if(in_subtree(ts)) return ts;
#ifndef C_AND_F
    std::cerr<<" blind ending in subtree cutting 1"<<std::endl;
#else
    printf(" blind ending in subtree cutting 1\n");
#endif
    return 0;
  }
  //----------------------------------------------------------------------------
  inline cell* find_cell(const cell* tC)
    // cell with one subtree cell  
  {
    LoopSubCells(tC,tc) if(in_subtree(tc)) {
      if(is_subtreecell(tc))   return tc;
      if(is_subtreezombie(tc)) return find_cell(tc);
    }
#ifndef C_AND_F
    std::cerr<<" blind ending in subtree cutting 2"<<std::endl;
#else
    printf(" blind ending in subtree cutting 2\n");
#endif
    return 0;
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  indx subtree<ssoul,scell>::link_sub_tree(const cell*tC, scell**Ci, ssoul**Si)
  {
    register cell  *zc;
    register scell *lastc=0, *C =(*Ci)++;        // C: current subtree cell     
    register ssoul *lasts=0;
    register indx   dpmax=1,dp;

    C->POS   = center(tC);                       // copy center from tree cell  
    LoopSubSouls(tC,ts)                          // loop subcells of tree cell  
      if(in_subtree(ts)) {                       //   if soul is also in subtree
#ifdef USE_BODIES
	(*Si)->link(mybody(ts));                 //     link to body            
#else
	(*Si)->link(myelem(ts));                 //     or link to element      
#endif
	(*Si)->pos() = cofm(ts);                 //     set position from soul  
	(*Si)->copy_body_flag(ts);               //     copy body flag from soul
	if(C->SOULS) lasts->thread(*Si);         //     add to linked list      
	else         C->SOULS = *Si;             //     of sub-tree sub-souls   
	lasts = (*Si)++;                         //     update last elem of list
	C->NUMBER++;                             //     increment number        
      }
    register soul *ts;
    LoopSubCells(tC,tc)                          // loop subcells of tree cell  
      if(is_subtreecell(tc)) {                   //   if cell is also in subtree
	if(C->CELLS) lastc->thread(*Ci);         //     add it to linked list   
	else         C->CELLS = *Ci;             //     of sub-tree cells       
	lastc = *Ci;                             //     update last elem of list
	dp = link_sub_tree(tc,Ci,Si);            //     recursively link subcell
	C->NUMBER += lastc->NUMBER;              //     add to number           
	if(dp>dpmax) dpmax=dp;                   //     update maximum depth    
      } else if(is_subtreezombie(tc)) {          //   else if cell=stree zombie 
	zc = find_cell(tc);                      //     find the subtree cell   
	if(C->CELLS) lastc->thread(*Ci);         //     add it to linked list   
	else         C->CELLS = *Ci;             //     of sub-tree cells       
	lastc = *Ci;                             //     update last elem of list
	dp = link_sub_tree(zc,Ci,Si);            //     recursively link subcell
	C->NUMBER += lastc->NUMBER;              //     add to number           
	if(dp>dpmax) dpmax=dp;                   //     update maximum depth    
      } else if(in_subtree(tc)) {                //   else if cell=subtree node 
	ts = find_soul(tc);                      //     find the subtree soul   
#ifdef USE_BODIES
	(*Si)->link(mybody(ts));                 //     link to body            
#else
	(*Si)->link(myelem(ts));                 //     or link to element      
#endif
	(*Si)->pos() = cofm(ts);                 //     set position from soul  
	(*Si)->copy_body_flag(ts);               //     copy body flag from soul
	if(C->SOULS) lasts->thread(*Si);         //     add to linked list      
	else          C->SOULS   = *Si;          //     of sub-tree sub-souls   
	lasts = (*Si)++;                         //     update last elem of list
	C->NUMBER++;                             //     increment number        
      }
    C->DEPTH = dpmax;                            // set depth                   
    if(C->CELLS)                                 // set first: if subcells      
      C->FIRST = C->CELLS->FIRST;                //   try first of first subcell
    if(C->SOULS &&                               // if subsouls                 
       (C->FIRST==0 || C->SOULS<C->FIRST))       //   try first subsoul         
      C->FIRST = C->SOULS;
    return dpmax + 1;                            // return our depth+1          
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  void subtree<ssoul,scell>::cut(fasttree::cell*tC0, fasttree::cell*tCN)
  {
    register indx nsun,nsuc;
    register uint ns=0u, nc=0u;
    register fasttree::cell *tCi;
    for(tCi=tCN-1; tCi>=tC0; tCi--) {            // loop tree cells backward
      tCi->unflag_subtree_flags();               //   reset subtree flags       
      nsun=0;                                    //   counter: subt subnodes    
      nsuc=0;                                    //   counter: subt subcells    
      LoopSubSouls(tCi,ts)                       //   loop subsouls of tcell   >
	if(in_subtree(ts)) nsun++;               //     count flagged for subt <
      ns += nsun;                                //   count subt souls          
      LoopSubCells(tCi,tc)                       //   loop subcells of tcell   >
	if(in_subtree(tc)) {                     //     IF tcell is subt node  >
	  ++nsun;                                //       count subt subnodes   
	  if(is_subtreecell(tc) ||               //       IF tcell is subt cell 
	     is_subtreezombie(tc) )              //       OR is subt zombie    >
	    ++nsuc;                              //         count subt subcells<
	}                                        //   < <                       
      if(nsun > 0) {                             //   IF tcell is subt node    >
	tCi->flag_for_subtree();                 //     flag it as such         
	if(nsun > 1) {                           //     IF tcell also subt cell>
	  tCi->flag_as_subtreecell();            //       flag it as such       
	  ++nc;                                  //       count subt cells      
	} else if(nsuc > 0)                      //     < ELSE tcell=subt zomb >
	  tCi->flag_as_subtreezombie();          //       so flag it as such   <
      }                                          //   <                         
    }                                            // <                           
    Ns  = ns;                                    // # subt souls                
    Nc  = nc;                                    // # subt cells                
    if(Ns==0 || Nc==0) {
      warning("subtree::cut(): empty subtree");
      return;
    }
    register ssoul *SI=S0=new ssoul[Ns];         // allocate subtree cells
    SN = S0+Ns;
    register scell *CI=C0=new scell[Nc];         // allocate subtree souls
    CN = C0+Nc;
    link_sub_tree(tC0,&CI,&SI);                  // Now: cut subtree out of tree
  }
  //============================================================================
#ifndef C_AND_F
  template<class scell>
  inline int number_of_cell(const scell *C, const scell*C0)
  { return C ? int(C-C0) : -99; }
  //----------------------------------------------------------------------------
  template<class ssoul>
  inline int number_of_soul(const ssoul *S, const ssoul*S0)
  { return S ? int(S-S0) : -9999; }
  //----------------------------------------------------------------------------
  template<class ssoul>
  inline void bsoul<ssoul>::dump_head(std::ostream&o)
  {
    o<<"     # flg  next          position";
  }
  //----------------------------------------------------------------------------
  template<class ssoul>
  inline void bsoul<ssoul>::dump(std::ostream&o, const ssoul*S0) const
  {
    o<<" "<<std::setw(5)<<number_of_soul<ssoul>((ssoul*)(this),S0)
     <<" "<<std::setw(3)<<flag(*this)
     <<" "<<std::setw(5)<<number_of_soul<ssoul>(NEXT,S0);
    for(register indx i=0; i<NDIM; i++)
      o<<" "<<std::setw(8)<<std::setprecision(4)<<POS[i];
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  inline void bcell<ssoul,scell>::dump_head(std::ostream&o)
  {
    o<<"     # flg  next cells souls first number de           position";
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell> inline
  void bcell<ssoul,scell>::dump(std::ostream&o,
				const ssoul*S0, const scell*C0) const
  {
    o<<" "<<std::setw(5)<<number_of_cell<scell>((scell*)(this), C0)
     <<" "<<std::setw(3)<<flag(*this)
     <<" "<<std::setw(5)<<number_of_cell<scell>(NEXT, C0)
     <<" "<<std::setw(5)<<number_of_cell<scell>(CELLS,C0)
     <<" "<<std::setw(5)<<number_of_soul<ssoul>(SOULS,S0)
     <<" "<<std::setw(5)<<number_of_soul<ssoul>(FIRST,S0)
     <<" "<<std::setw(6)<<NUMBER
     <<" "<<std::setw(2)<<DEPTH;
    for(register indx i=0; i<NDIM; i++)
      o<<" "<<std::setw(8)<<std::setprecision(4)<<POS[i];
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  void subtree<ssoul,scell>::dumpcells(std::ostream& o) const
  {
    scell::dump_head(o); o<<"\n";
    for(register scell *C=C0; C<CN; C++) { C->dump(o,S0,C0); o<<"\n"; }
    o.flush();
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  void subtree<ssoul,scell>::dumpcells(const char *file) const
  {
    std::ofstream out;
    if(open(out,file)) dumpcells(out);
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  void subtree<ssoul,scell>::dumpsouls(std::ostream& o) const
  {
    ssoul::dump_head(o); o<<"\n";
    for(register ssoul *S=S0; S<SN; S++) { S->dump(o,S0); o<<"\n"; }
    o.flush();
  }
  //----------------------------------------------------------------------------
  template<class ssoul, class scell>
  void subtree<ssoul,scell>::dumpsouls(const char *file) const
  {
    std::ofstream out;
    if(open(out,file)) dumpsouls(out);
  }
#endif // C_AND_F
}
//==============================================================================
#undef LoopSubCells
#undef LoopSubSouls
//==============================================================================
#endif // included_subt_h
