C------------------------------------------------------------------------------+
C                                                                              |
C  version 3.3.0                                                               |
C                                                                              |
C  treeF.f                                                                     |
C                                                                              |
C  header file for FORTRAN users                                               |
C  (C++ and C users, see files tree.h and treeC.h, respectively)               |
C                                                                              |
C  Copyright Walter Dehnen, 2000/2001                                          |
C  e-mail:   dehnen@mpia-hd.mpg.de                                             |
C  address:  Max-Planck Institut fr Astronomie,                               |
C            Knigstuhl 69117 Heidelberg, Germany                              |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C  FORTRAN routines implementing the tree code described by Dehnen (2000).     |
C  These FORTRAN routines call C routines declared in file treeC.h, which in   |
C  turn call the original C++ functions declared in file tree.h.               |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C  See file readme.tree for some guidelines on how to install, #include and    |
C  link this file and the code declared in it. You have the choice between     |
C  various options determining some properties of the code.                    |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C  CONTENTS                                                                    |
C  ========                                                                    |
C                                                                              |
C  0  Preliminaries                                                            |
C     0.1  Dimensionality                                                      |
C     0.2  Data type for I/O with the tree code                                |
C     0.3  FORTRAN specific problems                                           |
C  1  Initialisation and clearing up                                           |
C     1.1  Before using the code                                               |
C     1.2  After using the code                                                |
C  2  Meaning of the body flags                                                |
C  3  Force approximation and related                                          |
C     3.1  Generating a tree structure                                         |
C     3.2  Approximating the forces                                            |
C     3.3  Estimating mass- and number-density                                 |
C  4  Sticky-particle and SPH support                                          |
C     4.1  Sticky-particle support                                             |
C     4.2  SPH support                                                         |
C  5  Other features                                                           |
C  6  Known bugs and problems                                                  |
C     6.1  Test-bodies are not possible                                        |
C     6.2  Bodies at identical positions                                       |
C  References                                                                  |
C  A  Routines used for test purposes only                                     |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  0 PRELIMINARIES                                                             |
C  ===============                                                             |
C                                                                              |
C  0.1 Dimensionaliy                                                           |
C  -----------------                                                           |
C  The code allows for 2 or 3 (default) dimension, controlled by the macro     |
C  "TWODIMENSIONAL", which is set as compiler option, see file readme.tree.    |
C  Below, I'll explain the 3D versions. The only difference to the 2D versions |
C  is that arguments specifying the 3rd (=z) dimension of vectors are simply   |
C  omitted.                                                                    |
C                                                                              |
C                                                                              |
C  0.2 Data type for I/O with the tree code                                    |
C  ----------------------------------------                                    |
C  The tree code will take arrays of positions and accelerations. It then      |
C  copys their contents into its internal data types. The numerical            |
C  representation of the latter is either 32bit or 64bit, depending on the     |
C  macro "PRECISION" in file make.defs. The numerical representation of the    |
C  former, the arrays, can be chosen independently via the macro               |
C  "INPUTPRECISION" in the same file. In the comments below, I use             |
C  "INPUT_TYPE" for either REAL or DOUBLE PRECISION, depending on the value    |
C  of the makefile macro "INPUTPRECISION".                                     |
C                                                                              |
C                                                                              |
C  0.3 FORTRAN specific problems                                               |
C  -----------------------------                                               |
C  Note that with FORTRAN you won't recognize a wrong argument type at all:    |
C  the code will compile, run, and give wrong result, which are not            |
C  necessarily recognisable as such. As a consequence,                         |
C                                                                              |
C      +----------------------------------------------------------------+      |
C      |  YOU MUST BE EXTERMELY CAREFUL TO GET THE ARGUMENT TYPES RIGHT |      |
C      +----------------------------------------------------------------+      |
C                                                                              |
C  This is a generic FORTRAN problem and not a problem of the tree code. The   |
C  best and only secure way to avoid it, is to abolish the use of FORTRAN.     |
C                                                                              |
C  You have been warned!                                                       |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  1 INITIALISATION AND CLEARING UP                                            |
C  ================================                                            |
C                                                                              |
C  1.1 BEFORE USING THE CODE                                                   |
C  -------------------------                                                   |
C  In order to use the code, one has first to initialize it. This is done      |
C  using the following routine, which actually does no computation at all.     |
C                                                                              |
      EXTERNAL TREEF_INIT                !  initialize tree code               |
C                                                                              |
C  The syntax is a follows:                                                    |
C                                                                              |
C     INTEGER    N                                                             |
C     PARAMETER (N=10000)          number of bodies                            |
C     INTEGER    FL(N)             array with flags, see below                 |
C     INPUT_TYPE M(N)              array with masses                           |
C     INPUT_TYPE X(N),Y(N),Z(N)    arrays with position components             |
C     INPUT_TYPE E(N)              array with indiv softening lengths          |
C     INPUT_TYPE AX(N),AY(N),AZ(N) array for acceleration components           |
C     INPUT_TYPE P(N)              array for potentials                        |
C     INPUT_TYPE RH(N)             array for mass densities                    |
C     INPUT_TYPE EP                global softening length, see below          |
C     INPUT_TYPE TH                opening angle, see below                    |
C     INTEGER    K                 type of softening kernel                    |
C     INTEGER    D                 maximum tree depth, see below               |
C                                                                              |
C     CALL TREEF_INIT(FL,M,X,Y,Z,E,AX,AY,AZ,P,RH, N,EP,TH,K,D)                 |
C                                                                              |
C  The first 11 arguments specify the sink and source properties of the        |
C  bodies. Each body has the sink properties: position (x,y,z), mass,          |
C  softening length (for the case of individual softening lengths), and flag   |
C  (see section 2 below). The source properties are: acceleration(ax,ay,az),   |
C  potential and mass-density. If fixed softening is used, a NULL pointer may  |
C  be used instead of an array holding eps_i, i.e.                             |
C                                                                              |
C     CALL TREEF_INIT(FL,M,X,Y,Z,%VAL(0),AX,AY,AZ,P,RH, N,EP,TH,K,D)           |
C                                                                              |
C  The same applies to the arrays for density and potential: if a NULL         |
C  pointer is given, they will never be updated (but possibly computed).       |
C                                                                              |
C  The last 5 arguments are:                                                   |
C                                                                              |
C  N:                  size of arrays == number of bodies.                     |
C  EPS:   if(EPS>=0):  globally fixed softening length                         |
C         if(EPS< 0):  use individual softening lengths provided in array      |
C  THETA: if(THETA>0): theta = theta(M) with theta_min = THETA (Dehnen 2001b)  |
C         if(THETA<0): theta = THETA = const                                   |
C         RECOMMENDED: THETA = 0.4                                             |
C  KERN:  0,1,2,3:     Ferrers sphere of this index (Dehnen 2001a)             |
C         10,11,12,13  compensating kernel (Dehnen 2001a) of index (KERN-10)   |
C         20           Plummer softening kernel                                |
C         21           rho = a /[r^2+e^2]^(7/2)                                |
C  DMAX   maximum allowed tree depth (see 5.2 below)                           |
C         RECOMMENDED: DMAX = 50                                               |
C                                                                              |
C  If TREEF_INIT() is called more than once, only the last initialisation      |
C  applies, the older ones will be deleted. That is, you can have only one     |
C  tree at a time -- use the C++ version if you need more.                     |
C                                                                              |
C  With the functions                                                          |
C                                                                              |
      EXTERNAL TREEF_RESETSOFTENING      !  reset softening parameters         |
C     syntax:                                                                  |
C     INTEGER    K                         softening kernel (as above)         |
C     INPUT_TYPE EP                        global softening length             |
C     CALL TREEF_RESETSOFTENING(EP,K)      set softening properties            |
C                                                                              |
C  and                                                                         |
C                                                                              |
      EXTERNAL TREEF_RESETOPENING        !  reset opening criterion            |
C     syntax:                                                                  |
C     INPUT_TYPE TH                        opening angle                       |
C     CALL TREEF_RESETOPENING(TH)          set opening properties              |
C                                                                              |
C  it is possible to change, after initialisation, the softening length and    |
C  kernel as well as the opening criterion.                                    |
C                                                                              |
C                                                                              |
C                                                                              |
C  1.2 AFTER USING THE CODE                                                    |
C  ------------------------                                                    |
C  After using the code, one should delete allocated memory by calling         |
C                                                                              |
      EXTERNAL TREEF_CLEARUP             !  de-allocated memory                |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  2 MEANING OF THE BODY FLAGS                                                 |
C  ===========================                                                 |
C                                                                              |
C  The body flags are integers with the following meaning:                     |
C                                                                              |
C   bit  value     meaning                                                     |
C   --------------------------------------------------------------             |
C     1      1     this body is a sink, i.e. wants update                      |
C     2      2     don't load this body into the tree, ignore it               |
C     3      4     this body is a SPH particle                                 |
C     4      8     this body is a sticky particle                              |
C   i>4   2^(i-1)  not used                                                    |
C                                                                              |
C                                                                              |
C  The default, flag=0, represents a plain body that is not sink, but still    |
C  source of gravity.                                                          |
C  The flag is obtained by setting the bits or, equivalently, adding the       |
C  values. The flags are not used by TreeC_initialize() and will unfold their  |
C  effect only when The functions in section 3 and 4 below are called.         |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  3 FORCE APPROXIMATION AND RELATED                                           |
C  =================================                                           |
C                                                                              |
C  3.1 Generating a tree structure                                             |
C  -------------------------------                                             |
C  In order to establish a hierarchical tree structure from all bodies that    |
C  are not flagged to be ignored, and subsequently pre-compute the centers     |
C  of mass and sizes of the cells, use                                         |
C                                                                              |
      EXTERNAL TREEF_GROW                !  grow new tree structure            |
C     syntax:                                                                  |
C     INTEGER NCRIT                                                            |
C     CALL TREEF_GROW(NCRIT)                                                   |
C                                                                              |
C  or                                                                          |
C                                                                              |
      EXTERNAL TREEF_GROW_CENTERED       !  grow new tree structure            |
C     syntax:                                                                  |
C     INTEGER NCRIT                                                            |
C     INTEGER CEN(3)                                                           |
C     CALL TREEF_GROW_CENTERD(NCRIT,CEN)                                       |
C                                                                              |
C  In the second case, the root cell of the tree is forced to be centered on   |
C  the position given as argument (to avoid round-off errors, only integer-    |
C  valued positions are allowed).                                              |
C                                                                              |
C  Both create a tree from scratch. Cells containing NCRIT or less bodies      |
C  will NOT be splitted. Experiments showed that for a full force calculation  |
C  NCRIT = 6-8 is about optimal and results in a few % decrease in CPU time    |
C  consumption (compared to NCRIT=1) and about 20% reduction in memory.        |
C                                                                              |
C                                                                              |
C  You may instead also re-use an old tree structure and only re-compute the   |
C  centers of mass and sizes of the cells (since the bodies have moved), by    |
C  using                                                                       |
C                                                                              |
      EXTERNAL TREEF_REUSE               !  re-use olf tree structure          |
C     syntax:                                                                  |
C     CALL TREEF_REUSE()                                                       |
C                                                                              |
C  In this case, the logical linkage between cells and bodies is preserved     |
C  and no new tree structure is established. However, the code accounts for    |
C  the fact that the bodies might have moved. If the bodies have moved a lot,  |
C  the sizes of the cells will be much larger than the physical size of the    |
C  associated boxes, and the tree traversal will be very inefficient.          |
C  However, if the bodies have moved only little, the force compuation is      |
C  hardly slowed down and re-using an old tree saves the costs for             |
C  establishing a new one. Note that reuse() does not allow for any change in  |
C  the flags indicating whether or not a body shall be ignored: changing this  |
C  flag will have an effect only at the next call of grow().                   |
C                                                                              |
C                      +---------------------------------+                     |
C                      | USE THIS OPTION VERY CAREFULLY! |                     |
C                      +---------------------------------+                     |
C                                                                              |
C  You have been warned!                                                       |
C                                                                              |
C                                                                              |
C  3.2 Approximating the forces                                                |
C  ----------------------------                                                |
C  Once a tree structure is established, you can compute the forces by         |
C                                                                              |
      EXTERNAL TREEF_APPROX_GRAV         !  approximate, e.g., forces          |
C     syntax:                                                                  |
C     TREEF_APPROX_GRAV()                                                      |
C                                                                              |
C  after a call to TREEF_GROW, TREEF_GROW_CENTERED, or TREEF_REUSE.            |
C                                                                              |
C                                                                              |
C  3.3 Estimating mass- and number-density                                     |
C  ---------------------------------------                                     |
C  There is also the possibility to obtain a rough estimate of the mass- or    |
C  number density of bodies in the neighbourhood of every body flagged as      |
C  sink, via                                                                   |
C                                                                              |
      EXTERNAL TREEF_ESTIMATE_RHO        !  estimate the mass density          |
C     syntax:                                                                  |
C     INTEGER NX                                                               |
C     TREEF_ESTIMATE_RHO(NX)                                                   |
C                                                                              |
C  and                                                                         |
C                                                                              |
      EXTERNAL TREEF_ESTIMATE_N          !  estimate the number density        |
C     syntax:                                                                  |
C     INTEGER NX                                                               |
C     TREEF_ESTIMATE_N(NX)                                                     |
C                                                                              |
C  These estimates are simply the mean density within the smallest cells with  |
C  more then NX (1st arg) bodies. Note that when using test bodies (bodies     |
C  with zero or tiny mass), this guess for the mass density can have terrible  |
C  errors. Moreover, when the tree has not been grow()n but simply reuse()ed,  |
C  these estimate will not change. Be careful using these functions.           |
C  YOU HAVE BEEN WARNED.                                                       |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  4 STICKY-PARTICLE AND SPH SUPPORT                                           |
C  =================================                                           |
C                                                                              |
C  Once a tree structure is established (i.e. after a call to TREEF_GROW,      |
C  TREE_GROW_CENTERED, or TREEF_REUSE), you can also use it to create          |
C  interaction lists for SPH and sticky particles.                             |
C                                                                              |
C                                                                              |
C  4.1. Sticky-particle support                                                |
C  ----------------------------                                                |
C  This is provided by the routine                                             |
C                                                                              |
      EXTERNAL TREEF_STICKY              !  sticky particle support            |
C     syntax:                                                                  |
C     INTEGER    N                                                             |
C     PARAMETER (N=10000)             number of bodies                         |
C     INPUT_TYPE S(N)                 array with body sizes                    |
C     INPUT_TYPE VX(N),VY(N),VZ(N)    array for velocity components            |
C     INTEGER    NI                                                            |
C     PARAMETER (NI=1000)             physical size or interaction list        |
C     INTEGER    I1(NI)               array for indices: 1st of pair           |
C     INTEGER    I2(NI)               array for indices: 1st2nd of pair        |
C     INTEGER    NA                   actual size or interaction list          |
C     INPUT_TYPE TAU                  time step tau                            |
C     CALL TREEF_STICKY(TAU,I1,I2,NI,NA,VX,VY,VZ,S)                            |
C                                                                              |
C  which fills the interaction list (2nd/3rd arg) with all pairs {i,j} of      |
C  indices which satisfy the following three conditions.                       |
C                                                                              |
C     (1) both flags indicate sticky particles (see section 2),                |
C     (2) at least one flag indicates sink                                     |
C     (3) | (x_i+t|v_i)-(x_j+t|v_j) | < size_i+size_j  for some t in [0,tau]   |
C                                                                              |
C  In case of overflow, i.e. if the number of such pairs exceeds the size      |
C  (4th arg) of the list (2nd/3rd arg), the routine issues a warning to        |
C  stderr, but does not truncate the search, rather it no longer copies pairs  |
C  into the interaction list. In this case, the value returned for the actual  |
C  size (5th arg) exceeds the maximum size (4th arg), but reflects the actual  |
C  number of interactions found.                                               |
C                                                                              |
C  The arrays for velocities and sizes are assumed to be of the same size as   |
C  those used for positions (given to TREEF_INITIALIZE). However, the code     |
C  accesses velocities only for those bodies which are flagged as sticky, ie.  |
C  if only the first, say, 1000 of 10000 bodies are sticky, it is sufficient   |
C  if the size of the arrays with velocities and body sizes are 1000.          |
C                                                                              |
C  See file src/FORTRAN/TestPairF.F for an example application.                |
C                                                                              |
C                                                                              |
C  4.2 SPH particle support                                                    |
C  ------------------------                                                    |
C  This is provided by the routine                                             |
C                                                                              |
      EXTERNAL TREEF_SPH                 !  SPH particle support               |
C     syntax:                                                                  |
C     INTEGER    N                                                             |
C     PARAMETER (N=10000)             number of bodies                         |
C     INPUT_TYPE S(N)                 array with body sizes                    |
C     INTEGER    NI                                                            |
C     PARAMETER (NI=1000)             physical size or interaction list        |
C     INTEGER    I1(NI)               array for indices: 1st of pair           |
C     INTEGER    I2(NI)               array for indices: 1st2nd of pair        |
C     INTEGER    NA                   actual size or interaction list          |
C     CALL TREEF_STICKY(I1,I2,NI,NA,S)                                         |
C                                                                              |
C  which fills the interaction list (1st/2nd arg) with all pairs {i,j} of      |
C  indices which satisfy the following three conditions.                       |
C                                                                              |
C     (1) both are flags indicate SPH particles (see section 2),               |
C     (2) at least one flag indicates sink                                     |
C     (3) | x_i - x_j | < max(size_i,size_j)                                   |
C                                                                              |
C  In case of overflow, i.e. if the number of such pairs found exceeds the     |
C  size (3rd arg) of the list (1st/2nd arg), the routine issues a warning to   |
C  stderr, but does not truncate the search, rather it ceases copying pairs    |
C  into the interaction list. In this case, the value returned for the actual  |
C  size (4th arg) exceeds the maximum size (3rd arg), but reflects the actual  |
C  number of interactions found.                                               |
C                                                                              |
C  See file src/FORTRAN/TestPairF.F for an example application.                |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  5 OTHER FEATURES                                                            |
C  ================                                                            |
C                                                                              |
C  There are many further routines by which the user can obtain information    |
C  about the behaviour of the code.                                            |
C                                                                              |
      REAL     TREEF_ROOT_CENTER         !  gives Ith comp. of root center     |
      EXTERNAL TREEF_ROOT_CENTER         !                                     |
C     syntax:                                                                  |
C     INTEGER I                                                                |
C     INPUT_TYPE  C = TREEF_ROOT_CENTER(I)                                     |
C                                                                              |
      REAL     TREEF_ROOT_RADIUS         !  gives radius of root cell          |
      EXTERNAL TREEF_ROOT_RADIUS         !                                     |
C     syntax:                                                                  |
C     INPUT_TYPE    R = TREEF_ROOT_RADIUS()                                    |
C                                                                              |
      REAL     TREEF_CURRENT_THETA       !  gives theta(M)                     |
      EXTERNAL TREEF_CURRENT_THETA       !                                     |
C     syntax:                                                                  |
C     INPUT_TYPE    M,TH=TREEF_CURRENT_THETA(M)                                |
C                                                                              |
      REAL     TREEF_CURRENT_EPS         !  gives global eps                   |
      EXTERNAL TREEF_CURRENT_EPS         !                                     |
C     syntax:                                                                  |
C     INPUT_TYPE    EP=TREEF_CURRENT_EPS()                                     |
C                                                                              |
      INTEGER  TREEF_CURRENT_KERNEL      !  gives kernel info                  |
      EXTERNAL TREEF_CURRENT_KERNEL      !                                     |
C     syntax:                                                                  |
C     INTEGER K = TREEF_CURRENT_KERNEL()                                       |
C                                                                              |
      INTEGER  TREEF_SOFTENING           !  gives 0/1 : fixed/individual       |
      EXTERNAL TREEF_SOFTENING           !                                     |
C     syntax:                                                                  |
C     INTEGER S = TREEF_SOFTENING()                                            |
C                                                                              |
      INTEGER  TREEF_NO_CELLS            !  gives number of cells              |
      EXTERNAL TREEF_NO_CELLS            !                                     |
C     syntax:                                                                  |
C     INTEGER NCELLS = TREEF_NO_CELLS()                                        |
C                                                                              |
      INTEGER  TREEF_DEPTH               !  gives tree depth                   |
      EXTERNAL TREEF_DEPTH               !                                     |
C     syntax:                                                                  |
C     INTEGER DEPTH = TREEF_DEPTH()                                            |
C                                                                              |
      EXTERNAL TREEF_STATS               !  writes some stats to stdout        |
C     syntax:                                                                  |
C     CALL TREEF_STATS()                                                       |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  6 KNOWN BUGS AND PROBLEMS                                                   |
C  =========================                                                   |
C                                                                              |
C  6.1 Test-bodies are not possible                                            |
C  --------------------------------                                            |
C  A body that is loaded into the tree but has zero mass, will not acquire     |
C  any acceleration. This is because the code computes first the force = mass  |
C  times acceleration (it is symmetric and hence better suited for             |
C  computation of mutual interactions) and then divides by the mass to obtain  |
C  the acceleration. There are two possible solutions to this problem:         |
C  1. Test bodies may be allowed by setting the appopriate non-default         |
C     compiler option. Note, however, that this results in less efficient      |
C     code also for the case that no body has zero mass.                       |
C  2. Set the mass of potential test bodies much smaller than the masses of    |
C     source bodies. However, this must be done such that the gravity of test  |
C     bodies is everywhere neglible compared to that of the source bodies.     |
C     Note, however, that this work-around is wasteful: it computes the        |
C     forces generated by the test bodies.                                     |
C  (You have been warned!)                                                     |
C                                                                              |
C                                                                              |
C  6.2 Bodies at identical positions                                           |
C  ---------------------------------                                           |
C  The code cannot cope with more than Ncrit bodies at the same position       |
C  (within the floating point accuracy). This situation will lead to an        |
C  infinitely deep tree, i.e. the maximum allowed tree depth will be exceeded  |
C  and the code aborts with an error message.                                  |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  REFERENCES                                                                  |
C  ==========                                                                  |
C                                                                              |
C  Dehnen W., 2000, ApJ, 536, L39                                              |
C  Dehnen W., 2001a, MNRAS, in press; astro-ph/0011568                         |
C  Dehnen W., 2001b, submitted to J. Comp. Phys.                               |
C                                                                              |
C------------------------------------------------------------------------------+
C                                                                              |
C                                                                              |
C  A ROUTINES USED FOR TEST PURPOSES ONLY                                      |
C  ======================================                                      |
C                                                                              |
C  Routines not used by the tree, but by the test program TestTreeF.f          |
C                                                                              |
C  YOU MAY COMMENT THEM OUT IF YOU NEVER USE THEM.                             |
      INTEGER           CLOCK
      DOUBLE PRECISION  DRAND48,CLOCKS_PER_SECOND
      EXTERNAL          CLOCK,DRAND48,SRAND48,CLOCKS_PER_SECOND
C------------------------------------------------------------------------------+


