//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// written by Walter Dehnen, 2000/2001                                         |
// e-mail:   dehnen@mpia-hd.mpg.de                                             |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <nbdy.h>
extern "C" {
#include <fstream>
#include <iomanip>
}

const double Pi   = 3.14159265358979323846264338328;

using std::ostream;
using std::endl;
using std::ios;
using std::setw;
using std::setprecision;
using namespace fasttree;
////////////////////////////////////////////////////////////////////////////////
// class fasstree::nbody                                                        
////////////////////////////////////////////////////////////////////////////////
inline void nbody::reset_cpus() const
{
  CPU_PREDICT = zero;
  CPU_BUILD   = zero;
  CPU_GRAV    = zero;
  CPU_PEX     = zero;
  CPU_DENS    = zero;
  CPU_EPSI    = zero;
  CPU_STEP    = zero;
  N_FORCES    = 0u;
}
//------------------------------------------------------------------------------
inline void nbody::record_cpu(clock_t& c0, real& CPU) const
{
  register clock_t c1 = clock();
  CPU += (c1-c0)/real(CLOCKS_PER_SEC);
  c0 = c1;
}
//------------------------------------------------------------------------------
inline void nbody::set_allow_tree()
{
  const body* BN = bodies()+No_bodies();
  for(register body* Bi=bodies(); Bi<BN; Bi++) Bi->allow_tree_usage();
}
//------------------------------------------------------------------------------
nbody::nbody(                                    // constructor: initialize     
	           body           *b,            // I:   array with bodies      
             const unsigned        n,            // I:   N                      
	     const soft_type       sf,           // I:   softening              
	     const unsigned        ns,           // I:   Nsoft                  
	     const unsigned        nr,	         // I:   Nref                   
	     const real            e,            // I:   eps                    
	     const real            fac,          // I:   f                      
	     const real            tq,           // [I]: tol^2                  
	     const real            th,           // [I]: theta                  
	     const unsigned        nc,           // [I]: Ncrit                  
	     const kern_type       ker,          // [I]: kernel                 
	     const potext         *p)            // [I]: P_ex                   
  : tree(b,n, (ns==0 && e>=zero)? e : -abs(e), th, ker),
    BlockStep(),
    MYSCHEME((fac>zero)? rhos : dacc),
    PEX(p),
    READY(false),
    NSOFT(ns), NREF(nr), NCRIT(max(1u,nc)), EMAX(e),
    FAC(abs(fac)), TQ(tq), DIAG(false)
{
  reset_softening(sf,ns,e,ker);
  BlockStep::reset_fac(fac);
  set_allow_tree();
  if (fac==zero) error("time step-controlling factor = 0");
}
//------------------------------------------------------------------------------
void nbody::reset_stepping(const real fac)
{
  MYSCHEME = (fac > 0)? rhos : dacc;
  FAC = fac;
  BlockStep::reset_fac(fac);
}
//------------------------------------------------------------------------------
inline void nbody::set_pposs()
{
  register clock_t cpu = clock();
  const    body* BN = bodies()+No_bodies();
  set_predict();
  for(register body* Bi=bodies(); Bi<BN; Bi++) predict_pos(Bi);
  record_cpu(cpu,CPU_PREDICT);
}
//------------------------------------------------------------------------------
inline nbody::tree_use nbody::set_tree(const tree_use use_tree, vect *x)
{
  register clock_t cpu = clock();
  switch(use_tree) {
  case re_used:  
    tree::reuse();
    record_cpu(cpu,CPU_BUILD);
    return re_used;
  default:
    tree::grow(NCRIT,x);
    record_cpu(cpu,CPU_BUILD);
    return built;
  }
}
//------------------------------------------------------------------------------
void nbody::reset_softening(const soft_type sf,
			    const uint ns, const real e,
			    const kern_type ker)
{
  SOFTENING = sf;
  NSOFT     = ns;
  switch(SOFTENING) {
  case individual_adaptive:
  case individual_fixed:
    EMAX = e;
    tree::reset_softening(-one,ker);
    break;
  default:
    EMAX = zero;
    tree::reset_softening(e,ker);
  }
}
//------------------------------------------------------------------------------
void nbody::set_forces()
{
  register clock_t cpu = clock();
  const    body*  BN = bodies()+No_bodies();
  register body*  Bi;

  tree::approximate_gravity();                        //  compute sink forces
  record_cpu(cpu,CPU_GRAV);                           //  record CPU
  if(PEX) {                                           //  if external potential
    register vect A;
    for(Bi=bodies(); Bi<BN; Bi++) if(is_sink(Bi)) {   //    loop body sinks
      Bi->pot()+=PEX->pot_f(A,ppos(Bi),time());       //      add ext potential
      Bi->acc()+=A;                                   //      add ext force
    }
    record_cpu(cpu,CPU_PEX);                          //  record CPU
  }
}
//------------------------------------------------------------------------------
inline void nbody::setup_eps(body* Bi) const
{
  const double Pi = 3.14159265358979323846264338328;
#if NDIM==3
  const real fac = 0.75*NSOFT/Pi;
  Bi->eps() = min(EMAX,POW(fac/rho(Bi),third));
#else
  const real fac = NSOFT/Pi;
  Bi->eps() = min(EMAX,SQRT(fac/rho(Bi)));
#endif	
}
//------------------------------------------------------------------------------
void nbody::set_softening_lengths()
{
  const    body* BN = bodies()+No_bodies();
  register body* Bi;
  register clock_t  cpu = clock();                    // cpu time
  for(Bi=bodies(); Bi<BN; Bi++) Bi->flag_as_sink();   // flag every body
  estimate_n(NSOFT);                                  // estimate n_i
  for(Bi=bodies(); Bi<BN; Bi++) setup_eps(Bi);        // set eps_i
  record_cpu(cpu,CPU_EPSI);                           // record CPU
}
//==============================================================================
// routines to use with dacc scheme
//------------------------------------------------------------------------------
inline bool nbody::step_too_long(const body *B) const
{
  return may_go_shorter(B) && DaccScheme::step_too_long(this,B);
}
//------------------------------------------------------------------------------
inline bool nbody::may_go_longer(const body *B) const
{
  return  level(B)>0 && B->may_go_longer();
}
//------------------------------------------------------------------------------
inline BlockStep::too_short nbody::step_too_short(const body *B) const
{
  if(may_go_longer(B)) return DaccScheme::step_too_short(this,B);
  else                 return allow_short;
}
//------------------------------------------------------------------------------
inline fasttree::uint nbody::set_sink_flags(const indx low)
{
  const    body* BN = bodies()+No_bodies();
  register body* Bi;
  register uint  nf=0;
  for(Bi=bodies(); Bi<BN; Bi++) if(is_in_tree(Bi))    // loop bodies in tree
    if(low == 0 || level(Bi) >= low                   //   if level >= low
       || level(Bi)+1==low && may_go_shorter(Bi)) {   //      or may go shorter
         Bi->flag_as_sink();                          //     flag as sink
	 nf++;
    }
    else Bi->unflag_sink ();                          //   else    : no sink
  return nf;
}
//------------------------------------------------------------------------------
nbody::tree_use nbody::smallest_step_dacc(            // R: tree usage
                          const indx low,             // I: lowest level moving
                          const tree_use use_tree,    // I: how to construct tree
                          vect *x)                    // [I]: center of root
{
  if(low && !No_shorter_level(low-1)) {               // if nobody moves:
    clock_on();                                       //  increase actual clock
    move_times(low);                                  //  increase steps' clocks
    return un_used;                                   //  and we are DONE
  }
  const    body* BN = bodies()+No_bodies();
  register body* Bi;
  register tree_use use;
  clock_on();                                         // increase actual clock
  set_pposs();                                        // predict positions
  use = set_tree(use_tree,x);                         // initialize tree
  N_FORCES += set_sink_flags(low);                    // set flags
  set_forces();                                       // set accelerations
  for(Bi=bodies(); Bi<BN; Bi++)                       // loop bodies
    if(level(Bi) > low)                               // l>low
      switch(step_too_short(Bi)) {
      case go_longer:    move_proper(Bi); goto_longer(Bi);      break;
      case forbid_short: move_proper(Bi); Bi->forbid_shorter(); break;
      default:           move_proper(Bi); Bi->allow_shorter();  break;
      }
    else if(level(Bi) == low)                         // l=low
      switch(step_too_short(Bi)) {
      case go_longer:                     goto_longer(Bi);      break;
      case forbid_short: move_proper(Bi); Bi->forbid_shorter(); break;
      default:           move_proper(Bi); Bi->allow_shorter();  break;
      }
    else if(level(Bi)+1==low && step_too_long(Bi)) {  // l=low-1 && step too long
                         move_proper(Bi); goto_shorter(Bi);
    }
  move_times(low);                                    // increase steps' clocks
  return use;
}
//==============================================================================
// routines used with the rhos scheme
//------------------------------------------------------------------------------
inline bool nbody::test_for_longer(const indx low, const body*B) const
{
  return low>0 && may_go_longer(B) && level(B) == low;
}
//------------------------------------------------------------------------------
inline bool nbody::test_for_shorter(const indx low, const body*B) const
{
  return level(B)+1 == low;
}
//------------------------------------------------------------------------------
inline void nbody::estimate_mass_density()
{
  register clock_t  cpu = clock();                    // cpu time
  tree::estimate_rho(NREF);                           // estimate rho in tree
  if(PEX) {                                           // if external potential
    const    body* BN = bodies()+No_bodies();
    register body* Bi;
    for(Bi=bodies(); Bi<BN; Bi++) if(is_sink(Bi))     //   loop body sinks
      Bi->rho()+=PEX->rho(ppos(Bi),time());           //     add ext density
  }
  record_cpu(cpu,CPU_DENS);                           // record CPU
}
//------------------------------------------------------------------------------
void nbody::estimate_mass_densities()
{
  const body* BN = bodies()+No_bodies();
  for(register body*Bi=bodies(); Bi<BN; Bi++) Bi->flag_as_sink();
  estimate_mass_density();
}
//------------------------------------------------------------------------------
void nbody::estimate_surf_densities()
{
  const body* BN = bodies()+No_bodies();
  for(register body*Bi=bodies(); Bi<BN; Bi++) Bi->flag_as_sink();
  tree::estimate_sd(NREF);
}
//------------------------------------------------------------------------------
nbody::tree_use nbody::smallest_step_rhos(            // R: tree usage
                          const indx low,             // I: lowest level moving
                          const tree_use use_tree,    // I: how to construct tree
                          vect *x)                    // [I]: center of root
{
  if(low && !No_shorter_level(low-1)) {               // if nobody moves:
    clock_on();                                       //   increase actual clock
    move_times(low);                                  //   increase steps' clocks
    return un_used;                                   //   and we are DONE
  }
  const    body* BN = bodies()+No_bodies();
  register body* Bi;
  register uint  nf=0u;
  register tree_use use;
  clock_on();                                         // increase actual clock
  set_pposs();                                        // predict positions
  use = set_tree(use_tree,x);                         // initialize tree
  if(SOFTENING==individual_adaptive)                  // if individual adaptive:
    set_softening_lengths();                          //   then set eps_i
  if(low>0) {                                         // low>0: check time steps
    for(Bi=bodies(); Bi<BN; Bi++)                     //   loop bodies in tree
      if(test_for_longer (low,Bi) ||                  //     if may go slower
	 test_for_shorter(low,Bi) )                   //     OR faster
	   Bi->flag_as_sink();                        //       need guessed rho
      else Bi->unflag_sink ();                        //     else unflag
    estimate_mass_density();                          //   guess rho
    for(Bi=bodies(); Bi<BN; Bi++) {                   //   loop bodies in tree
      if     (test_for_longer (low,Bi) &&             //     if may go slower
	      RhoScheme::step_too_short(this,Bi) )    //       AND too fast
	goto_longer(Bi);                              //       change to l-1
      else if(test_for_shorter(low,Bi) &&             //     else if may go faster
	      RhoScheme::step_too_long(this,Bi) )     //       AND too slow
	goto_shorter(Bi);                             //       change to l+1
      if(level(Bi) >= low) {                          //     if l >= low:
	Bi->flag_as_sink();                           //       we need acc
	nf ++;                                        //       we count 
      } else                                          //     else
	Bi->unflag_sink ();                           //       unflag
    }
  } else {                                            // low=0: move all
    for(Bi=bodies(); Bi<BN; Bi++) Bi->flag_as_sink(); //   loop bodies in tree
    nf = No_bodies();
  }
  if(No_shorter_level(low)) {                         // if anybody moves:
    N_FORCES += nf;                                   //   # acc computations
    set_forces();                                     //   get accelerations
    for(Bi=bodies(); Bi<BN; Bi++) if(is_in_tree(Bi)){ //   loop bodies in tree
      if(level(Bi) >= low) {                          //     forall l >= low:
	move_proper(Bi);                              //       move
	Bi->allow_longer();                           //       allow change to l-1
      }
    }
  }
  move_times(low);                                    // increase steps' clocks
  return use;
}
//==============================================================================
// general routines
//==============================================================================
void nbody::long_step(const unsigned Nr, vect* x)
{
  if(!READY) error("nbody::long_step(): need to prepare() first");
  register clock_t  cpu = clock();                    // cpu time
  register unsigned ir  = 0u;                         // count tree::re_use()
  const    unsigned nn  = 1 << shortest_step();       // number of small steps
  register unsigned tn;                               // index of small step
  reset_cpus();                                       // reset CPU time counters
  switch(MYSCHEME) {
  case rhos:
    for(tn=1; tn<nn; tn++)                            // loop over all but last
      switch(smallest_step_rhos(longest_moving(tn),   //  smallest steps
			  (ir<Nr)?re_used:built, x)){ //  how was the tree made?
      case re_used:                                   //   reused:
	ir++; record_cpu(cpu,CPU_STEP); break;        //     count
      case built:                                     //   built:  
	ir=0u;record_cpu(cpu,CPU_STEP); break;        //     reset counter
      default: case un_used: break;
      }
    smallest_step_rhos(0,built,x);                    // last step: build tree
    break;
  default:
    for(tn=1; tn<nn; tn++)                            // loop over all but last
      switch(smallest_step_dacc(longest_moving(tn),   //  smallest steps
			   (ir<Nr)?re_used:built, x)){//  how was the tree made?
      case re_used:                                   //   reused:
	ir++; record_cpu(cpu,CPU_STEP); break;        //     count
      case built:                                     //   built:  
	ir=0u;record_cpu(cpu,CPU_STEP); break;        //     reset counter
      default: case un_used: break;
      }
    smallest_step_dacc(0,built,x);                    // last step: build tree
  }
  DIAG = false;                                       // diagnose out-of-date
  record_cpu(cpu,CPU_STEP);                           // record CPU time
  CPU_TOTAL += CPU_STEP;                              // record accumulated time
}
//------------------------------------------------------------------------------
void nbody::prepare(const real  t0,                   // I:   initial t
		    const int   h0,                   // I:   h0
		    const indx  nt,                   // I:   Nstps
		    const bool  need_init_levs,       // [I]: initialize levels?
		    const bool  need_init_epsi,       // [I]: initialize eps_i?
		          vect *x)                    // [I]: center of root
{
  const    body* BN = bodies()+No_bodies();
  register body* Bi;
  register clock_t cpu = clock();
  TINI = t0;                                          // remember t0
  reset_cpus();                                       // reset CPU counters
  BlockStep::reset_steps(h0,nt,t0);                   // set time steps
  for(Bi=bodies(); Bi<BN; Bi++) Bi->ppos()=tpos(Bi);  // set x=true position
  set_tree(built,x);                                  // initialize tree
  if(SOFTENING != global_fixed &&                     // if individual softening
     need_init_epsi) set_softening_lengths();         //   set eps_i
  for(Bi=bodies(); Bi<BN; Bi++) Bi->flag_as_sink();   // flag every body as sink
  set_forces();                                       // get accelerations
  for(Bi=bodies(); Bi<BN; Bi++) Bi->oacc()=acc(Bi);   // store accelerations
  diagnose();                                         // energy et al. at t=t0
  DIAG = true;
  BlockStep::reset_fac(FAC,TQ);                       // reset factors
  if(nt==1)                                           // only one time-step
    for(Bi=bodies(); Bi<BN; Bi++) add(Bi->level()=0); //   all bodies in level 0
  else {                                              // otherwise: work
    register bool use_old_levs = !need_init_levs;     //  try to use old levels:
    for(Bi=bodies(); Bi<BN && use_old_levs; Bi++)     //   loop bodies
      if(level(Bi)<nt) BlockStep::add(level(Bi));     //     level okay -> count
      else             use_old_levs = false;          //     otherwise -> stop AND
    if(!use_old_levs) {                               //  assign levels:
      BlockStep::reset_counts();                      //   reset counts
      switch(MYSCHEME) {                              //   which scheme ?
      case rhos:                                      //   rho scheme:
	estimate_mass_density();                      //     estimate mass density
	for(Bi=bodies(); Bi<BN; Bi++) {               //     loop bodies
	  RhoScheme::assign_to_level(this,Bi);        //       assign proper level
	  Bi->allow_longer();                         //       allow longer step
	}
	break;
      default:                                        //   delta(acc) scheme:
	clock_on();                                   //     advance t by tau_min
	set_pposs();                                  //     predict positions
	clock_back();                                 //     set t back 
 	set_tree(re_used);                            //     initialize tree
	set_forces();                                 //     get accelerations
	for(Bi=bodies(); Bi<BN; Bi++) {               //     loop bodies
	  DaccScheme::assign_to_level(this,Bi);       //       assign proper level
	  Bi->allow_longer();                         //       allow longer step
	}
      }
    }
  }
  READY     = true;
  record_cpu(cpu,CPU_STEP);                           // record total CPU
  CPU_TOTAL = CPU_STEP;                               // record accumulated time
}
//------------------------------------------------------------------------------
inline real trace(const real T[NDIM][NDIM])
{
  register real tr=zero;
  for(register int i=0; i<NDIM; i++) tr += T[i][i];
  return tr;
}
//------------------------------------------------------------------------------
#define LoopTensor for(i=0; i<NDIM; i++) for(j=0; j<NDIM; j++)
//------------------------------------------------------------------------------
void nbody::diagnose() const
{
  const    body*  BN = bodies()+No_bodies();
  register indx   i,j;
  register real   u=zero, m=zero;
  register vect   x=zero, v=zero;
  register amom   l=zero;
  register tensor t,w;
  LoopTensor t[i][j] = w[i][j] = zero;
  if(PEX) PEX->energies(u,w,t,x,v,l,m);
  for(register const body*Bi=bodies(); Bi<BN; Bi++) {
    m += mass(Bi);                                    // total mass
    u += mass(Bi) * pot(Bi);                          // potential energy
    LoopTensor {
      t[i][j] += mass(Bi) * vel (Bi)[i] * vel(Bi)[j]; // kin energy tensor
      w[i][j] += mass(Bi) * tpos(Bi)[i] * acc(Bi)[j]; // pot energy tensor
    }
    x += mass(Bi) * tpos(Bi);                         // dipole
    v += mass(Bi) * vel (Bi);                         // momentum
    l += mass(Bi) * tpos(Bi) ^ vel(Bi);               // total angular momentum
  }
  M    = m;                                           // total mass
  Ktot = half*trace(t);                               // total kin energy
  Utot = half*u;                                      // total pot energy
  TU   =-half*trace(t)/trace(w);                      // virial ratio
  L    = l;                                           // total angular momentum
  CMX  = x/m;                                         // center of mass
  CMV  = v/m;                                         // center of mass velocity
  LoopTensor {
    KT[i][j] = half * t[i][j];                        // kin energy tensor
    WT[i][j] =        w[i][j];                        // pos energy tensor
  }
}
#undef LoopTensor
//------------------------------------------------------------------------------
void nbody::update_diagnostics() const
{
  if(DIAG) return;
  diagnose();
  DIAG = true;                                        // diagnose is up-to-date
}
//------------------------------------------------------------------------------
const real  nbody::kin_energy    () const
{
  update_diagnostics();
  return Ktot;
}
//------------------------------------------------------------------------------
const real  nbody::pot_energy    () const
{
  update_diagnostics();
  return Utot;
}
//------------------------------------------------------------------------------
const real  nbody::total_energy    () const
{
  update_diagnostics();
  return Ktot+Utot;
}
//------------------------------------------------------------------------------
const amom& nbody::total_angmom    () const
{
  update_diagnostics();
  return L;
}
//------------------------------------------------------------------------------
const vect& nbody::total_momentum  () const
{
  update_diagnostics();
  return CMV;
}
//------------------------------------------------------------------------------
const vect& nbody::center_of_mass  () const
{
  update_diagnostics();
  return CMX;
}
//------------------------------------------------------------------------------
const real& nbody::kin_energy      (const indx i, const indx j) const
{
  update_diagnostics();
  return KT[i][j];
}
//------------------------------------------------------------------------------
const real& nbody::pot_energy      (const indx i, const indx j) const 
{
  update_diagnostics();
  return WT[i][j];
}
//------------------------------------------------------------------------------
const real  nbody::as_angmom    (const indx i, const indx j) const 
{
  update_diagnostics();
#if NDIM==3
  switch(i) {
  case 0:
    switch(j) {
    case 0:  return  zero;
    case 1:  return  L[2];
    default: return -L[1];
    }  
  case 1:
    switch(j) {
    case 0:  return -L[2];
    case 1:  return  zero;
    default: return  L[0];
    }  
  default:
    switch(j) {
    case 0:  return  L[1];
    case 1:  return -L[0];
    default: return  zero;
    }  
  }
#else
  return (i==j) ? zero : real(L);
#endif
}
//------------------------------------------------------------------------------
void nbody::long_stats(ostream& to) const
{
  if(!READY) error("nbody::long_stats(): not ready");
  to<<" cpu times [sec]: "
    <<"predict build tree/ext grav  eps_i  total   accum\n                  "
    <<setw(6)<<setprecision(4)<<cpu_predict()    <<" "
    <<setw(6)<<setprecision(4)<<cpu_build()      <<" "
    <<setw(6)<<setprecision(4)<<cpu_tree_force() <<" "
    <<setw(6)<<setprecision(4)<<cpu_ext_force()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_softening()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_longstep()   <<" "
    <<setw(8)<<setprecision(6)<<cpu_total()      <<"\n";
  BlockStep::short_stats(to); to<<"\n";
  if(time() != TINI) update_diagnostics();
  to<<" E="<<setw(8)<<setprecision(5)<<(Ktot+Utot)
    <<"; -T/U="<<setw(7)<<setprecision(5)<<TU
    <<"; L ="  
#if NDIM==3
    <<setw(7)<<setprecision(5)<<L[0]<<" "
    <<setw(7)<<setprecision(5)<<L[1]<<" "
    <<setw(7)<<setprecision(5)<<L[2]
#else
    <<setw(7)<<setprecision(5)<<L
#endif
    <<"; Vcm= "<<setw(7)<<setprecision(4)<<CMV[0]<<" "
    <<setw(7)<<setprecision(4)<<CMV[1]
#if NDIM==3
    <<" "<<setw(7)<<setprecision(4)<<CMV[2]
#endif
    <<endl;
}
//------------------------------------------------------------------------------
void nbody::short_stats(ostream& to) const
{
  if(!READY) error("nbody::short_stats(): not ready");
  to.setf(ios::left, ios::adjustfield);
  if(time() != TINI) update_diagnostics();
  to<<setprecision(5)<<setw(10)<<time()         <<" "
    <<setprecision(7)<<setw(11)<<(Ktot+Utot)    <<" "
    <<setprecision(min(5,max(1,5-int(-log10(TU)))))<<setw(7)<<TU<<" ";
#if NDIM==3
  to<<setprecision(5)<<setw(10)<<SQRT(norm(L))  <<" ";
#else
  to<<setprecision(5)<<setw(10)<<L              <<" ";
#endif
  to<<setprecision(2)<<setw(8)<<SQRT(norm(CMV))  <<" ";
  to.setf(ios::right, ios::adjustfield);
  if(shortest_step()) BlockStep::short_stats(to);
  if(MYSCHEME == rhos && !is_leap_frog() )
    to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
      <<setprecision(4)<<setw(6)<<cpu_density()  <<" "
      <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
      <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
  else
    to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
      <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
      <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
}
//------------------------------------------------------------------------------
void nbody::short_stat_head(ostream& to) const
{
  if(!READY) error("nbody::short_stats_head(): not ready");
  to<<"    time   "
    <<"  energy    "
    <<"  -T/U  "
#if NDIM==3
    <<"   |L|     "
#else
    <<"   L_z     "
#endif
    <<"  |v_cm| ";
  if(shortest_step())
    for(register int i=0, h=-h0(); i<=shortest_step(); i++, h--)
           if(h>13)  to<<"  #2^" <<     h  <<" ";
      else if(h> 9)  to<<"  #"   << (1<<h) <<" ";
      else if(h> 6)  to<<"   #"  << (1<<h) <<" ";
      else if(h> 3)  to<<"    #" << (1<<h) <<" ";
      else if(h>=0)  to<<"     #"<< (1<<h) <<" ";
      else if(h==-1) to<<"   #1/2 ";
      else if(h==-2) to<<"   #1/4 ";
      else if(h==-3) to<<"   #1/8 ";
      else if(h==-4) to<<"  #1/16 ";
      else if(h==-5) to<<"  #1/32 ";
      else if(h==-6) to<<"  #1/64 ";
      else if(h>-10) to<<"  #2^" <<     h  <<" ";
      else           to<<" #2^"  <<     h  <<" ";
  if(MYSCHEME == rhos &&  !is_leap_frog() )
    to<<" build    rho   force    step     accum"<<endl;
  else
    to<<" build   force    step     accum"<<endl;
}
//------------------------------------------------------------------------------
void nbody::short_stat_line(ostream &to) const
{
  to<<" --------------------------------------------------";
  if(shortest_step())
    for(register int i=0; i<=shortest_step(); i++) to<<"--------";
  if(MYSCHEME == rhos &&  !is_leap_frog() )
    to<<"---------------------------------------"<<endl;
  else
    to<<"--------------------------------"<<endl;
}
