//-----------------------------------------------------------------------------+
//                                                                             |
// nmio.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2001                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef  ALLOW_NEMO
#define  ALLOW_NEMO
#endif

#include <tree/nmio.h>
#include <iostream>

extern "C" {
#include <stdinc.h>
#include <filestruct.h>
#include <history.h>
#include <snapshot/snapshot.h>
}

#ifdef TWODIMENSIONAL
#define NDM 2
#else
#define NDM 3
#endif

using std::cerr;
using namespace fasttree;

//------------------------------------------------------------------------------
inline char* set_tag(const nemo_io::Set S)
{
  switch(S) {
  case nemo_io::snap:   return SnapShotTag;
  case nemo_io::param:  return ParametersTag;
  case nemo_io::bodies: return ParticlesTag;
  case nemo_io::diags:  return DiagnosticsTag;
  default: cerr<<"Unknown nemo_io::Set\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* array_tag(const nemo_io::Array A)
{
  switch(A) {
  case nemo_io::mass: return MassTag;
  case nemo_io::acc:  return AccelerationTag;
  case nemo_io::pot:  return PotentialTag;
  case nemo_io::aux:  return AuxTag;
  default: cerr<<"Unknown nemo_io::Array\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* matrix_tag(const nemo_io::Matrix M)
{
  switch(M) {
  case nemo_io::KinT: return KETensorTag;
  case nemo_io::PotT: return PETensorTag;
  case nemo_io::AmT:  return AMTensorTag;
  default: cerr<<"Unknown nemo_io::Matrix\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* vector_tag(const nemo_io::Vector V)
{
  switch(V) {
  case nemo_io::energy:  return EnergyTag;
  default: cerr<<"Unknown nemo_io::Vector\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline char* scalar_tag(const nemo_io::Scalar S)
{
  switch(S) {
  case nemo_io::time:    return TimeTag;
  case nemo_io::cputime: return "cputime";
  default: cerr<<"Unknown nemo_io::Scalar\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
inline int cosys_tag(const nemo_io::CoSys C)
{
  switch(C) {
  case nemo_io::cart:  return Cartesian;
  case nemo_io::spher: return Spherical;
  case nemo_io::scat:  return Scattering;
  default: cerr<<"Unknown nemo_io::CoSys\n"; exit(1);
  }
}
//------------------------------------------------------------------------------
void nemo_io::reset_N(const int n) const
{
  if(n != N) {
    if(AUXX)  delete[] AUXX;
    if(PHASE) delete[] PHASE;
    N = n;
    if(N) {
      AUXX  = new float[N];
      PHASE = new float[N*2*NDM];
    } else {
      AUXX  = 0;
      PHASE = 0;
    }
  }
}
//------------------------------------------------------------------------------
void nemo_io::open(char* file, char* control)
{
  if(STREAM) strclose(static_cast<stream>(STREAM));
  STREAM = static_cast<void*>(stropen(file,control));
}  
//------------------------------------------------------------------------------
void nemo_io::close()
{
  if(STREAM) strclose(static_cast<stream>(STREAM));
  STREAM = 0;
}  
//------------------------------------------------------------------------------
bool nemo_io::is_present(const Scalar S) const
{
  return get_tag_ok(static_cast<stream>(STREAM),scalar_tag(S));
}
//------------------------------------------------------------------------------
bool nemo_io::is_present(const Vector V) const
{
  return get_tag_ok(static_cast<stream>(STREAM),vector_tag(V));
}
//------------------------------------------------------------------------------
bool nemo_io::is_present(const Matrix M) const
{
  return get_tag_ok(static_cast<stream>(STREAM),matrix_tag(M));
}
//------------------------------------------------------------------------------
bool nemo_io::is_present(const Array A) const
{
  return get_tag_ok(static_cast<stream>(STREAM),array_tag(A));
}
//------------------------------------------------------------------------------
bool nemo_io::is_present(const Set S) const
{
  return get_tag_ok(static_cast<stream>(STREAM),set_tag(S));
}
//------------------------------------------------------------------------------
nemo_io::nemo_io(char* file, char* control)
{
  STREAM = static_cast<void*>(stropen(file,control));
  N     = 0;
  AUXX  = 0;
  PHASE = 0;
}
//------------------------------------------------------------------------------
nemo_io::~nemo_io()
{
  close();
  reset_N(0);
}
//------------------------------------------------------------------------------
void nemo_io::open_set(const Set S, const bool get) const
{
  if(get) get_set(static_cast<stream>(STREAM),set_tag(S));
  else    put_set(static_cast<stream>(STREAM),set_tag(S));
}
//------------------------------------------------------------------------------
void nemo_io::close_set(const Set S, const bool get) const
{
  if(get) get_tes(static_cast<stream>(STREAM),set_tag(S));
  else    put_tes(static_cast<stream>(STREAM),set_tag(S));
}
//------------------------------------------------------------------------------
int nemo_io::read_N() const
{
  int n;
  get_data(static_cast<stream>(STREAM),NobjTag,IntType,&n,0);
  reset_N(n);
  return N;
}
//------------------------------------------------------------------------------
float nemo_io::read_scalar(const Scalar S) const
{
  float scal;
  get_data_coerced(static_cast<stream>(STREAM),scalar_tag(S),FloatType,&scal,0);
  return scal;
}
//------------------------------------------------------------------------------
void nemo_io::read_vector(const Vector V) const
{
  get_data_coerced(static_cast<stream>(STREAM),
		   vector_tag(V),FloatType,VECTOR,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read_cofm() const
{
  get_data_coerced(static_cast<stream>(STREAM),
		   CMPhaseSpaceTag,FloatType,CMPHASE,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read_matrix(const Matrix M) const
{
  get_data_coerced(static_cast<stream>(STREAM),
		   matrix_tag(M),FloatType,MATRIX,NDM,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read_array(const Array A) const
{
  get_data_coerced(static_cast<stream>(STREAM),array_tag(A),FloatType,AUXX,N,0);
}
//------------------------------------------------------------------------------
void nemo_io::read_posvel() const
{
  get_data_coerced(static_cast<stream>(STREAM),
		   PhaseSpaceTag,FloatType,PHASE,N,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::read_history() const
{
  get_history(static_cast<stream>(STREAM));
}
//------------------------------------------------------------------------------
void nemo_io::write_N(const int n)
{
  reset_N(n);
  put_data(static_cast<stream>(STREAM),NobjTag,IntType,&N,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_coords(const CoSys C)
{
  CS = CSCode(cosys_tag(C), NDM, 2);
  put_data(static_cast<stream>(STREAM),CoordSystemTag,IntType,&CS,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_scalar(const Scalar S, const float scal)
{
  float s=scal;
  put_data(static_cast<stream>(STREAM),scalar_tag(S),FloatType,&s,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_vector(const Vector V)
{
  put_data(static_cast<stream>(STREAM),vector_tag(V),FloatType,VECTOR,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_cofm()
{
  put_data(static_cast<stream>(STREAM),CMPhaseSpaceTag,FloatType,CMPHASE,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_matrix(const Matrix M)
{
  put_data(static_cast<stream>(STREAM),matrix_tag(M),FloatType,MATRIX,NDM,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_array(const Array A)
{
  put_data(static_cast<stream>(STREAM),array_tag(A),FloatType,AUXX,N,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_posvel()
{
  put_data(static_cast<stream>(STREAM),PhaseSpaceTag,FloatType,PHASE,N,2,NDM,0);
}
//------------------------------------------------------------------------------
void nemo_io::write_history() const
{
  put_history(static_cast<stream>(STREAM));
}

#undef NDM
//------------------------------------------------------------------------------
