//-----------------------------------------------------------------------------+
//                                                                             |
// open.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000/2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <tree/open.h>
extern "C" {
#include <math.h>
}
using namespace fasttree;
////////////////////////////////////////////////////////////////////////////////
template<class S>
int hunt(const S*xarr, const int n, const S x, const int j)
// hunts the ordered table xarr for jlo such that xarr[jlo]<=x<xarr[jlo+1]
// on input j provides a guess for the final value of jlo.
// for an ascendingly ordered array, we return
//  -1 for         x < x[0]
//   i for x[i] <= x < x[i+1]  if  0<=i<n
// n-1 for         x == x[n-1]
// n   for         x >  x[n-1]
{
  int jm,jlo=j,jhi,l=n-1;
  int ascnd=(xarr[l]>xarr[0]);
  if(!ascnd && xarr[l]==xarr[0] ) return -1;             // x_0 = x_l
  if( ascnd && x<xarr[0] || !ascnd && x>xarr[0] ) return -1;
  if( ascnd && x>xarr[l] || !ascnd && x<xarr[l] ) return  n;
  if(jlo<0 || jlo>l) {                                   // input guess useless,
    jlo = -1;                                            //    go to bisection
    jhi = n;
  } else {
    int inc = 1;
    if(x>=xarr[jlo] == ascnd) {                         // hunt upward
      if(jlo == l) return (x==xarr[l])? l : n;
      jhi = jlo+1;
      while(x>=xarr[jhi] == ascnd) {                    // not done hunting
	jlo =jhi;
	inc+=inc;                                       // so double the increment
	jhi =jlo+inc;
	if(jhi>l) {                                     // off end of table
	  jhi=n;
	  break;
	}
      }
    } else {                                            // hunt downward
      if(jlo == 0) return -1;
      jhi = jlo;
      jlo-= 1;
      while(x<xarr[jlo] == ascnd) {                     // not done hunting
	jhi = jlo;
	inc+= inc;                                      // so double the increment
	jlo = jhi-inc;
	if(jlo < 0) {                                   // off end of table
	  jlo = 0;
	  break;
	}
      }
    }
  }
  while (jhi-jlo != 1) {                                // bisection phase
    jm=(jhi+jlo) >> 1;
    if(x>=xarr[jm] == ascnd) jlo=jm;
    else jhi=jm;
  }
  return jlo;
}
////////////////////////////////////////////////////////////////////////////////
template<class S>
inline int find_for_polev(int& j, const int n, const int m, const S *x,
			  const S xi)
{
    register int M=m;
    j = int( (xi-x[0]) / (x[n-1]-x[0]) * (n-1) );
    j = hunt(x,n,xi,j) - (m+1)/2 + 1;
    if(j>=0 && j<n && x[j]==xi)  M = 1;         // no interpolation required
    else if(j<0)                 j = 0;
    else if(j>n-M)               j = n-M;
    return M;
}

////////////////////////////////////////////////////////////////////////////////
template<class S, class T>
inline T polint(const S *xa, const T *ya, const int n, const S x)
// polynom interpolation using n values
{
    register int i,m;
    register T   y, *P=new T[n];

    for(i=0;i<n;i++) P[i]=ya[i];
    for(m=1;m<n;m++)
        for(i=0;i<n-m;i++) {
	  if(xa[i]==xa[i+m]) error("x's not distinct in OPENING");
            P[i]= ( (x-xa[i+m])*P[i] + (xa[i]-x)*P[i+1] ) / (xa[i] - xa[i+m]);
        }
    y = P[0];    
    delete[] P;
    return y;
}

////////////////////////////////////////////////////////////////////////////////
template<class S, class T>
inline T polev(S x, S*xarr, T*yarr, const int n, const int m=4)
// given the arrays xarr, yarr, polev returns y(x) using m of n values.
{
    int j, M=find_for_polev(j,n,m,xarr,x);
    return polint(xarr+j, yarr+j, M, x);
}

////////////////////////////////////////////////////////////////////////////////
// class open
void opening::set_table(const uint n, const real m)
{
  if(n != N) {
    if(Z)  delete[] Z;
    if(lM) delete[] lM;
    Z  = new double[n];
    lM = new double[n];
    N  = n;
  }
  register double alfa, zmin, zmax, dz;
  alfa = pow_int(TH0,5)/square(1-TH0);
  zmin = SQRT(POW(m,third)/alfa);
  zmin = SQRT(POW(m,third)/alfa*pow_uint(zmin+1.,5));
  zmax = (1-TH0)/TH0;
  dz   = (zmax - zmin)/double(n-1);
  for(register uint i=0; i<N; i++) {
    Z [i] = zmin + i*dz;
    lM[i] = 3 * log(alfa * square(Z[i])*pow_uint(Z[i]+1,3));
  }
}
//------------------------------------------------------------------------------
real opening::tabulated_theta(const real M) const
{
  return 1/(1+polev(log(M),lM,Z,N));
}
//------------------------------------------------------------------------------
real opening::tabulated_invtheta(const real M) const
{
  return 1+polev(log(M),lM,Z,N);
}
