//-----------------------------------------------------------------------------+
//                                                                             |
// stic.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// written by Walter Dehnen, 2001                                              |
// e-mail:   dehnen@mpia-hd.mpg.de                                             |
// address:  Max-Planck Institut fr Astronomie, Knigstuhl 69117 Heidelberg   |
//           Germany                                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// support for sticky particles                                                |
//                                                                             |
//-----------------------------------------------------------------------------+

#include <tree/stic.h>    // definition of sticky tree and interactor
#include <tree.h>         // definition of tree
#include <tree/iact.h>    // generic mutual interaction algorithm

using namespace fasttree;

//------------------------------------------------------------------------------
#define LoopAllSouls(C,X) for(register Scell::allsoul_iterator			\
			  X=C->begin_allsouls(); X!=C->end_allsouls(); ++X)
////////////////////////////////////////////////////////////////////////////////
// class sticky_finder

const indx N_STIC[3] = {100, 25, 100};

//------------------------------------------------------------------------------
inline void sticky_finder::add_pair(Ssoul*A,Ssoul*B) const
{
  if(N<NMAX)
#ifdef USE_BODIES
    if(mybody(A) < mybody(B)) {
      BLIST[N][0] = mybody(A);
      BLIST[N][1] = mybody(B);
    } else {
      BLIST[N][0] = mybody(B);
      BLIST[N][1] = mybody(A);
    }
#else  // USE_BODIES
    if(myelem(A) < myelem(B)) {
      BLIST[N][0] = myelem(A);
      BLIST[N][1] = myelem(B);
    } else {
      BLIST[N][0] = myelem(B);
      BLIST[N][1] = myelem(A);
    }
#endif // USE_BODIES
  N++;
  if(N==NMAX) warning("sticky particle interaction list overflow");
}
//------------------------------------------------------------------------------
void sticky_finder::interact(Ssoul*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job
  register vect R = pos(A)-pos(B);                   // R   = distance vector
  register real xq= square(size(A)+size(B));         // x^2 = (combined size)^2
  if(norm(R) < xq) return add_pair(A,B);             // |R|<x: overlap at t=0
  if(TAU==zero) return;                              // no times t>0
  register vect V = vel(A)-vel(B);                   // V   = velocity difference
  register real RV= R*V;                             // R*V = scalar product
  if(RV>zero) return;                                // R*V>0: diverging orbits
  register real t = min(TAU, -RV/norm(V));           // t of minimum distance
  if(norm(R+t*V) < xq) add_pair(A,B);                // d_min<x: overlapping
}
//------------------------------------------------------------------------------
#define CHECK_PAIR								\
  R = pos(A)-pos(B);                                 /* distance vector     */	\
  xq= square(size(A)+size(B));                       /* (combined size)^2   */	\
  if(norm(R) < xq) { add_pair(A,B); continue; }      /* overlap at t=0      */	\
  if(TAU==zero) continue;                            /* no times t>0        */	\
  V = vel(A)-vel(B);                                 /* velocity difference */	\
  RV= R*V;                                           /* scalar product      */	\
  if(RV>zero) continue;                              /* diverging orbits    */	\
  t = min(TAU,-RV/norm(V));                          /* time of min distance*/	\
  if(norm(R+t*V) < xq) add_pair(A,B);                /* overlap: add to list*/

inline void sticky_finder::many(const bool all, Ssoul*A,
				Ssoul*B0, const Ssoul*BN) const
{
  register vect   R,V;
  register real   xq,RV,t;
  register Ssoul *B;
  if(all) for(B=B0; B<BN; B++)                { CHECK_PAIR }
  else    for(B=B0; B<BN; B++) if(is_sink(B)) { CHECK_PAIR }
}
#undef CHECK_PAIR
//------------------------------------------------------------------------------
inline bool sticky_finder::many(Scell*A, Ssoul*B) const
{
  many(is_sink(B) || al_sink(A), B, A->begin_allsouls(), A->end_allsouls());
  return true;
}
//------------------------------------------------------------------------------
inline bool sticky_finder::many(Scell*A, Scell*B) const
{
  LoopAllSouls(B,Bi) many(A,Bi);
  return true;
}
//------------------------------------------------------------------------------
inline bool sticky_finder::many(Scell*A) const
{
  LoopAllSouls(A,Ai)
    many(al_sink(A),Ai,Ai+1,A->end_allsouls());
  return true;
}
//------------------------------------------------------------------------------
inline bool sticky_finder::discard(Scell*A, Ssoul*B) const
// true  if there cannot possibly be an interaction between any A and B
// false if an interaction between any A and B is possible
{
  register vect R = pos(A)-pos(B);                   // R   = distance vector
  register real
    Rq = norm(R),                                    // R^2 = distance^2
    x  = size(A)+size(B);                            // x   = combined size
  if(Rq < x*x) return false;                         // overlap at t=0
  if(TAU==zero) return true;                         // no times t>0
  register vect V = vel(A)-vel(B);                   // V   = velocity difference
  register real
    w  = vrad(A),                                    // w   = v-size
    wq = w*w,                                        // w^2 = (v-size)^2
    RV = R*V,                                        // R*V = scalar product
    RVq= RV*RV;                                      // (R*V)^2
  if(RV>zero && RVq>wq*Rq) return true;              // R*V/|R| > w: diverging
  register real
    Vq = norm(V),                                    // V^2 = (v-distance)^2
    t  = (wq>=Vq)? TAU :                             // v-size too large 
    min(TAU, (w*SQRT((Rq*Vq-RVq)/(Vq-wq))-RV)/Vq );  // t of min dist
  if(norm(R+t*V) < square(x+t*w)) return false;      // possible overlap
  return true;                                       // no overlap -> discard
}
//------------------------------------------------------------------------------
inline bool sticky_finder::discard(Scell*A, Scell*B) const
// true  if there cannot possibly be an interaction between any A and B
// false if an interaction between any A and B is possible
{
  register vect R = pos(A)-pos(B);                   // R   = distance vector
  register real
    Rq = norm(R),                                    // R^2 = distance^2
    x  = size(A)+size(B);                            // x   = combined size
  if(Rq < x*x) return false;                         // overlap at t=0
  if(TAU==zero) return true;                         // no times t>0
  register vect V = vel(A)-vel(B);                   // V   = velocity difference
  register real
    w  = vrad(A)+vrad(B),                            // w   = combined v-size
    wq = w*w,                                        // w^2 = (combined v-size)^2
    RV = R*V,                                        // R*V = scalar product
    RVq= RV*RV;                                      // (R*V)^2
  if(RV>zero && RVq>wq*Rq) return true;              // R*V/|R| > w: diverging
  register real
    Vq = norm(V),                                    // V^2 = (v-distance)^2
    t  = (wq>=Vq)? TAU :                             // v-size too large 
    min(TAU, (w*SQRT((Rq*Vq-RVq)/(Vq-wq))-RV)/Vq );  // t of min dist
  if(norm(R+t*V) < square(x+t*w)) return false;      // possible overlap
  return true;                                       // no overlap -> discard
}
//------------------------------------------------------------------------------
bool sticky_finder::interact(Scell*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)<N_STIC[0])          return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool sticky_finder::interact(Scell*A, Scell*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)*number(B)<N_STIC[1])return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool sticky_finder::interact(Scell*A) const
{
  if(!is_sink(A))                  return true;      // no sinks -> no job
  if(number(A)<N_STIC[2])          return many(A);   // try many single
                                   return false;     // must split
}
////////////////////////////////////////////////////////////////////////////////
// class neighbour_finder

const indx N_SPHM[3] = {100, 25, 100};

//------------------------------------------------------------------------------
inline void neighbour_finder::add_pair(Ssoul*A, Ssoul*B) const
{
  if(N<NMAX)
#ifdef USE_BODIES
    if(mybody(A) < mybody(B)) {
      BLIST[N][0] = mybody(A);
      BLIST[N][1] = mybody(B);
    } else {
      BLIST[N][0] = mybody(B);
      BLIST[N][1] = mybody(A);
    }
#else  // USE_BODIES
    if(myelem(A) < myelem(B)) {
      BLIST[N][0] = myelem(A);
      BLIST[N][1] = myelem(B);
    } else {
      BLIST[N][0] = myelem(B);
      BLIST[N][1] = myelem(A);
    }
#endif // USE_BODIES
  N++;
  if(N==NMAX) warning("SPH interaction list overflow");
}
//------------------------------------------------------------------------------
void neighbour_finder::interact(Ssoul*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job
  register real Rq = dist_sq(pos(A),pos(B));         // R^2 = distance^2        
  if(Rq < sizeq(A) || Rq < sizeq(B)) add_pair(A,B);  // |R| < max(s_A,s_B) ?
}
//------------------------------------------------------------------------------
#define CHECK_PAIR								\
  Rq = dist_sq(pos(A),pos(B));                       /* R^2 = distance^2     */	\
  if(Rq < sizeq(A) || Rq < sizeq(B)) add_pair(A,B);  /* |R| < max(s_A,s_B) ? */

inline void neighbour_finder::many(const bool all, Ssoul*A,
			    Ssoul*B0, const Ssoul*BN) const
{
  register real Rq;
  register Ssoul *B;
  if(all) for(B=B0; B<BN; B++)                { CHECK_PAIR }
  else    for(B=B0; B<BN; B++) if(is_sink(B)) { CHECK_PAIR }
}
#undef CHECK_PAIR
//------------------------------------------------------------------------------
inline bool neighbour_finder::many(Scell*A, Ssoul*B) const
{
  many(is_sink(B) || al_sink(A), B, A->begin_allsouls(), A->end_allsouls());
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_finder::many(Scell*A, Scell*B) const
{
  LoopAllSouls(B,Bi) many(A,Bi);
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_finder::many(Scell*A) const
{
  LoopAllSouls(A,Ai)
    many(al_sink(A),Ai,Ai+1,A->end_allsouls());
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_finder::discard(Scell*A, Ssoul*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2
    x  = max(rmax(A)+size(B),size(A));               // max size distance
  if(Rq < square(x)) return false;                   // |R| < max(s_A,r_A+s_B)
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_finder::discard(Scell*A, Scell*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2
    x  = max(rmax(A)+size(B),rmax(B)+size(A));       // max size distance
  if(Rq < square(x)) return false;                   // |R| < max(r_b+s_A,r_A+s_B)
  return true;
}
//------------------------------------------------------------------------------
bool neighbour_finder::interact(Scell*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)<N_SPHM[0])          return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool neighbour_finder::interact(Scell*A, Scell*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)*number(B)<N_SPHM[1])return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool neighbour_finder::interact(Scell*A) const
{
  if(!is_sink(A))                  return true;      // no sinks -> no job
  if(number(A)<N_SPHM[2])          return many(A);   // try many single
                                   return false;     // must split
}
////////////////////////////////////////////////////////////////////////////////
// class neighbour_counter
void neighbour_counter::interact(Ssoul*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B)) return;             // no sinks -> no job
  register real Rq = dist_sq(pos(A),pos(B));         // R^2 = distance^2  
  if(is_sink(A) && Rq < sizeq(A)) A->inc();          // A is sink && |R|<siza(A)
  if(is_sink(B) && Rq < sizeq(B)) B->inc();          // B is sink && |R|<siza(B)
}
//------------------------------------------------------------------------------
inline void neighbour_counter::many(Ssoul*A, const bool A_sink, 
				    Ssoul*B0, const Ssoul*BN,
				    const bool al_B_sink) const
{
  register real Rq;
  register Ssoul *B;
  if(A_sink)
    if(al_B_sink)
      for(B=B0; B<BN; B++) {
	Rq = dist_sq(pos(A),pos(B));
	if(Rq < sizeq(A)) A->inc();
	if(Rq < sizeq(B)) B->inc();
      }
    else
      for(B=B0; B<BN; B++) {
	Rq = dist_sq(pos(A),pos(B));
	if(              Rq < sizeq(A)) A->inc();
	if(is_sink(B) && Rq < sizeq(B)) B->inc();
      }
  else
    if(al_B_sink)
      for(B=B0; B<BN; B++) {
	Rq = dist_sq(pos(A),pos(B));
	if(Rq < sizeq(B)) B->inc();
      }
    else
      for(B=B0; B<BN; B++) if(is_sink(B)) {
	Rq = dist_sq(pos(A),pos(B));
	if(Rq < sizeq(B)) B->inc();
      }
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::many(Scell*A, Ssoul*B) const
{
  many(B, is_sink(B), A->begin_allsouls(), A->end_allsouls(), al_sink(A));
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::many(Scell*A, Scell*B) const
{
  if(al_sink(B))
    LoopAllSouls(B,Bi)
      many(Bi, true, A->begin_allsouls(), A->end_allsouls(), al_sink(A));
  else
    LoopAllSouls(B,Bi)
      many(Bi, is_sink(Bi), A->begin_allsouls(), A->end_allsouls(), al_sink(A));
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::many(Scell*A) const
{
  if(al_sink(A))
    LoopAllSouls(A,Ai)
      many(Ai, true, Ai+1, A->end_allsouls(), true);
  else
    LoopAllSouls(A,Ai)
      many(Ai, is_sink(Ai), Ai+1, A->end_allsouls(), false);
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::discard(Scell*A, Ssoul*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2
    x  = max(rmax(A)+size(B),size(A));               // max size distance
  if(Rq < square(x)) return false;                   // |R| < max(s_A,r_A+s_B)
  return true;
}
//------------------------------------------------------------------------------
inline bool neighbour_counter::discard(Scell*A, Scell*B) const
{
  register real
    Rq = dist_sq(pos(A),pos(B)),                     // R^2 = distance^2
    x  = max(rmax(A)+size(B),rmax(B)+size(A));       // max size distance
  if(Rq < square(x)) return false;                   // |R| < max(r_b+s_A,r_A+s_B)
  return true;
}
//------------------------------------------------------------------------------
bool neighbour_counter::interact(Scell*A, Ssoul*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)<N_SPHM[0])          return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool neighbour_counter::interact(Scell*A, Scell*B) const
{
  if(!is_sink(A) && !is_sink(B))   return true;      // no sinks -> no job
  if(discard(A,B))                 return true;      // try to discard
  if(number(A)*number(B)<N_SPHM[1])return many(A,B); // try many single
                                   return false;     // must split
}
//------------------------------------------------------------------------------
bool neighbour_counter::interact(Scell*A) const
{
  if(!is_sink(A))                  return true;      // no sinks -> no job
  if(number(A)<N_SPHM[2])          return many(A);   // try many single
                                   return false;     // must split
}
////////////////////////////////////////////////////////////////////////////////
#define LoopSubCells         for(register Scell::subcell_iterator		\
			     c=Ci->begin_subcells(); c!=Ci->end_subcells(); ++c)
#define LoopSubSouls         for(register Scell::subsoul_iterator		\
			     s=Ci->begin_subsouls(); s!=Ci->end_subsouls(); ++s)
#define LoopCellsDown(TYPE)  for(register TYPE *Ci=C0;   Ci< CN; Ci++)
#define LoopCellsUp(TYPE)    for(register TYPE *Ci=CN-1; Ci>=C0; Ci--)
#define LoopSouls(TYPE)      for(register TYPE *Si=S0;   Si< SN; Si++)
////////////////////////////////////////////////////////////////////////////////
// class Stree
void Stree::prepare_sticky()
{
  if(S0==0) return;
  // 1. update souls with size and vel (body flag & pos known from tree souls)
#ifdef USE_BODIES
  LoopSouls(Ssoul) Si->set_sticky();
#else  // USE_BODIES
  LoopSouls(Ssoul) Si->set_sticky(S,V);
#endif // USE_BODIES
  // 2. for all cells: compute Sum x_i and Sum v_i, pass sink flag up
  register vect spos,svel;
  LoopCellsUp(Scell) {
    spos = zero;
    svel = zero;
    Ci->reset_flag();
//     LoopSubSouls { spos+= pos(s); svel+= vel(s); Ci->add_body_flag(s); }
    LoopSubSouls { spos+= pos(s); svel+= vel(s); Ci->add_body_flag(s); }
    LoopSubCells { spos+= pos(c); svel+= vel(c); Ci->add_body_flag(c); }
    Ci->pos() = spos;
    Ci->vel() = svel;
  }
  // 3. for all cells: normalize pos & vel; compute size & vrad
  register real iN,dmax,vmax;
  LoopCellsUp(Scell) {
    iN = 1./real(number(Ci));
    Ci->pos() *= iN;
    Ci->vel() *= iN;
    dmax = zero;
    vmax = zero;
//     LoopSubSouls {
    LoopSubSouls {
      update_max(dmax,SQRT(dist_sq(pos(Ci),pos(s))) + size(s));
      update_max(vmax,SQRT(dist_sq(vel(Ci),vel(s))));
    }
    LoopSubCells {
      update_max(dmax,SQRT(dist_sq(pos(Ci),pos(c))) + size(c));
      update_max(vmax,SQRT(dist_sq(vel(Ci),vel(c))) + vrad(c));
    }
    Ci->size() = dmax;
    Ci->vrad() = vmax;
  }
}
//------------------------------------------------------------------------------
void Stree::prepare_sph()
{
  if(S0==0) return;
  // 1. update souls with size and size^2 (sink flag & pos known from tree souls)
#ifdef USE_BODIES
  LoopSouls(Ssoul) Si->set_sph();
#else  // USE_BODIES
  LoopSouls(Ssoul) Si->set_sph(S);
#endif // USE_BODIES
  // 2. for all cells: compute Sum x_i, pass sink flag up
  register vect spos;
  LoopCellsUp(Scell) {
    spos = zero;
    Ci->reset_flag();
//     LoopSubSouls { spos+= pos(s); Ci->add_body_flag(s); }
    LoopSubSouls { spos+= pos(s); Ci->add_body_flag(s); }
    LoopSubCells { spos+= pos(c); Ci->add_body_flag(c); }
    Ci->pos() = spos;
  }
  // 3. for all cells: normalize pos, compute size
  register real iN,dmax,smax,d;
  LoopCellsUp(Scell) {
    iN = 1./real(number(Ci));
    Ci->pos() *= iN;
    dmax = zero;
    smax = zero;
//     LoopSubSouls {
    LoopSubSouls {
      d = SQRT(dist_sq(pos(Ci),pos(s)));
      update_max(dmax, d + size(s));
      update_max(smax, d);
    }
    LoopSubCells {
      d = SQRT(dist_sq(pos(Ci),pos(c)));
      update_max(dmax, d + size(c));
      update_max(smax, d + rmax(c));
    }
    Ci->size() = dmax;
    Ci->rmax() = smax;
  }
}
//------------------------------------------------------------------------------
#ifdef USE_BODIES
void Stree::set_num(body* B0) const
{
  LoopSouls(Ssoul) if(is_sink(Si))
    (B0+int(mybody(Si)-B0))->rho() = Si->num();
}
#else  // USE_BODIES
void Stree::set_num(areal*NUM) const
{
  LoopSouls(Ssoul) if(is_sink(Si)) NUM[myelem(Si)] = Si->num();
}
#endif // USE_BODIES
////////////////////////////////////////////////////////////////////////////////
#undef LoopSubCells
#undef LoopSubSouls
#undef LoopCellsDown
#undef LoopCellsUp
#undef LoopSouls
#undef LoopAllSouls
////////////////////////////////////////////////////////////////////////////////
