//-----------------------------------------------------------------------------+
//                                                                             |
// tree.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines all function members of class tree                                  |
//                                                                             |
// Contents                                                                    |
//                                                                             |
// 1  Constructor, destructor and such                                         |
// 2  tree growth and re-use                                                   |
// 3  information and statistics                                               |
// 4  estimation of density and such                                           |
// 5  approximation of gravity                                                 |
// 6  neighbour finding and counting                                           |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <tree.h>                   // class tree                               
#ifdef USE_BODIES
#include <body.h>                   // class body and all its members           
#endif
#include <tree/buil.h>              // class builder                            
#include <tree/soul.h>              // class soul and all its members           
#include <tree/cell.h>              // class cell and all its members           
#include <tree/tayl.h>              // classes TaylorCoeffs and TaylorGrav      
#include <tree/open.h>              // class opening                            
#include <tree/grav.h>              // class interactor                         
#include <tree/iact.h>              // generic interaction algorithm            
#include <tree/stic.h>              // support for SPH & sticky particles       
#include <tree/stat.h>              // statistics & error messages              
#ifndef C_AND_F
#include <iomanip.h>                // C++ I/O formatting etc                   
#endif // C_AND_F
using std::ostream;                 // we'll use ostream                        
using std::ofstream;                // and ofstream                             
using namespace fasttree;           // make all tree stuff available            
////////////////////////////////////////////////////////////////////////////////
// some useful macros                                                           
////////////////////////////////////////////////////////////////////////////////

#define LoopSubCells  for(register cell::subcell_iterator			\
			  c=Ci->begin_subcells(); c!=Ci->end_subcells(); c++)
#define LoopSubSouls  for(register cell::subsoul_iterator			\
			  s=Ci->begin_subsouls(); s!=Ci->end_subsouls(); s++)
#define LoopCellsDown for(register cell* Ci=C0;   Ci< CN; Ci++)
#define LoopCellsUp   for(register cell* Ci=CN-1; Ci>=C0; Ci--)
#define LoopSouls     for(register soul* Si=S0;   Si< SN; Si++)

// Note the difference between looping over the cells up or downwards. Since the
// the cells are aligned in memory (such that descendent cells are behind their 
// parents), looping upwards (downwards) ensures that the loop has already      
// visited all the descendents (ancestors) of the current cell.                 
// Thus, by looping the proper way, recursive calls can largely be avoided.     

//=============================================================================+
//                                                                             |
// 1 CONSTRUCTOR, DESTRUCTOR AND SUCH                                          |
// ==================================                                          |
//                                                                             |
//=============================================================================+

#ifdef USE_BODIES
tree::tree(      body*     b,
	   const uint      n,
	   const real      e,
	   const real      th,
	   const kern_type t,
	   const uint      d)
  : Nb(n), Ns(0), Nc(0), Nz(0), B0(b), BN(b+Nb), S0(0), SN(0), C0(0), CN(0),
    OPENING(new opening(abs(th),(NDIM==2 || th<zero))),
    STATS(new TreeStats()),
    DMAX(d)
{
  reset_softening(e,t);
}
#else  // USE_BODIES
//------------------------------------------------------------------------------
tree::tree(const int* f, const areal* m, const areal*x[NDIM], const areal* ep,
	   areal* a[NDIM], areal* ph, areal* rh, const unsigned n,
	   const real e, const real th, const kern_type t, const unsigned d)
  : Nb(n), Ns(0), Nc(0), Nz(0), FL(f),M(m),EP(ep),X(x),PH(ph),RH(rh),A(a),
    S0(0),SN(0), C0(0),CN(0),
    OPENING(new opening(abs(th),(NDIM==2 || th<zero))),
    STATS(new TreeStats()),
    DMAX(d)
{
  reset_softening(e,t);
}
#endif
//------------------------------------------------------------------------------
tree::~tree()
{
  if(C0) delete[] C0;
  if(S0) delete[] S0;
  delete OPENING;
  delete STATS;
}
//------------------------------------------------------------------------------
void tree::reset_softening (const real e, const kern_type t) const
{
  SOFTENING = (e<zero)? individual : global;
  EPS       = (e<zero)? zero : e;
  KERNEL    = t;
}
//------------------------------------------------------------------------------
void tree::reset_opening(const real t) const
{
  OPENING->reset(abs(t),t<zero);
}
//------------------------------------------------------------------------------
inline cell* tree::root()
{
  return C0;
}
//------------------------------------------------------------------------------
inline const cell* tree::root() const
{
  return C0;
}
//=============================================================================+
//                                                                             |
// 2 TREE GROWTH AND RE-USE                                                    |
// ========================                                                    |
//                                                                             |
//=============================================================================+
void tree::make(const uint  Ncrit, const vect* pc)
  // This routines builds a new tree with root centered on *pc (if pc!=0) and   
  // with cell divived until they contain not more than Ncrit souls             
  // The routines works as follows:                                             
  // 1 A box-leaf tree is established:                                          
  //   - the active bodies are counted and a linked list of leafs is prepared.  
  //   - the root box is established (center, radius, list of leafs).           
  //   - a tree is build by recursive subdivision of boxes                      
  // 2 A cell-soul tree is established:                                         
  //   - the (known) appropriate number of souls & cells is allocated           
  //   - souls & cells are linked to the tree using the box-leaf tree           
  //   - the box leaf tree is discarded                                         
{
#ifdef USE_BODIES
  builder PM(Ncrit,B0,Nb);                         // initiliaze tree builder   
#else  // USE_BODIES
  builder PM(Ncrit,X,M,FL,Nb);                     // initiliaze tree builder   
#endif // USE_BODIES
  if(PM.N_leafs() == 0) {                          // IF(no leaf in tree) >     
    warning(" no bodies in tree");                 //   issue a warning         
    STATE = unset;                                 //   set state               
    return;                                        //   and return undone       
  }                                                // <                         
  PM.build_tree(pc,DMAX);                          // build box-leaf tree       
  if(C0) delete[] C0;                              // deallocate cells if needed
  Nc = PM.N_needed();                              // # cells = # active boxes  
  C0 = new cell[Nc];                               // allocate cells            
  CN = C0 + Nc;                                    // pointer past last cell    
  if(S0) delete[] S0;                              // deallocate souls if needed
  Ns = PM.N_leafs();                               // # souls = # active bodies 
  S0 = new soul[Ns];                               // allocate souls            
  SN = S0+Ns;                                      // pointer past last soul    
  PM.link_tree(C0,S0);                             // link box-leaf <> cell-soul
  Nz = PM.N_zombies();                             // # zombie cells            
}
//------------------------------------------------------------------------------
inline real bmax(const cell *C)
  // This routines returns the distance from the cell's cofm to its most        
  // distant corner.                                                            
{
  register real bq=zero;                           // initialize return variable
  for(register indx i=0; i<NDIM; i++)              // loop dimensions           
    bq += square(radius(C)+abs(cofm(C)[i]-center(C)[i])); // add in square      
  return SQRT(bq);                                 // return sqrt of total      
}
//------------------------------------------------------------------------------
inline void eval_basic_source(const bool reuse, cell*C0, cell*CN)
  // This routines computes for all cells of TREE:                              
  //  - mass, 1/mass, cofm, rmax (needed for almost everything)                 
  // To be uses AFTER souls have been updated                                   
{
  register real  mon;                              // monopole                  
  register vect  dip;                              // dipole                    
  LoopCellsUp {                                    // loop cells upwards >      
#ifdef TESTBODIES
    Ci->reset_source_flag();                       //   reset cell's source flag
#endif // TESTBODIES
    mon = zero;                                    //   reset monopole          
    dip = zero;                                    //   reset dipole            
    LoopSubCells {                                 //   loop sub-cells c >      
#ifdef TESTBODIES
      Ci->add_source_flag(c);                      //     update source flag    
      if(is_source(c)) {                           //     IF(c is source) >     
#endif // TESTBODIES
	mon += mass  (c);                          //       sum up monopole     
	dip += cofm  (c);                          //       sum up dipole       
#ifdef TESTBODIES
      }                                            //     <                     
#endif // TESTBODIES
    }                                              //   <                       
    LoopSubSouls {                                 //   loop sub-souls s >      
#ifdef TESTBODIES
      Ci->add_source_flag(s);                      //     update source flag    
      if(is_source(s)) {                           //     IF(s is source) >     
#endif // TESTBODIES
	mon += mass(s);                            //       sum up monopole     
	dip += mass(s)*cofm(s);                    //       sum up dipole       
#ifdef TESTBODIES
      }                                            //     <                     
#endif // TESTBODIES
    }                                              //   <                       
    Ci->mass() = mon;                              //   set mass                
    Ci->imass()= (mon==zero)? zero:one/mon;        //   set 1/mass              
    Ci->cofm() = dip;                              //   set dipole              
  }                                                // <                         
  register real dmax, x, Xq;                       // d_max, auxiliary vars     
  LoopCellsUp {                                    // loop cells upwards >      
#ifdef TESTBODIES
    if(is_source(Ci)) Ci->cofm()*=imass(Ci);       //  cofm = dipole/mass      
    else              Ci->cofm() =center(Ci);      //       = center if mass=0 
#else  // TESTBODIES
    Ci->cofm() *= imass(Ci);                       //   cofm = dipole/mass      
#endif // TESTBODIES
    dmax = zero;                                   //   reset dmax              
    LoopSubSouls                                   //   loop sub-souls s >      
      update_max(dmax,dist_sq(cofm(s),cofm(Ci)));  //     update d_max^2 <      
    if(has_subsouls(Ci)) dmax = SQRT(dmax);        //   d_max due to sub-souls  
    LoopSubCells {                                 //   loop sub-cells c >      
      Xq = dist_sq(cofm(c),cofm(Ci));              //     distance^2            
      x  = dmax - rmax(c);                         //     d=distance + r_i      
      if(zero > x || Xq>x*x)                       //     IF(d>d_max)      >    
	dmax = SQRT(Xq) + rmax(c);                 //       set d_max = d  <    
    }                                              //   <                       
    Ci->rmax() = reuse? dmax:min(dmax,bmax(Ci));   //   assign r_max       
  }                                                // <                         
}
//------------------------------------------------------------------------------
void tree::reuse()
{
  // This routine uses the existing tree and                                    
  // -  re-sets the souls' positions, masses, and source flags                  
  // -  re-evaluates the basic source properties of the cells                   
  // and pre-compute basic source properties of cells                           
#ifdef USE_BODIES
  LoopSouls Si->set_up();                          // re-set souls              
#else
  LoopSouls Si->set_up(M,X);                       // re-set souls              
#endif // USE_BODIES
  eval_basic_source(true,C0,CN);                   // re-evaluate cells' props  
  STATE=built;                                     // set tree state            
}
//------------------------------------------------------------------------------
void tree::grow(const uint n, const vect*p)
  // This routines grows a fresh tree & evaluates the cells' basic source       
  // properties.                                                                
{
  make(n,p);                                       // grow tree from scratch    
  eval_basic_source(false,C0,CN);                  // evaluate cells' props     
  STATE=built;                                     // set tree state            
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 3 INFORMATION AND STATISTICS                                                |
// ============================                                                |
//                                                                             |
//-----------------------------------------------------------------------------+
const char* tree::opening_method() const
{
  return OPENING->method();
}
//------------------------------------------------------------------------------
real tree::theta(const real m) const
{
  return OPENING->theta(m);
}
//------------------------------------------------------------------------------
fasttree::uint tree::No_zombies() const
{
  register uint  n=0;
#ifdef USE_BODIES
  for(register soul* Si=S0; Si<SN; Si++)
    if(!is_in_tree(mybody(Si))) n++;
#else  // USE_BODIES
  for(register soul* Si=S0; Si<SN; Si++)
    if(!is_in_tree(FL[myelem(Si)])) n++;
#endif // USE_BODIES
  return n;
}
//------------------------------------------------------------------------------
#ifndef C_AND_F
namespace fasttree {
  inline void dump_soul(const soul *Si, ostream&o, const soul*S0)
  {
    o <<setw(5)<< int(Si-S0);
#ifdef SUBSOUL_LLIST
    o <<setw(5)<< (Si->NEXT? int(Si->NEXT-S0) : -9999)
#endif
    for(register indx i=0; i<NDIM; i++)
      o<<" "<<setw(8)<<setprecision(4)<<cofm(Si)[i];
    o<<" "<<setw(8)<<setprecision(4)<<mass(Si)<<"\n";
  }
  //----------------------------------------------------------------------------
  inline void dump_cell(const cell*C, ostream& o, const cell* C0, const soul* S0)
  {
    o     <<setw(5)<< int(C-C0)
     <<" "<<setw(5)<< (C->CELLS? int(C->CELLS-C0) : -99)
#ifdef SUBCELL_LLIST
     <<" "<<setw(5)<< (C->NEXT ? int(C->NEXT -C0) : -99)
#else
     <<" "<<setw(5)<< C->NCELLS
#endif
     <<" "<<setw(5)<< (C->SOULS? int(C->SOULS-S0) : -9999)
#ifndef SUBSOUL_LLIST
     <<" "<<setw(5)<< C->NSOULS
#endif
     <<" "<<setw(5)<< C->NUMBER
     <<" "<<setw(2)<< C->DEPTH
     <<" "<<setw(4)<< flag(*C)
     <<" "<<setw(8)<<setprecision(4)<<radius(C);
    register indx i;
    for(i=0; i<NDIM; i++) o<<" "<<setw(9)<<setprecision(6)<<center(C)[i];
    o<<" "<<setw(8)<<mass(C);
    for(i=0; i<NDIM; i++) o<<" "<<setw(8)<<setprecision(4)<<cofm(C)[i];
    o<<" "<<setw(12)<<rmax(C)
     <<"\n";
  }
  //----------------------------------------------------------------------------
  inline void dump_nodes(cell *Ci, ostream* oc, ostream* ob,
			 const cell*C0, const soul*S0)
    // RECURSIVE
  {
    if(oc) dump_cell(Ci,*oc,C0,S0);
    LoopSubCells dump_nodes(c,oc,ob,C0,S0);
    if(ob) LoopSubSouls dump_soul (s,*ob,S0);
  }
  //----------------------------------------------------------------------------
  void tree::dump_nodes(const char* fcells, const char* fleafs) const
  {
    ofstream *oc = fcells? new ofstream(fcells) : 0;
    ofstream *ob = fleafs? new ofstream(fleafs) : 0;
#ifdef SUBCELL_LLIST
#ifdef SUBSOUL_LLIST
    if(oc) (*oc)<<"    # cells  next souls number "
		<<"de flag   radius             center"
		<<"               mass               cofm               rmax\n";
    if(ob) (*ob)<<"    #  next "
		<<"             cofm               mass\n";
#else
    if(oc) (*oc)<<"    # cells  next souls nsoul number "
		<<"de flag   radius             center"
		<<"               mass               cofm               rmax\n";
    if(ob) (*ob)<<"    # "
		<<"             cofm               mass\n";
#endif
#else
#ifdef SUBSOUL_LLIST
    if(oc) (*oc)<<"    # cells ncell souls number "
		<<"de flag   radius             center"
		<<"               mass               cofm               rmax\n";
    if(ob) (*ob)<<"    #  next "
		<<"             cofm               mass\n";
#else
    if(oc) (*oc)<<"    # cells ncell souls nsoul number "
		<<"de flag   radius             center"
		<<"               mass               cofm               rmax\n";
    if(ob) (*ob)<<"    # "
		<<"             cofm               mass\n";
#endif
#endif
    if(oc || ob) ::dump_nodes(C0,oc,ob,C0,S0);
    if(oc) { oc->flush(); delete oc; }
    if(ob) { ob->flush(); delete ob; }
  }
}
//------------------------------------------------------------------------------
inline int trick(const real x, const int w)
{
  if(x<one)    return w-2;
  if(x<ten)    return w-1;
  return w;
}
//------------------------------------------------------------------------------
inline int trick2(const real x, const int w)
{
  if(x<one)       return w;
  if(x<ten)       return w+1;
  if(x<100.)      return w+2;
  if(x<1000.)     return w+3;
  if(x<10000.)    return w+4;
  if(x<100000.)   return w+5;
  if(x<1000000.)  return w+6;
  return w+7;
}
//------------------------------------------------------------------------------
const char* tree::describe(const kern_type KER)
{
  switch(KER) {
  case k0: return "K0";
  case k1: return "K1";
  case k2: return "K2";
  case k3: return "K3";
  case f0: return "F0";
  case f1: return "F1";
  case f2: return "F2";
  case f3: return "F3";
  case p0: return "P0==Plummer";
  case p1: return "P1";
  case p2: return "P2";
  case p3: return "P3";
  default: return "F1";
  }
}
//------------------------------------------------------------------------------
void tree::stats(ostream& out) const
{
  out<<"\n state:                ";
  if(STATE&unset)  out<<" unset";
  if(STATE&built)  out<<" built";
  if(STATE&reused) out<<" reused";
  out<<"\n";
  if(STATE!=unset) {
    out<<" root center:           "<<center(root())       <<"\n"
       <<" root radius:           "<<radius(root())       <<"\n"
       <<" bodies loaded:         "<<number(root())       <<"\n"
       <<" total mass:            "<<mass  (root())       <<"\n"
       <<" cells used:            "<<Nc                   <<"\n"
       <<" zombies elimitated:    "<<Nz                   <<"\n"
       <<" maximum depth:         "<<depth(root())        <<"\n"
       <<" current theta :        "<<OPENING->theta_min() <<"\n"
       <<" current opening :      "<<OPENING->method()    <<"\n";
    if(SOFTENING == global) 
      out<<" softening:             global\n"
	 <<" softening length:      "<< EPS               <<"\n";
    else
      out<<" softening:             individual\n";
    out<<" softening kernel:      "<<describe_kernel()    <<"\n";
  }
  out<<" interaction statitics:\n"
     <<"    type            interactions\n"
     <<" # body-body : "
     <<setw(10)<<STATS->BB_interactions()
     <<setprecision(trick(STATS->BB_ia_part(),4))<<setw(7)<<STATS->BB_ia_part()
     <<"%\n # many-body : "
     <<setw(10)<<STATS->MB_interactions()
     <<setprecision(trick(STATS->MB_ia_part(),4))<<setw(7)<<STATS->MB_ia_part()
     <<"%\n # cell-body : "
     <<setw(10)<<STATS->CB_interactions()
     <<setprecision(trick(STATS->CB_ia_part(),4))<<setw(7)<<STATS->CB_ia_part()
     <<"%\n # cell-cell : "
     <<setw(10)<<STATS->CC_interactions()
     <<setprecision(trick(STATS->CC_ia_part(),4))<<setw(7)<<STATS->CC_ia_part()
     <<"%\n # total     : "
     <<setw(10)<<STATS->total_interactions()<<" 100.00%\n";
}
#endif // C_AND_F
//------------------------------------------------------------------------------
fasttree::uint tree::BB_interactions() const { return STATS->BB_interactions(); }
fasttree::uint tree::MB_interactions() const { return STATS->MB_interactions(); }
fasttree::uint tree::CB_interactions() const { return STATS->CB_interactions(); }
fasttree::uint tree::CC_interactions() const { return STATS->CC_interactions(); }
fasttree::uint tree::total_interactions () const
                                          { return STATS->total_interactions(); }
const fasttree::vect& tree::root_center() const { return center(root()); }
const fasttree::real& tree::root_radius() const { return radius(root()); }
const fasttree::indx& tree::root_depth () const { return depth (root()); }
const fasttree::uint& tree::root_number() const { return number(root()); }
const fasttree::real& tree::root_mass  () const { return mass  (root()); }
//-----------------------------------------------------------------------------+
//                                                                             |
// 4 ESTIMATION OF DENSITY AND SUCH                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace fasttree {
  inline void pass_flags_up(cell*C0, cell*CN, soul*S0, soul*SN
#ifndef USE_BODIES
			    , const int *FL
#endif // USE_BODIES
			    )
  {
#ifdef USE_BODIES
    LoopSouls Si->reset_flag();
#else  // USE_BODIES
    LoopSouls Si->reset_flag(FL);
#endif // USE_BODIES
    LoopCellsUp {
      static_cast<cell*>(Ci)->reset_sink_flag();
      LoopSubCells static_cast<cell*>(Ci)->add_sink_flag(c);
      LoopSubSouls static_cast<cell*>(Ci)->add_sink_flag(s);
    }
  }
  //----------------------------------------------------------------------------
  inline void guess_N(cell*Ci, real n, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci) / (NSUB*cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || n==zero) n = number(Ci) / (NSUB*square(radius(Ci)));
#endif // NDIM==3
    LoopSubCells if(is_sink(c)) guess_N(c,n,Nx);
    LoopSubSouls if(is_sink(s)) s->rho() = n;
  }
  //----------------------------------------------------------------------------
  inline void guess_sd(cell*Ci, real sd, const uint& Nx)
  {
    if(number(Ci)>Nx || sd==zero) sd = mass(Ci) / (4*square(radius(Ci)));
    LoopSubCells if(is_sink(c)) guess_sd(c,sd,Nx);
    LoopSubSouls if(is_sink(s)) s->rho() = sd;
  }
  //----------------------------------------------------------------------------
  inline void guess_rho(cell*Ci, real rh, const uint& Nx)
  {
#if NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci) / (NSUB*cube(radius(Ci)));
#else  // NDIM==3
    if(number(Ci)>Nx || rh==zero) rh = mass(Ci) / (NSUB*square(radius(Ci)));
#endif // NDIM==3
    LoopSubCells if(is_sink(c)) guess_rho(c,rh,Nx);
    LoopSubSouls if(is_sink(s)) s->rho() = rh;
  }
}
//------------------------------------------------------------------------------
void tree::estimate_n(const uint Nx)
{
#ifdef USE_BODIES
  pass_flags_up(C0,CN,S0,SN);                        // update flags
  guess_N(root(),zero,Nx);                           // guess n_i
#else  // USE_BODIES
  pass_flags_up(C0,CN,S0,SN,FL);                     // update flags
  guess_N(root(),zero,Nx);                           // guess n_i
  LoopSouls if(is_sink(Si))
    Si->update_dens(RH);                             // copy n_i into array RH
#endif // USE_BODIES
}
//------------------------------------------------------------------------------
void tree::estimate_sd(const uint Nx)
{
#ifdef USE_BODIES
  pass_flags_up(C0,CN,S0,SN);                        // update flags
  guess_sd(root(),zero,Nx);                          // guess SD_i
#else  // USE_BODIES
  pass_flags_up(C0,CN,S0,SN,FL);                     // update flags
  guess_sd(root(),zero,Nx);                          // guess SD_i
  LoopSouls if(is_sink(Si))
    Si->update_dens(RH);     // copy rho_i into array RH
#endif // USE_BODIES
}
//------------------------------------------------------------------------------
void tree::estimate_rho(const uint Nx)
{
#ifdef USE_BODIES
  pass_flags_up(C0,CN,S0,SN);                        // update flags
  guess_rho(root(),zero,Nx);                         // guess rho_i
#else  // USE_BODIES
  pass_flags_up(C0,CN,S0,SN,FL);                     // update flags
  guess_rho(root(),zero,Nx);                         // guess rho_i
  LoopSouls if(is_sink(Si))
    Si->update_dens(RH);     // copy rho_i into array RH
#endif // USE_BODIES
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 5 APPROXIMATION OF GRAVITY                                                  |
// ==========================                                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
namespace fasttree {
  inline uint pass_flags_and_eps_up(const opening* OPENING,
				    const tree_soft_type& SOFTENING,
				    cell*C0,cell*CN, soul*S0, soul*SN
#ifndef USE_BODIES
				    , const int*FL, const areal*EP
#endif // USE_BODIES
				    )
    // preparation: passing flags & eps up the tree, set R_crit                 
  {
#ifdef USE_BODIES
    LoopSouls Si->reset_grav();
#else  // USE_BODIES
    if(EP) { LoopSouls Si->reset_grav(FL,EP); }
    else   { LoopSouls Si->reset_grav(FL); }
#endif // USE_BODIES
    register uint n=0u;
    if(SOFTENING == individual) {
      register real Emin, Emax;
#ifdef SOFT_MEAN
      register real Eps;
#endif
      LoopCellsUp {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->reset_sink_flag();
	Emin = zero;
	Emax = zero;
#ifdef SOFT_MEAN
	Eps  = zero;
#endif
	LoopSubCells {
	  Ci->add_sink_flag(c);
	  if(Emin==zero || emin(c)<Emin) Emin = emin(c);
	  if(Emax==zero || emax(c)>Emax) Emax = emax(c);
#ifdef SOFT_MEAN
	  Eps += eps(c);
#endif
	}
	LoopSubSouls {
	  Ci->add_sink_flag(s);
	  if(Emin==zero || eps(s)<Emin) Emin = eps(s);
	  if(Emax==zero || eps(s)>Emax) Emax = eps(s);
#ifdef SOFT_MEAN
	  Eps += mass(s)*eps(s);
#endif
	}
	if(is_sink(Ci)) n++;
	Ci->emin() = Emin;
	Ci->emax() = Emax;
#ifdef SOFT_MEAN
	Ci->eps()  = Eps;
#else
	Ci->eps()  = Emin;
#endif
      }
#ifdef SOFT_MEAN
      LoopCellsUp Ci->eps() *= imass(Ci);
#endif
    } else { // SOFTENING == fixed, no cell-intern epsila
      LoopCellsUp {
	Ci->set_rcrit(OPENING->invtheta(mass(Ci)));
	Ci->reset_sink_flag();
	LoopSubCells Ci->add_sink_flag(c);
	LoopSubSouls Ci->add_sink_flag(s);
	if(is_sink(Ci)) n++;
      }
    }
    return n;
  }
  //----------------------------------------------------------------------------
  inline void eval_quadrupoles(cell*C0,cell*CN)
    // pre-computes quadrupole moments for all cells (needed for gravity only)  
  {
    register vect Xi;
    register matr quad,Qi;
    LoopCellsUp {                                  // loop tree upwards >       
#ifdef TESTBODIES
      if(is_source(Ci)) {                          //   IF source cells >       
#endif
	quad = zero;                               //     reset quadrupole      
	LoopSubSouls                               //     loop over subsouls    
#ifdef TESTBODIES
	if(is_source(s))                           //     that are sources      
#endif
	  {                                        //     >                     
	    Xi    = cofm(s)-cofm(Ci);              //       vector to subsoul   
	    Qi.outprod(Xi);                        //       add quad due to     
	    Qi   *= mass(s);                       //       sub-node's          
	    quad += Qi;                            //       position            
	  }                                        //     <                     
	LoopSubCells                               //     loop over subcells    
#ifdef TESTBODIES
	if(is_source(c))                           //     that are sources      
#endif
	  {                                        //     >                     
	    Xi    = cofm(c)-cofm(Ci);              //       vector to subcell   
	    quad += quadru(c) ;                    //       add sub-qu          
	    Qi.outprod(Xi);                        //       add quad due to     
	    Qi   *= mass(c);                       //       sub-node's          
	    quad += Qi;                            //       position            
	  }                                        //     <                     
	Ci->quadru() = quad;                       //     assign quadrupole     
#ifdef TESTBODIES
      } else                                       //   < ELSE            >     
	Ci->quadru() = zero;                       //     quadrupole = 0  <     
#endif
    }                                              // <                         
    LoopCellsDown                                  // loop cells                
#ifdef TESTBODIES
      if(is_source(Ci))                            // that are sources       >  
#endif
	Ci->quadru() *= imass(Ci);                 //   normalize quadrupole <  
  }
  //----------------------------------------------------------------------------
  inline void collect_grav(cell        *Ci,
			   TaylorGrav   G)
    // collection algorithm for gravity                                         
  {
    G.shift_and_add(cofm(Ci),coeffs(Ci),imass(Ci));
    LoopSubSouls if(is_sink(s)) G.extract_grav(s);
    LoopSubCells if(is_sink(c)) collect_grav(c,G);
  }
}
//------------------------------------------------------------------------------
// approximation: prepare, interact, and collect
void tree::approximate_gravity()
{
#ifdef USE_BODIES
  register uint Nu = pass_flags_and_eps_up(OPENING,SOFTENING,C0,CN,S0,SN);
#else  // USE_BODIES
  register uint Nu = pass_flags_and_eps_up(OPENING,SOFTENING,C0,CN,S0,SN,FL,EP);
#endif // USE_BODIES
  if(Nu==0) {
    warning(" tree::approximate_gravity(): no body wants updating");
    return;
  }
  T0 = new TaylorCoeffs[Nu];
  register TaylorCoeffs *Ti=T0;
  LoopCellsDown if(is_sink(Ci)) Ci->coeffs() = Ti++;
  STATS->reset();
  gravity_kernel K(STATS,KERNEL,EPS,SOFTENING);
  eval_quadrupoles(C0,CN);
  interact<gravity_kernel,cell,soul>(root(),&K);
  TaylorGrav G(cofm(root()));
  LoopSouls if(is_sink(Si)) Si->normalize_grav();
  collect_grav(root(),G);
  delete[] T0;
  T0 = 0;
#ifndef USE_BODIES
  LoopSouls if(is_sink(Si)) Si->update_grav(A,PH);
#endif // USE_BODIES
}
//-----------------------------------------------------------------------------+
//                                                                             |
// 6 NEIGHBOUR FINDING AND COUNTING                                            |
// ================================                                            |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifdef USE_BODIES
void tree::make_iaction_list(body_pair *bl,
			     const uint nl,
			     uint      &na,
			     const real tau)
{
  if(tau < zero) {             // SPH
    LoopSouls if(is_sph(mybody(Si))) {
      Si->flag_for_subtree();
      Si->add_body_flag();
    }
    Stree            stree(C0,CN);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    interact<neighbour_finder,Scell,Ssoul>(stree.root(),&sfind);
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls if(is_sticky(mybody(Si))) {
      Si->flag_for_subtree();
      Si->add_body_flag();
    }
    Stree         stree(C0,CN);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    interact<sticky_finder,Scell,Ssoul>(stree.root(),&sfind);
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void tree::count_sph_neighbours()
{
  LoopSouls if(is_sph(mybody(Si))) {
    Si->flag_for_subtree();
    Si->add_body_flag();
  }
  Stree             stree (C0,CN);
  neighbour_counter scount;
  stree.prepare_sph();
  interact<neighbour_counter,Scell,Ssoul>(stree.root(),&scount);
  stree.set_num(B0);
}
//-----------------------------------------------------------------------------+
#else  // USE_BODIES
//-----------------------------------------------------------------------------+
void tree::make_iaction_list(body_pair   *bl,
			     const uint   nl,
			     uint        &na,
			     const areal *s,
			     const areal *v[NDIM],
			     const real   tau)
{
  if(tau < zero) {             // SPH
    LoopSouls if(is_sph(FL[myelem(Si)])) {
      Si->flag_for_subtree();
      Si->add_body_flag(FL);
    }
    Stree            stree(C0,CN,s,v);
    neighbour_finder sfind(nl,bl);
    stree.prepare_sph();
    interact<neighbour_finder,Scell,Ssoul>(stree.root(),&sfind);
    na = sfind.actual_size_of_list();
  } else {                     // sticky particles
    LoopSouls if(is_sticky(FL[myelem(Si)])) {
      Si->flag_for_subtree();
      Si->add_body_flag(FL);
    }
    Stree         stree(C0,CN,s,v);
    sticky_finder sfind(tau,nl,bl);
    stree.prepare_sticky();
    interact<sticky_finder,Scell,Ssoul>(stree.root(),&sfind);
    na = sfind.actual_size_of_list();
  }
}
//-----------------------------------------------------------------------------+
void tree::count_sph_neighbours(const areal* s, areal* NUM)
{
  LoopSouls if(is_sph(FL[myelem(Si)])) {
    Si->flag_for_subtree();
    Si->add_body_flag(FL);
  }
  Stree             stree(C0,CN,s,0);
  neighbour_counter scount;
  stree.prepare_sph();
  interact<neighbour_counter,Scell,Ssoul>(stree.root(),&scount);
  stree.set_num(NUM);
}
#endif // USE_BODIES
////////////////////////////////////////////////////////////////////////////////
#undef LoopCellsDown
#undef LoopCellsUp
#undef LoopSubCells
#undef LoopSubSouls
#undef LoopSouls
