//-----------------------------------------------------------------------------+
//                                                                             |
//  treeC.cc                                                                   |
//                                                                             |
//  copyright Walter Dehnen, 2000/2001                                         |
//  e-mail:   dehnen@mpia.de                                                   |
//  address:  Max-Planck Institut fr Astronomie,                              |
//            Knigstuhl 17, D69117 Heidelberg Germany                         |
//                                                                             |
//-----------------------------------------------------------------------------+

extern "C" {
#include <treeC.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
}
#include <tree.h>

using namespace fasttree;

#if NDIM==3
#define POSIS   const areal*X,  const areal*Y,  const areal*Z
#define VELIS   const areal*VX, const areal*VY, const areal*VZ
#define ACCIS         areal*AX,       areal*AY,       areal*AZ
#else
#define POSIS   const areal*X,  const areal*Y
#define VELIS   const areal*VX, const areal*VY
#define ACCIS         areal*AX,       areal*AY
#endif

//==============================================================================
// auxiliary data
//==============================================================================
static tree        *TREE      = 0;     // pointer to `our' tree
static const areal *POS[NDIM] = {0};   // pointer to position arrays
static const areal *VEL[NDIM] = {0};   // pointer to velocity arrays
static areal       *ACC[NDIM] = {0};   // pointer to acceleration arrays
static bool         BUILT     = false; // is our tree build or not?
//==============================================================================
// auxiliary routines
//==============================================================================
inline void TreeC_error(const char* name)
{
  if(TREE==0) {
    printf(" TREE ERROR: %s() called before TreeC_initialize()\n",name);
    exit(1);
  }
}
//------------------------------------------------------------------------------
inline int TreeC_warning(const char* name)
{
  if(TREE==0) {
    printf("TREE WARNING: %s() called before TreeC_initialize()\n",name);
    return 1;
  }
  return 0;
}
//------------------------------------------------------------------------------
inline void TreeC_grown(const char* name)
{
  if(!BUILT) {
    printf("TREE WARNING: %s() called before a tree has been grown\n",name);
    printf("              I will grow the tree (via TreeC_grow()) first\n");
    TREE->grow();
    BUILT=true;
  }
}
//------------------------------------------------------------------------------
inline fasttree::kern_type ktype(const int K)
{
  switch(K) {
  case  0: return fasttree::f0;
  case  1: return fasttree::f1;
  case  2: return fasttree::f2;
  case  3: return fasttree::f3;
  case 10: return fasttree::k0;
  case 11: return fasttree::k1;
  case 12: return fasttree::k2;
  case 13: return fasttree::k3;
  case 20: return fasttree::p0;
  case 21: return fasttree::p1;
  case 22: return fasttree::p2;
  case 23: return fasttree::p3;
  default: printf("TREE WARNING: unknown kernel code %d; defaulting to 21",K);
           return fasttree::p1;
  }
}
//------------------------------------------------------------------------------
inline int ktype(const tree::kern_type K)
{
  switch(K) {
  case tree::f0: return  0;
  case tree::f1: return  1;
  case tree::f2: return  2;
  case tree::f3: return  3;
  case tree::k0: return 10;
  case tree::k1: return 11;
  case tree::k2: return 12;
  case tree::k3: return 13;
  case tree::p0: return 20;
  case tree::p1: return 21;
  case tree::p2: return 22;
  case tree::p3: return 23;
  default:       printf("TREE WARNING: unknown kern_type; defaulting to 21");
                 return 21;
  }
}
//==============================================================================
// routines defined in treeC.h
//==============================================================================
void TreeC_initialize(const int *F, const areal*M, POSIS, const areal*E, ACCIS,
		      areal*PH, areal*RH, const int N, const areal EPS,
		      const areal TH, const int K, const int DMAX)
{
  if(TREE) delete TREE;
  POS[0] = X; ACC[0] = AX;
  POS[1] = Y; ACC[1] = AY;
#if NDIM==3
  POS[2] = Z; ACC[2] = AZ;
#endif
  TREE  = new tree(F,M,POS,E,ACC,PH,RH,abs(N),real(EPS),real(TH),ktype(K),DMAX);
  BUILT = false;
}
//------------------------------------------------------------------------------
void TreeC_resetsoftening(const areal EPS, const int K)
{
  if(TreeC_warning("TreeC_resetsoftening")) return;
  TREE->reset_softening(real(EPS),ktype(K));
}
//------------------------------------------------------------------------------
void TreeC_resetopening(const areal TH)
{
  if(TreeC_warning("TreeC_setopening")) return;
  TREE->reset_opening(real(TH));
}  
//------------------------------------------------------------------------------
void TreeC_clearup()
{
  if(TREE) delete TREE;
  TREE =  0;
  BUILT= false;
}
//------------------------------------------------------------------------------
void TreeC_grow(const int Nc)
{
  TreeC_error("TreeC_grow");
  TREE->grow(Nc);
  BUILT=true;
}
//------------------------------------------------------------------------------
void TreeC_grow_centered(const int Nc, const int X[NDIM])
{
  TreeC_error("TreeC_grow_centered");
  vect C; 
  for(register indx i=0; i<NDIM; i++) C[i] = real(X[i]);
  TREE->grow(Nc,&C);
  BUILT=true;
}
//------------------------------------------------------------------------------
void TreeC_reuse()
{
  TreeC_error("TreeC_reuse");
  if(!BUILT) {
    printf("TREE WARNING: TreeC_reuse() called before a tree has been grown\n");
    printf("              I will use TreeC_grow() instead\n");
    TREE->grow();
    BUILT=true;
  } else
    TREE->reuse();
}
//------------------------------------------------------------------------------
void TreeC_approx_grav()
{
  TreeC_error("TreeC_approx_gravity");
  TreeC_grown("TreeC_approx_gravity");
  TREE->approximate_gravity();
}  
//------------------------------------------------------------------------------
void TreeC_estimate_rho(const int nx)
{
  TreeC_error("TreeC_estimate_rho");
  TreeC_grown("TreeC_estimate_rho");
  TREE->estimate_rho(nx);
}  
//------------------------------------------------------------------------------
void TreeC_estimate_n(const int nx)
{
  TreeC_error("TreeC_estimate_n");
  TreeC_grown("TreeC_estimate_n");
  TREE->estimate_n(nx);
}  
//------------------------------------------------------------------------------
void TreeC_iactionlist(int*I1, int*I2, const int Ni, int* Nt, const areal* S,
		       const areal tau, VELIS)
{
  TreeC_error("TreeC_iactionlist");
  TreeC_grown("TreeC_iactionlist");
  fasttree::uint na;
  tree::body_pair *bl = new tree::body_pair[Ni];
  VEL[0] = VX;
  VEL[1] = VY;
#if NDIM==3
  VEL[2] = VZ;
#endif
  TREE->make_iaction_list(bl,Ni,na,S,VEL,tau);
  const fasttree::uint iend = min(na,fasttree::uint(Ni));
  for(register fasttree::uint i=0; i<iend; i++) {
    I1[i] = bl[i][0];
    I2[i] = bl[i][1];
  }
  delete[] bl;
  *Nt=na;
}
//------------------------------------------------------------------------------
void TreeF_ialist(int*I1, int*I2, const int Ni, int* Nt, const areal* S,
		  const areal tau, VELIS)
{
  TreeC_error("TreeF_ialist");
  TreeC_grown("TreeF_ialist");
  fasttree::uint na;
  tree::body_pair *bl = new tree::body_pair[Ni];
  VEL[0] = VX;
  VEL[1] = VY;
#if NDIM==3
  VEL[2] = VZ;
#endif
  TREE->make_iaction_list(bl,Ni,na,S,VEL,tau);
  const fasttree::uint iend = min(na,fasttree::uint(Ni));
  for(register fasttree::uint i=0; i<iend; i++) {
    I1[i] = 1+bl[i][0];           // FORTRAN indices = 1 + C indices
    I2[i] = 1+bl[i][1];
  }
  delete[] bl;
  *Nt=na;
}
//------------------------------------------------------------------------------
areal TreeC_root_center(const int i)
{
  if(TreeC_warning("TreeC_root_center")) return 0.;
  return areal((TREE->root_center())[i]);
}
//------------------------------------------------------------------------------
areal TreeC_root_radius()
{
  if(TreeC_warning("TreeC_root_radius")) return 0.;
  return areal((TREE->root_radius()));
}
//------------------------------------------------------------------------------
areal TreeC_current_theta(const areal m)
{
  if(TreeC_warning("TreeC_current_theta")) return 0.;
  return areal(TREE->theta(real(m)));
}
//------------------------------------------------------------------------------
areal TreeC_current_eps()
{
  if(TreeC_warning("TreeC_current_eps")) return 0.;
  return areal(TREE->softening_length());
}
//------------------------------------------------------------------------------
int TreeC_current_kernel()
{
  if(TreeC_warning("TreeC_current_kernel")) return 0;
  return ktype(TREE->kernel());
}
//------------------------------------------------------------------------------
int TreeC_softening()
{
  if(TreeC_warning("TreeC_softening")) return 0;
  switch(TREE->softening()) {
  case fasttree::global:              return 0;
  case fasttree::individual: default: return 1;
  }
}
//------------------------------------------------------------------------------
int TreeC_No_cells()
{
  if(TreeC_warning("TreeC_No_cells")) return 0;
  return TREE->No_cells_used();
}
//------------------------------------------------------------------------------
int TreeC_depth()
{
  if(TreeC_warning("TreeC_depth")) return 0;
  return TREE->root_depth();
}
//------------------------------------------------------------------------------
void TreeC_stats()
{
  if(TreeC_warning("TreeC_stats")) return;
  if(!BUILT) {
    printf(" no tree built\n");
    return;
  }
  printf(" root center:      %f %f %f\n",TREE->root_center()[0],
	                                 TREE->root_center()[1],
	  		       	         TREE->root_center()[2]);
  printf(" root radius:      %f\n",      TREE->root_radius());
  printf(" bodies loaded:    %d\n",      TREE->root_number());
  printf(" total mass:       %f\n",      TREE->root_mass  ());
  printf(" cells used:       %d\n",      TREE->No_cells_used());
  printf(" zombies elimated  %d\n",      TREE->No_zombie_cells());
  printf(" tree depth:       %d\n",      TREE->root_depth ());
  printf(" current theta:    %f\n",      TREE->theta(TREE->root_mass()));
  printf(" current opening:  %s\n",      TREE->opening_method());
  if(TREE->softening() == tree::global) {
    printf(" softening:        global\n");
    printf(" softening lengh:  %f\n",    TREE->softening_length());
  } else {
    printf(" softening:        individual\n");
  }
  printf(" softening kernel: ");
  int kern = ktype(TREE->kernel());
  if(kern < 10)      printf("F%d\n",kern);
  else if(kern < 20) printf("K%d\n",kern-10);
  else if(kern ==20) printf("Plummer\n");
  else               printf("rho = a / [r^2+e^2]^(7/2)\n");
  printf("\n interaction statistic:\n");
  printf(" # body-body : %d\n",TREE->BB_interactions());
  printf(" # many body : %d\n",TREE->MB_interactions());
  printf(" # cell-body : %d\n",TREE->CB_interactions());
  printf(" # cell-cell : %d\n",TREE->CC_interactions());
  printf(" # total     : %d\n",TREE->total_interactions());
}
//------------------------------------------------------------------------------
#undef POSIS
#undef VELIS
#undef ACCIS
